//=========================================================
// H8SX/1655 Project
//=========================================================
// File Name : sdcard.c
// Function  : Control SD Card
//---------------------------------------------------------
// Rev.01 2009.09.18 Takanaga Yamazaki
//---------------------------------------------------------
// Copyright (C) 2009, 2010 Takanaga Yamazaki
//=========================================================
// ---- License Information -------------------------------
// Anyone can FREELY use this code fully or partially
// under conditions shown below.
// 1. You may use this code only for individual purpose,
//    and educational purpose.
//    Do not use this code for business even if partially.
// 2. You should use this code under the GNU GPL.
// 3. You should remain this header text in your codes
//   including Copyright credit and License Information.
// 4. Your codes should inherit this license information.
//=========================================================
// ---- Patent Notice -------------------------------------
// I have not cared whether this system (hw + sw) causes 
// infringement on the patent, copyright, trademark,
// or trade secret rights of others. You have all 
// responsibilities for determining if your designs
// and products infringe on the intellectual property
// rights of others, when you use technical information
// included in this system for your business.
//=========================================================
// ---- Disclaimers ---------------------------------------
// The function and reliability of this system are not 
// guaranteed. They may cause any damages to loss of
// properties, data, money, profits, life, or business.
// By adopting this system even partially, you assume
// all responsibility for its use.
//=========================================================
//=========================================================
// --- Note on SD Interface License -----------------------
// If you use SD Card for your commercial project,
// I strongly recommend you to contact to 
// the SD Card Association (www.sdcard.org)
// to confirm whether licensing is required or not
// even if you use only SPI mode (MMC style).
//=========================================================

#include <stdio.h>
#include "api.h"

//============================
// Define SD Setup State
//============================
enum SD_SETUP_STATE
{
    NONE,  // no SD card
    POWER, // powering up
    INIT,  // initializing (doing CMD0)
    READY, // ready to accept commands
};

//============================
// Define Constant
//============================
#define SD_REMOVE_CHATTER       (100 / TICK_UNIT * 4) // 0.1sec : Removal Time for SD sense Chattering * margin
#define SD_POWERUP_COUNT        (400 / TICK_UNIT * 4) // power up wait time
#define SD_RESPONSE_TIME_OUT    (           1000 * 4) // if MARK-response is returned this count times, no more waiting response * margin
#define SD_COMMAND_TIME_OUT     (           1000 * 4) // command time out * margin
#define SD_READY_TIME_OUT       (500 / TICK_UNIT * 4) // CMD1 to READY time out 500ms * margin
#define SD_BLOCK_READ_TIME_OUT  (100 / TICK_UNIT * 4) // block read  time out 100ms   * margin
//#define SD_BLOCK_WRITE_TIME_OUT (250 / TICK_UNIT * 4) // block write time out 250ms   * margin
#define SD_BLOCK_WRITE_TIME_OUT (4000 / TICK_UNIT * 4) // block write time out 250ms   * margin
#define MARK 0xff
//#define SD_BPS 400 // kbps
#define SD_BPS 100 // kbps
#define INSERT 0x01
#define UNLOCK 0x02
#define BLOCK_SIZE 512

//============================
// Global Variable
//============================
volatile UI08 gSD_Sense = 0; // bit1: Unlock, bit0: Insert
volatile UI08 gSD_Ready = 0; // SD is inserted and in Ready state

//============================
// Define SD Commands
//============================
// Each last byte shows CRC code which is effective in only CMD00 in SPI mode.
UI08 CMD00[6] = {0x40, 0x00, 0x00, 0x00, 0x00, 0x95}; // CMD00 : GO_IDLE_STATE(reset); last 0x95 is CRC code
UI08 CMD01[6] = {0x41, 0x00, 0x00, 0x00, 0x00, 0xff}; // CMD01 : SEND_OP_COND
UI08 CMD16[6] = {0x50, 0x00, 0x00, 0x00, 0x00, 0xff}; // CMD16 : SET_BLOCK_LEN
UI08 CMD17[6] = {0x51, 0x00, 0x00, 0x00, 0x00, 0xff}; // CMD17 : READ_SINGLE_BLOCK
UI08 CMD24[6] = {0x58, 0x00, 0x00, 0x00, 0x00, 0xff}; // CMD24 : WRITE_BLOCK
//
UI08 CMD09[6] = {0x49, 0x00, 0x00, 0x00, 0x00, 0xff}; // CMD09 : SEND_CSD
UI08 CMD10[6] = {0x4a, 0x00, 0x00, 0x00, 0x00, 0xff}; // CMD10 : SEND_CID
UI08 CMD27[6] = {0x5b, 0x00, 0x00, 0x00, 0x00, 0xff}; // CMD27 : PROGRAM_CSD
UI08 CMD58[6] = {0x7a, 0x00, 0x00, 0x00, 0x00, 0xff}; // CMD58 : READ_OCR

//============================
// SD Card Registers
//============================
UI08 OCR[ 4]; // Operating Condition Register
UI08 CSD[16]; // Card Specific Data Register
UI08 CID[16]; // Card Identification Register

//============================
// Define Response
//============================
#define R1_IDLE_STATE      0x01 // R1: The card is in an idle state and running initializing process.
#define R1_ERASE_RESET     0x02 // R1: An erase sequence was cleared before execution because an out-of-erase sequence command was received.
#define R1_ILLEGAL_COMMAND 0x04 // R1: An illegal command code was detected.
#define R1_COM_CRC_ERROR   0x08 // R1: The CRC check of the last command failed.
#define R1_ERASE_SEQ_ERROR 0x10 // R1: An error in the sequence of erase commands occured.
#define R1_ADDRESS_ERROR   0x20 // R1: A mis-aligned address that did not match the block length was used in the command.
#define R1_PARAMETER_ERROR 0x40 // R1: The command's argument (e.g., address, block length) was out of the allowed range for this card.
#define R1_MSB             0x80 // R1: The MSB should be always Zero in valid response.

#define R2_CARD_IS_LOCKED  0x01 // R2: This bit is set when the user locks the card. It is reset when it is unlocked.
#define R2_WP_ERASE_SKIP   0x02 // R2: This status bit has two functions. It is set when the host attempts to erase a write protected sector or if a sequence or password error occured during a card lock/unlock operation.
#define R2_ERROR           0x04 // R2: A general or an unknown error occurred during the operation.
#define R2_CC_ERROR        0x08 // R2: Internal card controller error.
#define R2_CARD_ECC_FAILED 0x10 // R2: The card's internal ESS was applied but failed to correct the data.
#define R2_WP_VIOLATION    0x20 // R2: The command tried to write to a write-protected block.
#define R2_ERASE_PARAMETER 0x40 // R2: An invalid selection, sectors, or group for erase.
#define R2_OUT_OF_RANGE    0x80 // R2: This bit has two functions. It is set if the command argument was out of its valid range, or if the host is trying to change the ROM section or reverse the copy bit (set as original) or permanent WP bit (un-protect) of the CSD register.

//=======================
// SD Read Block
//=======================
UI08 SD_Read_Block(UI32 addr, UI08 *buffer)
{
    UI32 i;
    UI08 data;
    UI08 response;
    UI08 error;
    //-----------------------------
    // Set SPI Baud Rate
    //-----------------------------
    SPI_Set_BaudRate(SD_BPS);
    //-----------------
    // Assert CS
    //-----------------
    SPI_Control_CS(SPI_CS_SD, SPI_CS_ASSERT);
    //--------------------------------
    // Send CMD17:READ_SINGLE_BLOCK
    //--------------------------------
    SD_Command_Argument(CMD17, addr);
    response = SD_Write_Command(CMD17);
    if (response != 0)
    {
        SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
        return response;
    }
    //-------------------------------------------
    // Wait for Token of Start Data Block (0xfe)
    //-------------------------------------------
    error = 17;
    for (i = 0; i < SD_BLOCK_READ_TIME_OUT; i++)
    {
        data = SPI_Rx();
        if (data == 0xfe) {error = 0; break;}
        TMR_Wait_One_Tick();
    }
    if (error)
    {
        SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
        return error;
    }
    //--------------------------------
    // Read All Bytes in the Block
    //--------------------------------
    for (i = 0; i < BLOCK_SIZE; i++)
    {
        *buffer++ = SPI_Rx();
    }
    //--------------------------------
    // Read CRC data (16bits)
    //--------------------------------
    data = SPI_Rx();
    data = SPI_Rx();
    //--------------------------------------------------------
    // SD Card requires extra 8 clocks to accept next command.
    //--------------------------------------------------------
    SPI_Tx(MARK);
    //-----------------
    // Negate CS
    //-----------------
    SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
    //-------------------
    // Return Success
    //-------------------
    return 0;
}

//=======================
// SD Write Block
//=======================
UI08 SD_Write_Block(UI32 addr, UI08 *buffer)
{
    UI32 i;
    UI08 data;
    UI08 response;
    UI08 error;

    //-----------------------------
    // Set SPI Baud Rate
    //-----------------------------
    SPI_Set_BaudRate(SD_BPS);
    //-----------------
    // Assert CS
    //-----------------
    SPI_Control_CS(SPI_CS_SD, SPI_CS_ASSERT);
    //-----------------------------
    // Send CMD24:WRITE_BLOCK
    //-----------------------------
    SD_Command_Argument(CMD24, addr);
    response = SD_Write_Command(CMD24);
    if (response != 0)
    {
        SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
        return response;
    }
    //----------------------------------------
    // Write Token of Start Data Block (0xfe)
    //----------------------------------------
    SPI_Tx(0xfe);
    //------------------------------
    // Write All Bytes in the Block
    //------------------------------
    for (i = 0; i < BLOCK_SIZE; i++)
    {
        SPI_Tx(*buffer++);
    }
    //------------------------------
    // Write CRC data (16bits)
    //------------------------------
    SPI_Tx(0xff);
    SPI_Tx(0xff);
    //------------------------------
    // Check Data Response
    //------------------------------
    response = SPI_Rx() & 0x1f;
    if (response != 0x05) // Data accepted?
    {
        SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
        return response;
    }
    //------------------------------
    // Wait until not-Busy
    //------------------------------
    error = 24;
    for (i = 0; i < SD_BLOCK_WRITE_TIME_OUT; i++)
    {
        data = SPI_Rx();
        if (data == MARK) {error = 0; break;}
    }
    if (error)
    {
        SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
        return error;
    }
    //--------------------------------------------------------
    // SD Card requires extra 8 clocks to accept next command.
    //--------------------------------------------------------
    SPI_Tx(MARK);
    //-----------------
    // Negate CS
    //-----------------
    SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
    //--------------------
    // Return Success
    //--------------------
    return 0;
}

//=======================
// SD Read Registers
//=======================
UI08 SD_Read_Register(void)
{
    UI32 i;
    UI08 data;
    UI08 response;
    UI08 error;
    //-----------------------------
    // Set SPI Baud Rate
    //-----------------------------
    SPI_Set_BaudRate(SD_BPS);
    //-----------------
    // Assert CS
    //-----------------
    SPI_Control_CS(SPI_CS_SD, SPI_CS_ASSERT);
    //--------------------
    // Send CMD09:SEND_CSD
    //--------------------
    response = SD_Write_Command(CMD09);
    if (response)
    {
        SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
        return response;
    }
    //-------------------------------------------
    // Wait for Token of Start Data Block (0xfe)
    //-------------------------------------------
    error = 9;
    for (i = 0; i < 8; i++)
    {
        data = SPI_Rx();
        if (data == 0xfe) {error = 0; break;}
    }
    if (error)
    {
        SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
        return error;
    }
    //-------------------------------------------
    // Read remained CSD bytes
    //-------------------------------------------
    for (i = 0; i < 16; i++)
    {
        CSD[i] = SPI_Rx();
    }
    //-------------------------------------------
    // Read CRC data (16bits)
    //-------------------------------------------
    data = SPI_Rx();
    data = SPI_Rx();
    //---------------------------------------------------------
    // SD Card requires extra 8 clocks to accept next command.
    //---------------------------------------------------------
    SPI_Tx(MARK);
    //--------------------
    // Send CMD10:SEND_CID
    //--------------------
    response = SD_Write_Command(CMD10);
    if (response)
    {
        SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
        return response;
    }
    //---------------------------------------------------------
    // Wait for Token of Start Data Block (0xfe)
    //---------------------------------------------------------
    error = 10;
    for (i = 0; i < 8; i++)
    {
        data = SPI_Rx();
        if (data == 0xfe) {error = 0; break;}
    }
    if (error)
    {
        SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
        return error;
    }
    //---------------------------------------------------------
    // Read remained CID bytes
    //---------------------------------------------------------
    for (i = 0; i < 16; i++)
    {
        CID[i] = SPI_Rx();
    }
    //-------------------------
    // Read CRC data (16bits)
    //-------------------------
    data = SPI_Rx();
    data = SPI_Rx();
    //---------------------------------------------------------
    // SD Card requires extra 8 clocks to accept next command.
    //---------------------------------------------------------
    SPI_Tx(MARK);
    //--------------------
    // Send CMD58:READ_OCR
    //--------------------
    response = SD_Write_Command(CMD58);
    if (response)
    {
        SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
        return response; // Response R1
    }
    //-------------------------------
    // Read OCR bytes as Response R3
    //-------------------------------
    for (i = 0; i < 4; i++)
    {
        OCR[i] = SPI_Rx();
    }
    //---------------------------------------------------------
    // SD Card requires extra 8 clocks to accept next command.
    //---------------------------------------------------------
    SPI_Tx(MARK);
    //-----------------
    // Negate CS
    //-----------------
    SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
    //----------------
    // Return Success
    //----------------
    return 0;
}

//=======================
// SD Write CSD Register
//=======================
UI08 SD_Write_CSD(void)
{
    UI32 i;
    UI08 data;
    UI08 response;
    UI08 error;
    //-----------------------------
    // Set SPI Baud Rate
    //-----------------------------
    SPI_Set_BaudRate(SD_BPS);
    //-----------------
    // Assert CS
    //-----------------
    SPI_Control_CS(SPI_CS_SD, SPI_CS_ASSERT);
    //-----------------------------
    // Send CMD27:PROGRAM_CSD
    //-----------------------------
    response = SD_Write_Command(CMD27);
    if (response)
    {
        SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
        return response;
    }
    //---------------------------------------
    // Write Token of Start Data Block (0xfe)
    //---------------------------------------
    SPI_Tx(0xfe);
    //-----------------------------
    // Write All Bytes in CSD
    //-----------------------------
    for (i = 0; i < 16; i++)
    {
        SPI_Tx(CSD[i]);
    }
    //-----------------------------
    // Write CRC data (16bits)
    //-----------------------------
    SPI_Tx(0xff);
    SPI_Tx(0xff);
    //-----------------------------
    // Wait until not-Busy
    //-----------------------------
    error = 27;
    for (i = 0; i < 8; i++)
    {
        data = SPI_Rx();
        if (data == MARK) {error = 0; break;}
    }
    if (error)
    {
        SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
        return error;
    }
    //--------------------------------------------------------
    // SD Card requires extra 8 clocks to accept next command.
    //--------------------------------------------------------
    SPI_Tx(MARK);
    //-----------------
    // Negate CS
    //-----------------
    SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
    //-----------------------------
    // Return Success
    //-----------------------------
    return 0;
}

//=======================
// SD Initialization
//=======================
UI08 SD_Init(void)
{
    UI32 i;
    UI08 response;
    UI08 error;
    //-----------------------------
    // Set SPI Baud Rate
    //-----------------------------
    SPI_Set_BaudRate(SD_BPS);
    //-----------------
    // Assert CS
    //-----------------
    SPI_Control_CS(SPI_CS_SD, SPI_CS_ASSERT);
    //---------------------------------
    // Send CMD00:GO_IDLE_STATE(reset)
    // If /CS is asserted during this command,
    // the SD Card will enter the SPI Mode.
    //---------------------------------
    error = 100;
    for (i = 0; i < SD_COMMAND_TIME_OUT; i++)
    {
        for (i = 0; i < 128; i++) SPI_Tx(MARK); // send 1024 clocks
        response = SD_Write_Command(CMD00);
        if (response == R1_IDLE_STATE) {error = 0; break;} // waiting for entering IDLE
    }
    if (error)
    {
        SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
        return error;
    }
    //---------------------------------
    // Send CMD01:SEND_OP_COND
    //---------------------------------
    error = 1;
    for (i = 0; i < SD_READY_TIME_OUT; i++)
    {
        response = SD_Write_Command(CMD01);
        if ((response & (R1_IDLE_STATE | R1_MSB)) == 0) {error = 0; break;} // waiting for end of IDLE
    }
    if (error)
    {
        SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
        return error;
    }
    //---------------------------------
    // Send CMD16:SET_BLOCKLEN
    //---------------------------------
    SD_Command_Argument(CMD16, BLOCK_SIZE);
    response = SD_Write_Command(CMD16);
    if (response)
    {
        SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
        return response;
    }
    //-----------------
    // Negate CS
    //-----------------
    SPI_Control_CS(SPI_CS_SD, SPI_CS_NEGATE);
    //---------------------------------
    // Read all SD Card Registers
    //---------------------------------
#if 0
    response = SD_Read_Register();
    if (response) return response;
    //
    // Print out Register Contents
    //
    printf("<< SD Card Registers >>\n");
    printf("OCR = ");
    for (i = 0; i <  4; i++) printf("%02x ", OCR[i]);
    printf("\nCSD = ");
    for (i = 0; i < 16; i++) printf("%02x ", CSD[i]);
    printf("\nCID = ");
    for (i = 0; i < 16; i++) printf("%02x ", CID[i]);
    printf(" \n");
#endif
    //---------------------------------
    // Return Success
    //---------------------------------
    return 0;
}

//================================
// Set Argument in Command String
//================================
void SD_Command_Argument(UI08 *command, UI32 argument)
{
    command[1] = (argument & 0xff000000) >> 24;
    command[2] = (argument & 0x00ff0000) >> 16;
    command[3] = (argument & 0x0000ff00) >>  8;
    command[4] = (argument & 0x000000ff) >>  0;
}

//=======================
// Check SD Ready
//=======================
UI08 SD_Check_Ready(void)
{
    return gSD_Ready;
}

//=======================
// SD Setup Control
//=======================
void SD_Setup_Control(void)
{
    static UI08 sdState = NONE; // State of SD Control
    static UI32 sdPowerUpCount;
    UI08 error;

    switch (sdState)
    {
        //------------
        // No SD Card
        //------------
        case NONE :
        {
            if (SD_Read_Socket(INSERT)) // inserted?
            {
                sdState = POWER;
                sdPowerUpCount = 0;
            }
            else
            {
                gSD_Ready = 0;
                sdState = NONE;
            }
            break;
        }
        //------------
        // Powering Up
        //------------
        case POWER :
        {
            if (SD_Read_Socket(INSERT)) // inserted?
            {
                // Powering Up
                if (sdPowerUpCount < SD_POWERUP_COUNT)
                {
                    sdPowerUpCount++;
                    TMR_Wait_One_Tick();
                    break;
                }
                // finished power-up
                sdState = INIT;
            }
            else
            {
                sdState = NONE;
            }
            break;
        }
        //----------------
        // Initialization
        //----------------
        case INIT :
        {
            if (SD_Read_Socket(INSERT)) // inserted?
            {
                // Initialize SD Card
                error = SD_Init();
                if (error)
                {
                  //printf("***** SD Card Setup Error ***** %d\n", error);
                    sdState = NONE;  // somthing error
                }
                else
                {
                    sdState = READY; // initialization OK
                }
            }
            else
            {
                sdState = NONE;
            }
            break;
        }
        //----------------
        // Ready
        //----------------
        case READY :
        {
            if (SD_Read_Socket(INSERT)) // inserted?
            {
                gSD_Ready = 1;
            }
            else
            {
                gSD_Ready = 0;
                sdState = NONE;
            }
            break;
        }
        //------------------------
        // Others (illegal state)
        //------------------------
        default:
        {
            sdState = NONE; // return back to NONE because of illegal
            gSD_Ready = 0;
            break;
        }
    }
}

//=======================
// SD Write Command
//=======================
UI08 SD_Write_Command(UI08 *command)
{
    UI32 i;
    UI08 response;

    //------------------------
    // Send Command (6bytes)
    //------------------------
    for (i = 0; i < 6; i++)
    {
        SPI_Tx(*command++);
    }
    //------------------------
    // Wait for Response
    //------------------------
    for (i = 0; i < SD_RESPONSE_TIME_OUT; i++)
    {
        response = SPI_Rx();
        if (response != MARK) break;
    }
    //---------------------------------------------------------
    // SD Card requires extra 8 clocks to accept next command.
    //---------------------------------------------------------
    SPI_Tx(MARK);
    //---------------------------------
    // Return response even if time up
    //---------------------------------
    return response;
}

//=================================
// SD Read Socket Status
//=================================
UI08 SD_Read_Socket(UI08 what)
{
    return (gSD_Sense & what);
}

//=================================
// SD Sense (Called every TMR Tick)
//=================================
void SD_Sense(void)
{
    static UI08 sdResult = 0;
    static UI08 sdPrev   = 0;
    static UI32 sdCount  = 0;
    UI08 sdRaw;

    //---------------------------------
    // If card removed, return quickly
    //---------------------------------
    if (!Check_SD_Insert())
    {
        sdResult  = 0;
        sdPrev    = 0;
        sdCount   = 0;
        gSD_Sense = 0;
        return;
    }
    //---------------------------------
    // Remove Chattering on Sense Switch of Socket
    //---------------------------------
    if (sdCount < SD_REMOVE_CHATTER)
    {
        sdCount++;
        gSD_Sense = sdResult;
        return;
    }
    //---------------------------------
    // Make qualified Result
    //---------------------------------
    sdRaw = Check_SD_Unlock() * 2 + Check_SD_Insert();
    //
    sdCount = 0;
    //
    if (sdPrev == sdRaw)
    {
        sdResult = sdPrev;
    }
    else
    {
        sdPrev = sdRaw;
    }
    //
    gSD_Sense = sdResult;
}

//=================================
// Check SD Card Insert
//=================================
UI08 Check_SD_Insert(void)
{
    UI08 result;
    
    result = (P5.PORT.BIT.B4)? 0 : 1; // P54 is low, when inserted
    return result;
}

//=================================
// Check SD Card Unlock
//=================================
UI08 Check_SD_Unlock(void)
{
    UI08 result;

    result = (P5.PORT.BIT.B5)? 0 : 1; // P55 is low, when write enabled
    return result;
}

//======================================================
// End of Program
//======================================================
