//=========================================================
// H8SX/1655 Project
//=========================================================
// File Name : spi.c
// Function  : Control SPI (SCI2)
//---------------------------------------------------------
// Rev.01 2009.09.05 Takanaga Yamazaki
//---------------------------------------------------------
// Copyright (C) 2009, 2010 Takanaga Yamazaki
//=========================================================
// ---- License Information -------------------------------
// Anyone can FREELY use this code fully or partially
// under conditions shown below.
// 1. You may use this code only for individual purpose,
//    and educational purpose.
//    Do not use this code for business even if partially.
// 2. You should use this code under the GNU GPL.
// 3. You should remain this header text in your codes
//   including Copyright credit and License Information.
// 4. Your codes should inherit this license information.
//=========================================================
// ---- Patent Notice -------------------------------------
// I have not cared whether this system (hw + sw) causes 
// infringement on the patent, copyright, trademark,
// or trade secret rights of others. You have all 
// responsibilities for determining if your designs
// and products infringe on the intellectual property
// rights of others, when you use technical information
// included in this system for your business.
//=========================================================
// ---- Disclaimers ---------------------------------------
// The function and reliability of this system are not 
// guaranteed. They may cause any damages to loss of
// properties, data, money, profits, life, or business.
// By adopting this system even partially, you assume
// all responsibility for its use.
//=========================================================

#include "api.h"

//=========================
// Define Initial Baud Rate
//=========================
#define SPI_INITIAL_BAUD_RATE 100 // kbps

//=============================
// Initialize SPI (SCI2)
//=============================
void Init_SPI(void)
{
    //----------------------------
    // Disable INterrupt
    //----------------------------
    INTC.IPRL.BIT._SCI2 = 0;
    //------------------------
    // Start SCI2 System Clock
    //------------------------
    MSTP.CRB.BIT._SCI2 = 0;
    //----------------------------
    // P10 : TxD2 MOSI(SD/TFT/RTC)
    // P11 : RxD2 MISO(SD/TFT/RTC)
    // P12 : SCK2 SCK (SD/TFT/RTC)
    //----------------------------
    SCI2.SCR.BIT.TE = 0;
    SCI2.SCR.BIT.RE = 0;
    //----------------------------
    // Clocked Serial Mode
    //----------------------------
    SCI2.SMR.BYTE      = 0x80; // 10000000
    SCI2.SCMR.BIT.SDIR = 1;    // MSB first
    //----------------------------
    //Baud Rate
    //----------------------------
    SPI_Set_BaudRate(SPI_INITIAL_BAUD_RATE);
    //----------------------------
    // Set TE & RE
    //----------------------------
    SCI2.SCR.BIT.TE = 1;
    SCI2.SCR.BIT.RE = 1;
}

//=====================
// Set SPI Baud Rate
//=====================
void SPI_Set_BaudRate(UI32 kbps)
{
    static UI32 kbps_curr = SPI_INITIAL_BAUD_RATE;
    SI32 brr;

    //---------------------
    // If kbps is changed?
    //---------------------
    if (kbps != kbps_curr)
    {
        //--------------------
        // Store Current kbps
        //--------------------
        kbps_curr = kbps;
        //--------------
        // Wait for Tx
        //--------------
        // Wait for TDRE = 1
        while(SCI2.SSR.BIT.TDRE == 0);
        // Wait for TEND = 1
        while(SCI2.SSR.BIT.TEND == 0);
        //----------------
        // Clear TE & RE
        //----------------
        SCI2.SCR.BIT.TE = 0;
        SCI2.SCR.BIT.RE = 0;
        //----------------------------
        // Wait for 1bit period
        //----------------------------
        CPU_Wait_N_us(1000 / kbps);
        //---------------
        // Clear Flags
        //---------------
        SCI2.SSR.BIT.RDRF = 0;
        SCI2.SSR.BIT.ORER = 0;
        SCI2.SSR.BIT.FER  = 0;
        SCI2.SSR.BIT.PER  = 0;
        //----------------------------
        // Set Baud Rate
        //   PCLK = 24MHz
        //   CKS  = 0
        //   BRR = 24 * 1000 * 1000 / (8 * (2^0-1) * 1000 * kbps) - 1
        //----------------------------
        brr = (6000 / (SI32) kbps) - 1; //(24 * 1000) / (8 / 2 * kbps) - 1;
        if (brr <   0) brr = 0;
        if (brr > 255) brr = 255;
        //
        SCI2.SCR.BIT.CKE = 0;
        SCI2.SMR.BIT.CKS = 0;
        SCI2.BRR = (UI08) brr;
        //----------------------------
        // Wait for 1bit period
        //----------------------------
        CPU_Wait_N_us(1000 / kbps);
        //----------------
        // Set TE & RE
        //----------------
        SCI2.SCR.BIT.TE = 1;
        SCI2.SCR.BIT.RE = 1;
    }
}

//=====================
// SPI Control CS
//=====================
void SPI_Control_CS(UI08 which, UI08 how)
{
    //
    // Set each CS level
    //   P13 : /CS of SD Card
    //   PM2 : /CS of LCD_QVGA
    //   PA2 :  CS of RTC(DS1305)
    //
    if (how == SPI_CS_NEGATE)
    {
        P1.DR.BIT.B3 = 1; // low active
        while(P1.PORT.BIT.B3 == 0);
        PM.DR.BIT.B2 = 1; // low active
        while(PM.PORT.BIT.B2 == 0);
        PA.DR.BIT.B2 = 0; // high active
        while(PA.PORT.BIT.B2 == 1);
    }
    else if (which == SPI_CS_SD)
    {
        PM.DR.BIT.B2 = 1; // low active
        while(PM.PORT.BIT.B2 == 0);
        PA.DR.BIT.B2 = 0; // high active
        while(PA.PORT.BIT.B2 == 1);
        //
        P1.DR.BIT.B3 = 0; // low active
        while(P1.PORT.BIT.B3 == 1);
    }
    else if (which == SPI_CS_TFT)
    {
        P1.DR.BIT.B3 = 1; // low active
        while(P1.PORT.BIT.B3 == 0);
        PA.DR.BIT.B2 = 0; // high active
        while(PA.PORT.BIT.B2 == 1);
        //
        PM.DR.BIT.B2 = 0; // low active
        while(PM.PORT.BIT.B2 == 1);
    }
    else if (which == SPI_CS_RTC)
    {
        P1.DR.BIT.B3 = 1; // low active
        while(P1.PORT.BIT.B3 == 0);
        PM.DR.BIT.B2 = 1; // low active
        while(PM.PORT.BIT.B2 == 0);
        //
        PA.DR.BIT.B2 = 1; // high active
        while(PA.PORT.BIT.B2 == 0);
    }
    else
    {
        P1.DR.BIT.B3 = 1; // low active
        while(P1.PORT.BIT.B3 == 0);
        PM.DR.BIT.B2 = 1; // low active
        while(PM.PORT.BIT.B2 == 0);
        PA.DR.BIT.B2 = 0; // high active
        while(PA.PORT.BIT.B2 == 1);
    }
}

//==========================
// SPI Bit Reverse (unused)
//==========================
UI08 SPI_Bit_Reverse(UI08 in)
{
#if 0
    UI08 i;
    UI08 b;
    UI08 out;

    out = 0;
    for (i = 0; i < 8; i++)
    {
        b = (in >> i) & 0x01;
        out = out | (b << (7 - i));
    }
    return out;
#else
    UI08 out;
    asm ("mov.l er0,@-er7");
    asm ("mov %0l, r0l"::"r"(in));
    asm ("rotxl.b r0l");
    asm ("rotxr.b r0h");
    asm ("rotxl.b r0l");
    asm ("rotxr.b r0h");
    asm ("rotxl.b r0l");
    asm ("rotxr.b r0h");
    asm ("rotxl.b r0l");
    asm ("rotxr.b r0h");
    asm ("rotxl.b r0l");
    asm ("rotxr.b r0h");
    asm ("rotxl.b r0l");
    asm ("rotxr.b r0h");
    asm ("rotxl.b r0l");
    asm ("rotxr.b r0h");
    asm ("rotxl.b r0l");
    asm ("rotxr.b r0h");
    asm ("mov r0h, %0l":"=r"(out):);
    asm ("mov.l @er7+,er0");
    return out;
#endif
}

//=========================
// SPI Tx and Rx a Byte
//=========================
UI08 SPI_TxRx(UI08 byte_tx)
{
    UI08 byte_rx;
    //
    // Tx
    //
    // Wait for TDRE = 1
    while(SCI2.SSR.BIT.TDRE == 0);
    // Set TDR
    SCI2.TDR = byte_tx;
    // Clear TDRE
    SCI2.SSR.BIT.TDRE = 0; // read 1 and write 0
    // Wait for TEND = 1
    while(SCI2.SSR.BIT.TEND == 0);
    // Clear TEND
    SCI2.SSR.BIT.TEND = 0; // read 1 and write 0
    //
    // Rx
    //
    // Wait for RDRF = 1
    while(SCI2.SSR.BIT.RDRF == 0);
    // If Over Run found, Clear ORER
    if (SCI2.SSR.BIT.ORER) SCI2.SSR.BIT.ORER = 0;
    // Get RDR
    byte_rx = SCI2.RDR;
    // Clear RDRF
    SCI2.SSR.BIT.RDRF = 0; // read 1 and write 0

    return byte_rx;
}

//=========================
// SPI Tx a Byte
//=========================
void SPI_Tx(UI08 byte)
{
    SPI_TxRx(byte);
}

//=========================
// SPI Rx a Byte
//=========================
UI08 SPI_Rx(void)
{
    return SPI_TxRx(0xff);
}

//======================================================
// End of Program
//======================================================
