//=========================================================
// H8SX/1655 Project
//=========================================================
// File Name : fat16.c
// Function  : FAT16 File System
//---------------------------------------------------------
// Rev.01 2009.09.19 Takanaga Yamazaki
//---------------------------------------------------------
// Copyright (C) 2009, 2010 Takanaga Yamazaki
//=========================================================
// ---- License Information -------------------------------
// Anyone can FREELY use this code fully or partially
// under conditions shown below.
// 1. You may use this code only for individual purpose,
//    and educational purpose.
//    Do not use this code for business even if partially.
// 2. You should use this code under the GNU GPL.
// 3. You should remain this header text in your codes
//   including Copyright credit and License Information.
// 4. Your codes should inherit this license information.
//=========================================================
// ---- Patent Notice -------------------------------------
// I have not cared whether this system (hw + sw) causes 
// infringement on the patent, copyright, trademark,
// or trade secret rights of others. You have all 
// responsibilities for determining if your designs
// and products infringe on the intellectual property
// rights of others, when you use technical information
// included in this system for your business.
//=========================================================
// ---- Disclaimers ---------------------------------------
// The function and reliability of this system are not 
// guaranteed. They may cause any damages to loss of
// properties, data, money, profits, life, or business.
// By adopting this system even partially, you assume
// all responsibility for its use.
//=========================================================

#include <stdio.h>
#include <string.h>
#include "api.h"

//----------------
// Define Constant
//----------------
#define SD_BLOCK_SIZE 9 // 2^9=512

//--------------
// Mount Command
//--------------
#define DO_UNMOUNT  0
#define DO_MOUNT    1

//--------------------
// Define FAT contents
//--------------------
#define UNUSED_CLUSTER   0x0000 // unused cluster
#define RESERVED_CLUSTER 0x0001 // reserved cluster
#define BAD_CLUSTER      0xfff7 // corrupted cluster
#define EOC              0xfff8 // end of clusters (equal to or larger than this)
#define END_OF_CLUSTER   0xffff // end of clusters (use this when write FAT table)

//----------------
// File Attributes
//----------------
#define ATTR_READ_ONLY 0x01
#define ATTR_HIDDEN    0x02
#define ATTR_SYSTEM    0x04
#define ATTR_VOLUME_ID 0x08
#define ATTR_DIRECTORY 0x10
#define ATTR_ARCHIVE   0x20
#define ATTR_LONG_NAME 0x0f

//-----------------------------------
// Define Partition Entry Description 
//-----------------------------------
typedef struct
{
    UI08 bootDescriptor;       // 1byte  : 0x00=Non-bootable Device, 0x80=Bootable Device
    UI32 firstPartitionSector; // 3bytes : Address of first sector
    UI08 fileSystemDescriptor; // 1byte  : 0=Empty, 1=FAT12(<16MB), 4=FAT16(<32MB), 5=Extended DOS, 6=FAT16(>=32MB)
    UI32 lastPartitionSector;  // 3bytes : Address of last sector
    UI32 firstSectorPosition;  // 4bytes : Number of First Sector (linear address)
    UI32 numberOfSectors;      // 4bytes : Number of Sectors in the Partition
} sPartitionEntry;

//-------------------------------------------------------
// Define Boot Sector (BS) and BIOS Parameter Block (BPB)
//-------------------------------------------------------
typedef struct
{
  //UI08 BS_jmpBoot[3];    // Jump Instruction to boot code     (um... it is x86 culture... x_x;)
  //UI08 BS_OEMName[8];    // "MSWIN4.1" is recommended setting (um... it is MS  culture... x_x;)
    UI16 BPB_BytsPerSec;   // Count of bytes per Sector
    UI08 BPB_SecPerClus;   // Number of Sectors per Allocation Unit which must be a power of 2
    UI16 BPB_RsvdSecCnt;   // Number of Reserved Sectors in the Reserved Region of the volume starting at the first sector of the volume
    UI08 BPB_NumFATs;      // Count of FAT data structures on the volume
    UI16 BPB_RootEntCnt;   // Count of 32byte directory entries in the root directory for FAT12 and FAT16 volumes
    UI16 BPB_TotSec16;     // Old 16-bit total count of sectors on the volume
    UI08 BPB_Media;        // 0xF8 for fixed media, 0xF0 for removable media
    UI16 BPB_FATSz16;      // FAT12/FAT16 16bit-count of sectors occupied by ONE FAT
  //UI16 BPB_SecPerTrk;    // Sectors per Track for interrupt 0x13 (um... it is Wintel culture... x_x;)
  //UI16 BPB_NumHeads;     // Number of Heads for interrupt 0x13   (um... it is Wintel culture... x_x;)
    UI32 BPB_HiddSec;      // Count of Hidden Sectors preceding the partition that contains this FAT volume
    UI32 BPB_TotSec32;     // New 32-bit total count of sectors on the volume
    // Following structure is prepared for FAT12 and FAT16 (not for FAT32)
  //UI08 BS_DrvNum;        // Int0x13 drive number   (um... it is Wintel culture... x_x;)
  //UI08 BS_Reserved1;     // Reserved for WindowsNT (um... it is MS culture... x_x;)
  //UI08 BS_BootSig;       // Extended Boot Signature (0x29) which indicates following three fields in the boot sector are present
  //UI32 BS_VolID;         // Volume Serial Number
  //UI08 BS_VolLab[11];    // Volume Label
  //UI08 BS_FilSysType[8]; // Type of File System (it is no more than a hint)
} sBootSectorInfo;

//-----------------------
// Define FAT Information
//-----------------------
typedef struct
{
    UI32 RootDirSectors;     // Count of Sectors occupied by the Root Directory
    UI32 FATSz;              // FAT16 16bit-count of sectors occupied by ONE FAT
    UI32 FirstDataSector;    // Start of the data region, the first sector of cluster 2
    UI32 TotSec;             // Total count of sectors on the volume
    UI32 DataSec;            // Count of Sectors in the data region of the volume
    UI32 CountofClusters;    // Count of Clusters
    UI32 FirstRootDirSecNum; // First Sector of Root Directory
} sFATInfo;

//-----------------------------
// Define Directory Information
//-----------------------------
typedef struct
{
    UI08 DIR_Name[11];     // Short Name
    UI08 DIR_Attr;         // File Attribute
  //UI08 DIR_NTRes;        // Reserved for use by WindowsNT (um... it is MS  culture... x_x;)
    UI08 DIR_CrtTimeTenth; // Milisecond stamp at file creation time. Shows counts of 100ms (0~199) in 2 seconds.
    UI16 DIR_CrtTime;      // Time file was created.
                           //     bit15-11 Hours  , valid range 0-23 inclusive
                           //     bit10- 5 Minutes, valid range 0-59 inclusive
                           //     bit 4- 0 2-seconds count, valid range 0-29 inclusive (0-58 seconds)
    UI16 DIR_CrtDate;      // Date file was created.
                           //     bit15- 9 Count of years from 1980, valid range 0-127 inclusive (1980-2107), found 2107 problem!
                           //     bit 8- 5 Month of year, 1=January, valid range 1-12 inclusive
                           //     bit 4- 0 Day of month, valid range 1-31 inclusive.
    UI16 DIR_LstAccDate;   // Last Access Date (same as DIR_WrtDate in this application)
    UI16 DIR_FstClusHI;    // High word of this entry's first cluster number (always 0 for a FAT12/FAT16 volume)
    UI16 DIR_WrtTime;      // Time of last write
    UI16 DIR_WrtDate;      // Date of last write
    UI16 DIR_FstClusLO;    // Low word of this entry's first cluster number
    UI32 DIR_FileSize;     // 32bit DWORD holding this file's size in bytes
} sDirInfo;

//--------------------
// Define Sector Cache
//--------------------
typedef struct
{
    UI08 buffer[512]; // Sector Buffer
    UI32 sector;      // Tagged Sector Number
    UI08 valid;       // Valid Flag
    UI08 dirty;       // Dirty Flag
} sSectorCache;

//------------------------
// Global File Information
//------------------------
sPartitionEntry gPT; // Volume Partition Entry
sBootSectorInfo gBS; // Boot Sector and BIOS Parameter Block
sFATInfo        gFI; // FAT Structure Information

//---------------------
// Global Sector Buffer
//---------------------
sSectorCache gFatCache;  // FAT Table Cache
sSectorCache gDatCache;  // Sector Data Cache

//-----------------
// Global Variables
//-----------------
UI08 *gpStrTargetDir = (UI08*) "H8SX1655   "; // Target Directory Name (must be 8 chars + 3 spaces)
extern sDateTime gDateTime; // Current Clock Database

//=====================
// Define Prototype
//=====================
//UI08 Load_File_from_TargetDir(UI08 *pFileName_11, UI08 (*Load_Data)(UI08 data, UI08 end));
//UI08 Save_File_in_TargetDir(UI08 *pFileName_11, UI08 (*Save_Data)(UI08 *data));
UI08 Get_File_Info_from_TargetDir(UI32 nth, sDirInfo *psDirInfo);
UI08 Load_File_from_TargetDir_by_DirInfo(sDirInfo *psDirInfo, UI08 (*Load_Data)(UI08 data, UI08 end));
UI08 Make_New_File_Name_in_TargetDir(UI08 *pFileName_11);
UI08 Search_TargetDir_from_RootDir(sDirInfo *psDirInfo);
UI08 Create_TargetDir_in_RootDir(sDirInfo *psDirInfo);
void Fill_Date_Time_in_DirInfo(sDirInfo *psDirInfo);
UI08 Read_Dir_Info(sSectorCache *cache, UI32 sector, UI32 entry, sDirInfo *psDirInfo);
UI08 Write_Dir_Info(sSectorCache *cache, UI32 sector, UI32 entry, sDirInfo *psDirInfo);
UI08 Read_FAT_Value(UI32 cluster, UI16 *FAT16ClusEntryVal);
UI08 Write_FAT_Value(UI32 cluster, UI16 FAT16ClusEntryVal);
UI32 First_Sector_of_Cluster(UI32 cluster);
UI08 Check_FAT_Data_Structure(void);
UI08 Mount_Volume(UI08 mount);
UI08 Purge_Cache(sSectorCache *cache);
UI08 Read_Data_from_Cache(sSectorCache *cache, UI32 sector, UI32 offset, UI08 *data);
UI08 Write_Data_to_Cache(sSectorCache *cache, UI32 sector, UI32 offset, UI08 data);
UI08 Read_Volume_Info(void);
void Print_Volume_Info(void);
void Print_Dir_Info(sDirInfo *psDirInfo);
void Dump_Sector(UI32 sector);
void Dump_Cache(sSectorCache *cache, UI32 sector);

//---------------------------------------------------------------------------------
// Load a File from the Target Directory
//---------------------------------------------------------------------------------
// In the argument list of this function,
// you can find interface function to receive the data stream to be saved;
//     void (*Load_Data)(UI08 data, UI08 end) ... it is pointer of a function
// The specification is shown below.
//     Input  : data ... a loaded streaming byte for the system
//              end .... if not 0, this "data" is final one
//     Output : if 1, stop the loading
//---------------------------------------------------------------------------------
UI08 Load_File_from_TargetDir(UI08 *pFileName_11, UI08 (*Load_Data)(UI08 data, UI08 end))
{
    sDirInfo    dirInfo;
    UI32        fileIndex;
    UI08        found;
    UI08        error;
    //
    // Find the File Name from Target Directory
    //
    found = 0;
    fileIndex = 0;
    while(1)
    {
        error = Get_File_Info_from_TargetDir(fileIndex, &dirInfo);
        if (error) break; // no file
        fileIndex++;
        //
        if (strncmp((char*) pFileName_11, (char*) dirInfo.DIR_Name, 11) == 0)
        {
            found = 1;
            break;
        }
    }
    //
    // If not found ,return 10
    //
    if (found == 0) return 10;
    //
    // Load the File
    //
    error = Load_File_from_TargetDir_by_DirInfo(&dirInfo, Load_Data);
    return error;
}
//
UI08 Load_File_from_TargetDir_by_DirInfo(sDirInfo *psDirInfo, UI08 (*Load_Data)(UI08 data, UI08 end))
{
    UI08     error;
    UI32     cluster;
    UI32     lastCluster;
    UI16     fat;
    UI32     sector;
    UI32     offset;
    UI08     data;
    UI32     fileSize;
    UI08     finish;
    //
    // Search all Clusters for the File Body along with FAT 
    // until End of File (judged by DIR_FileSize)
    //
    cluster = psDirInfo->DIR_FstClusLO; // first cluster of File Body
    fileSize = 0;
    finish = 0;
    while(1)
    {
        lastCluster = cluster;
        //
        // search all sectors in the cluster
        //
        for (sector = First_Sector_of_Cluster(cluster); sector < (First_Sector_of_Cluster(cluster) + gBS.BPB_SecPerClus); sector++)
        {
            //
            // load each byte from the sector
            //
            for (offset = 0; offset < gBS.BPB_BytsPerSec; offset++)
            {
                //
                // EOF?
                //
                if (fileSize >= psDirInfo->DIR_FileSize)
                {
                    return 0; // already reached to the size, success end
                }
                //
                // Read one byte from the File
                //
                error = Read_Data_from_Cache(&gDatCache, sector, offset, &data);
                if (error) return 2; // Data Cache Read Error
                //
                // Increment file size
                //
                fileSize = fileSize + 1;
                //
                // Send the data to system
                //
                finish = (*Load_Data)(data, ((fileSize >= psDirInfo->DIR_FileSize)));
                if (finish) break;
            }
            if (finish) break;
        }
        if (finish) break;
        //
        // next cluster
        //
        error = Read_FAT_Value(lastCluster, &fat);
        if (fat >= EOC) break; // no more cluster, just EOF
        cluster = (UI32) fat;
    }
    //
    // Success
    //
    return 0;
}

//---------------------------------------------------------------------------------
// Save a File in the Target Directory
//---------------------------------------------------------------------------------
// In this system, newly saved file always has unique name in the target directory.
// Therefore, this routine does not have "over-write" function.
// If you want to implement "over-write", please make "Delete" routine,
// and call it before this "Save File" routine if same name file exists.
//---------------------------------------------------------------------------------
// In the argument list of this function,
// you can find interface function to receive the data stream to be saved;
//     UI08 (*Save_Data)(UI08 *data) ... it is pointer of a function
// The specification is shown below.
//     Input  : *data .......... a streaming byte to be saved from the system
//     Output : Return value ... 0=continuing stream data
//                               1=no more stream data (*data is last one)
//---------------------------------------------------------------------------------
UI08 Save_File_in_TargetDir(UI08 *pFileName_11, UI08 (*Save_Data)(UI08 *data))
{
    UI08 error;
    sDirInfo dirInfo;
    UI32     cluster;
    UI32     lastCluster;
    UI32     sector;
    UI32     entry = 0; // initialize to avoid compile warning
    UI32     dirSector;
    UI32     dirEntry;
    UI32     offset;
    UI16     fat;
    UI08     found;
    UI08     i;
    UI08     bodyData;
    UI08     bodyStop;
    UI08     finish;
    //
    // Search Target Directory from Root Directory
    // if not found, Create Target Directory
    //
    error =  Search_TargetDir_from_RootDir(&dirInfo);
    if (error == 1)
    {
         error = Create_TargetDir_in_RootDir(&dirInfo);
         if (error) return 2; // can not create target directory
    }
    else if (error) return 2; // Directory access error
    //
    // Search all Clusters for the Target Directory along with FAT 
    // until finding vacant cluster for the file information
    //
    found = 0;
    cluster = dirInfo.DIR_FstClusLO; // first cluster
    //
    while(1)
    {
        lastCluster = cluster;
        //
        // search all sectors in the cluster
        //
        for (sector = First_Sector_of_Cluster(cluster); sector < (First_Sector_of_Cluster(cluster) + gBS.BPB_SecPerClus); sector++)
        {
            //
            // search all entries in the sector
            //
            for (entry = 0; entry < gBS.BPB_BytsPerSec / 32; entry++)
            {
                error = Read_Dir_Info(&gDatCache, sector, entry, &dirInfo); // get directory information
                if (error) return 2; // Directory read error
                if (dirInfo.DIR_Name[0] == 0)    {found = 1; break;} // found vacant entry
                if (dirInfo.DIR_Name[0] == 0xe5) {found = 1; break;} // found vacant entry 
            }
            if (found) break;
        }
        if (found) break;
        //
        // final cluster?
        //
        error = Read_FAT_Value(cluster, &fat);
        if (error) return 2; // FAT read error
        if (fat >= EOC) break;
        //
        // next cluster
        //
        cluster = (UI32) fat; // next cluster
    }
    //
    // Found the Room? or Not?
    //
    if (found) // use current sector and entry to save Directory Information
    {
        dirSector = sector;
        dirEntry  = entry;
    }
    else // add new cluster for the target directory
    {
        //
        // Search vacant Cluster in FAT Table
        //
        found = 0;
        for (cluster = 2; cluster < gFI.CountofClusters + 2; cluster++)
        {
            error = Read_FAT_Value(cluster, &fat);
            if (error) return 2; // FAT Read Error
            if (fat == UNUSED_CLUSTER)
            {
                found = 1;
                break;
            }
        }
        if (found == 0)
        {
            error = Mount_Volume(DO_UNMOUNT); // Unmount Volume
            if (error) return 3; // Volume can not be Un-Mounted
            return 1; // may be Volume Full
        }
        //
        // Update FAT Table of Target Directory to make the link
        //
        error = Write_FAT_Value(lastCluster, (UI16) cluster);
        if (error) return 2; // FAT Write Error
        error = Write_FAT_Value(cluster, END_OF_CLUSTER);
        if (error)  return 2;// FAT Write Error
        //
        // Clear all sector data in new cluster of Target Directory
        //
        for (sector = First_Sector_of_Cluster(cluster); sector < (First_Sector_of_Cluster(cluster) + gBS.BPB_SecPerClus); sector++)
        {
            for (offset = 0; offset < gBS.BPB_BytsPerSec; offset++)
            {
                error = Write_Data_to_Cache(&gDatCache, sector, offset, 0); // zero clear
                if (error) return 2; // Dir Write Error
            }
        }
        //
        dirSector = First_Sector_of_Cluster(cluster);
        dirEntry  = 0;
    }
    //
    // Search 1st vacant Cluster in FAT Table for File Body
    //
    found = 0;
    for (cluster = 2; cluster < gFI.CountofClusters + 2; cluster++)
    {
        error = Read_FAT_Value(cluster, &fat);
        if (error)  return 2;// FAT Read Error
        if (fat == UNUSED_CLUSTER) {found = 1; break;}
    }
    if (found == 0) return 1; // may be Volume Full
    //
    // Write END_OF_CLUSTER to the FAT entry
    //
    error = Write_FAT_Value(cluster, END_OF_CLUSTER);
    if (error) return 2; // FAT Write Error
    //
    // So far, Write File Information in Target Directory
    //
    for (i = 0; i < 11; i++) dirInfo.DIR_Name[i] = pFileName_11[i];
    dirInfo.DIR_Attr      = ATTR_ARCHIVE;
    dirInfo.DIR_FstClusHI = 0;
    dirInfo.DIR_FstClusLO = cluster;
    dirInfo.DIR_FileSize  = 0; // size=0, first of all
    Fill_Date_Time_in_DirInfo(&dirInfo); // save create time       
    //
    error = Write_Dir_Info(&gDatCache, dirSector, dirEntry, &dirInfo);
    if (error)  return 2;  // Dir Write Error
    //
    // Write the File Body
    //
    finish = 0;
    while(1)
    {
        lastCluster = cluster;
        //
        // Write each data to the cluster (to sectors in the cluster)
        //
        for (sector = First_Sector_of_Cluster(cluster); sector < (First_Sector_of_Cluster(cluster) + gBS.BPB_SecPerClus); sector++)
        {
            for (offset = 0; offset < gBS.BPB_BytsPerSec; offset++)
            {
                //
                // Receive one byte from system
                //
                bodyStop = (*Save_Data)(&bodyData);
                //
                // write to data cache
                //
                error = Write_Data_to_Cache(&gDatCache, sector, offset, bodyData);
                if (error) return 2; // Data Write Error
                //
                // Increment File Size
                //
                dirInfo.DIR_FileSize = dirInfo.DIR_FileSize + 1; // unit is byte
                //
                // EOF ?
                //
                if (bodyStop) {finish = 1; break;}
            }
            if (finish) break;
        }
        if (finish) break;
        //         
        // One cluster has been filled, so add a new Cluster in FAT table
        // Now, Search vacant cluster from current cluster in FAT Table
        //
        found = 0;
        for (cluster = lastCluster; cluster < gFI.CountofClusters + 2; cluster++)
        {
            error = Read_FAT_Value(cluster, &fat);
            if (error) return 2; // FAT Read Error
            if (fat == UNUSED_CLUSTER)
            {
                found = 1;
                break;
            }
        }
        if (found == 0) return 1; // may be Volume Full
        //
        // Update FAT Table of File Body to make the link
        //
        error = Write_FAT_Value(lastCluster, (UI16) cluster);
        if (error) return 2; // FAT Write Error
        error = Write_FAT_Value(cluster, END_OF_CLUSTER);
        if (error) return 2; // FAT Write Error
    }    
    //
    // Update Directory Information (DIR_FileSize)
    //
    error = Write_Dir_Info(&gDatCache, dirSector,  dirEntry, &dirInfo);
    if (error) return 2; // Dir Write Error
    //
    // Success
    //
    return 0;
}

//------------------------------------------------
// Get N-th File Information from Target Directory
//------------------------------------------------
UI08 Get_File_Info_from_TargetDir(UI32 nth, sDirInfo *psDirInfo)
{
    UI08     error;
    sDirInfo dirInfo;
    UI32     cluster;
    UI32     lastCluster;
    UI16     fat;
    UI32     sector;
    UI32     entry;
    UI32     fileInfoCount;
    //
    // Search Target Directory from Root Directory
    //
    error =  Search_TargetDir_from_RootDir(&dirInfo);
    if (error) return error; // no Target Directory or something error
    //
    // Search all Clusters for the Target Directory along with FAT
    // until finding nth (0, 1, 2,...) file information
    //
    fileInfoCount = 0;
    cluster = dirInfo.DIR_FstClusLO; // first cluster
    //
    while(1)
    {
        lastCluster = cluster;
        //
        // search all sectors in the cluster
        //
        for (sector = First_Sector_of_Cluster(cluster); sector < (First_Sector_of_Cluster(cluster) + (UI32) gBS.BPB_SecPerClus); sector++)
        {
            //
            // search all entries in the sector
            //
            for (entry = 0; entry < gBS.BPB_BytsPerSec / 32; entry++)
            {
                error = Read_Dir_Info(&gDatCache, sector, entry, psDirInfo); // get directory information
                if (error) return 2;  // Directory read error
                if (psDirInfo->DIR_Name[0] == 0) return 1; // no more directory data, can not find required information
                if (psDirInfo->DIR_Name[0] == 0xe5) continue;                           // skip it 
                if (psDirInfo->DIR_Attr == ATTR_LONG_NAME) continue;                    // skip it
                if ((psDirInfo->DIR_Attr & ATTR_DIRECTORY) == ATTR_DIRECTORY) continue; // skip it
                //
                // found a file
                //
                if (nth == fileInfoCount) return 0; // here you are! please use psDirInfo.
                //
                // it is not your request, ok search next
                //
                fileInfoCount = fileInfoCount + 1;
            }
        }
        //
        // go to next cluster chain
        //
        error = Read_FAT_Value(lastCluster, &fat);
        if (error) return 2;  // FAT read error
        if (fat >= EOC) return 1; // no more cluster for Target Directory, can not find your request
        cluster = (UI32) fat;
    }
    //
    // Success
    //
    return 0;
}

//-------------------------------------------
// Make New File Name in the Target Directory
//-------------------------------------------
UI08 Make_New_File_Name_in_TargetDir(UI08 *pFileName_11)
{
    UI08     error;
    sDirInfo dirInfo;
    UI32     cluster;
    UI32     sector;
    UI32     entry;
    UI16     fat;
    UI32     maxFileNameNum;
    UI32     fileNameNum;
    UI08     finish;
    UI08     i;
    UI08     outRange;
    //
    // Search Target Directory from Root Directory
    // if not found, Create Target Directory
    //
    error =  Search_TargetDir_from_RootDir(&dirInfo);
    if (error == 1)
    {
         error = Create_TargetDir_in_RootDir(&dirInfo);
         if (error) return 2; // can not create target directory
    }
    else if (error) return 2; // Directory access error
    //
    // Search all Clusters for the Target Directory along with FAT
    //
    maxFileNameNum = 0; // number to be converted to file name 
    finish = 0;
    cluster = dirInfo.DIR_FstClusLO; // first cluster
    while(1)
    {
        //
        // search all sectors in the cluster
        //
        for (sector = First_Sector_of_Cluster(cluster); sector < (First_Sector_of_Cluster(cluster) + (UI32) gBS.BPB_SecPerClus); sector++)
        {
            //
            // search all entries in the sector
            //
            for (entry = 0; entry < gBS.BPB_BytsPerSec / 32; entry++)
            {
                error = Read_Dir_Info(&gDatCache, sector, entry, &dirInfo); // get directory information
                if (error) return 2;  // Directory read error
                if (dirInfo.DIR_Name[0] == 0) {finish = 1; break;} // no more directory data
                if (dirInfo.DIR_Name[0] == 0xe5) continue; // skip it 
                if (dirInfo.DIR_Attr == ATTR_LONG_NAME) continue; // skip it
                if ((dirInfo.DIR_Attr & ATTR_DIRECTORY) == ATTR_DIRECTORY) continue; // skip it
                //
                // Check File Name ********.AWV (* must be 0~9)
                //
                outRange = 0;
                for (i = 0; i < 8; i++)
                {
                    if (dirInfo.DIR_Name[i] <  '0') {outRange = 1; break;}
                    if (dirInfo.DIR_Name[i] >  '9') {outRange = 1; break;}
                }
                if (outRange) continue; 
                if (dirInfo.DIR_Name[ 8] != 'A') continue;
                if (dirInfo.DIR_Name[ 9] != 'W') continue;
                if (dirInfo.DIR_Name[10] != 'V') continue;
                //
                // now, found candidate, so check max name
                //
                fileNameNum = 0;
                for (i = 0; i < 8; i++)
                {
                    fileNameNum = fileNameNum << 4;
                    fileNameNum = fileNameNum + (dirInfo.DIR_Name[i] - '0');
                }
                if (maxFileNameNum < fileNameNum) maxFileNameNum = fileNameNum;
            }
            if (finish) break;
        }
        if (finish) break;
        //
        // final cluster?
        //
        error = Read_FAT_Value(cluster, &fat);
        if (error) return 2;  // FAT read error
        if (fat >= EOC) break;
        //
        // next cluster
        //
        cluster = (UI32) fat; // next cluster
    }
    //
    // Make New File Name
    //
    maxFileNameNum = maxFileNameNum + 1;
    //
    for (i = 0; i < 8; i++) // convert to BCD code
    {
        if ((maxFileNameNum & (0x0000000f<<(i*4))) > (0x00000009<<(i*4))) maxFileNameNum = maxFileNameNum + (0x00000006<<(i*4));
        //
        pFileName_11[7-i] = ((maxFileNameNum & (0x0000000f<<(i*4))) >> (i*4)) + '0';
    }
    pFileName_11[ 8] = 'A';
    pFileName_11[ 9] = 'W';
    pFileName_11[10] = 'V';
    //
    // Success
    //
    return 0;
}

//--------------------------------------------
// Search Target Directory from Root Directory
//--------------------------------------------
UI08 Search_TargetDir_from_RootDir(sDirInfo *psDirInfo)
{
    UI32 sector; // sector of root directory
    UI32 entry;  // directory entry (16 entries per sector)
    UI32 entrybase;
    UI32 offset;
    UI08 error;
    UI08 data;
    UI08 i;
    UI08 found;
    UI08 nameHit;
    //
    // search all sectors for root directory
    //
    nameHit = 0;
    for (sector = gFI.FirstRootDirSecNum; sector < (gFI.FirstRootDirSecNum + gFI.RootDirSectors); sector++)
    {
        //
        // search all directory entries in the sector
        //
        for (entry = 0; entry < gBS.BPB_BytsPerSec / 32; entry++)
        {
            entrybase = entry * 32;
            //
            // check DIR_Name[0]
            //
            offset = 0;
            error = Read_Data_from_Cache(&gDatCache, sector, entrybase + offset, &data);
            if (error) return 2;        // Cache Read Error
            if (data == 0) return 1;    // No more directory entres, Not Found the directory
            if (data == 0xe5) continue; // search next entry
            //
            // check DIR_Attr
            //            
            offset = 11;
            error = Read_Data_from_Cache(&gDatCache, sector, entrybase + offset, &data);
            if (error) return 2;                  // Cache Read Error
            if (data != ATTR_DIRECTORY) continue; // search next entry
            //
            // compare DIR_Name with the Target Directory Name
            //
            found = 1;
            for (i = 0; i < 11; i++)
            {
                offset = i;
                error = Read_Data_from_Cache(&gDatCache, sector, entrybase + offset, &data);
                if (error) return 2;     // Cache Read Error
                if (data != gpStrTargetDir[i]) {found = 0; break;}
            }
            if (found) // found the name
            {
                nameHit = 1;
                error = Read_Dir_Info(&gDatCache, sector, entry, psDirInfo); // get directory information
                if (error) return 2;     // Cache Read Error
                break; // found 
            }
        }
        if (nameHit) break; // found the name
    }
    if (nameHit == 0) return 1; // not found
    //
    // Success
    //
    return 0; // Found the directory
}

//------------------------------------------
// Create Target Directory in Root Directory
//------------------------------------------
UI08 Create_TargetDir_in_RootDir(sDirInfo *psDirInfo)
{
    UI32 cluster;
    UI32 sector;
    UI32 offset;
    UI16 fat;
    UI08 error;
    UI08 found;
    sDirInfo dirInfoDot; // dot and dotdor directory
    UI08 i;
    UI32 rootSector;
    UI32 rootEntry;
    UI08 data;
    //
    // Search vacant Cluster in FAT Table
    //
    found = 0;
    for (cluster = 2; cluster < gFI.CountofClusters + 2; cluster++)
    {
        error = Read_FAT_Value(cluster, &fat);
        if (error) return 2; // FAT Read Error
        if (fat == UNUSED_CLUSTER) {found = 1; break;}
    }
    if (found == 0) return 1; // may be Volume Full
    //
    // Clear all sector data in the cluster of Target Directory
    //
    for (sector = First_Sector_of_Cluster(cluster); sector < (First_Sector_of_Cluster(cluster) + (UI32) gBS.BPB_SecPerClus); sector++)
    {
        for (offset = 0; offset < gBS.BPB_BytsPerSec; offset++)
        {
            error = Write_Data_to_Cache(&gDatCache, sector, offset, 0); // zero clear
            if (error) return 2; // Dir Write Error
        }
    }
    //
    // Create "dot" directory
    //
    dirInfoDot.DIR_Name[0] = '.'; // make "."
    for (i = 1; i < 11; i++) dirInfoDot.DIR_Name[i] = ' ';
    dirInfoDot.DIR_Attr = ATTR_DIRECTORY;
    dirInfoDot.DIR_FileSize = 0;
    dirInfoDot.DIR_FstClusHI = 0;
    dirInfoDot.DIR_FstClusLO = cluster & 0x0000ffff;
    Fill_Date_Time_in_DirInfo(&dirInfoDot);
    //
    error = Write_Dir_Info(&gDatCache, First_Sector_of_Cluster(cluster),  0, &dirInfoDot);
    if (error) return 2; // Dir Write Error
    //
    // Create "dot-dot" directory
    //
    dirInfoDot.DIR_Name[1] = '.'; // make ".."
    dirInfoDot.DIR_FstClusLO = 0; // because parent directorty is root directory
    //
    error = Write_Dir_Info(&gDatCache, First_Sector_of_Cluster(cluster), 1, &dirInfoDot);
    if (error) return 2; // Dir Write Error
    //
    // Write EOC mark on the FAT (so far, occupy 1 cluster for new Target Directory)
    //
    error = Write_FAT_Value(cluster, END_OF_CLUSTER);
    if (error) return 2; // FAT Write Error    
    //
    // Write Root Directory
    // Search vacant entry in Root Directory
    //
    found = 0;
    for (rootSector = gFI.FirstRootDirSecNum; rootSector < (gFI.FirstRootDirSecNum + gFI.RootDirSectors); rootSector++)
    {
        //
        // search all directory entries in the sector
        //
        for (rootEntry = 0; rootEntry < gBS.BPB_BytsPerSec / 32; rootEntry++)
        {
            error = Read_Data_from_Cache(&gDatCache, rootSector, rootEntry * 32, &data); // search DIR_Name[0]
            if (error) return 2; // FAT Write Error
            if ((data == 0) | (data == 0xe5)) {found = 1; break;}
        }
        if (found) break;
    }
    if (found == 0) return 1; // may be Volume Full
    //
    // Make Root Directory Information
    //
    for (i = 0; i < 11; i++) psDirInfo->DIR_Name[i] = gpStrTargetDir[i];
    psDirInfo->DIR_Attr = ATTR_DIRECTORY;
    psDirInfo->DIR_FileSize = 0;
    psDirInfo->DIR_FstClusHI = 0;
    psDirInfo->DIR_FstClusLO = cluster & 0x0000ffff;
    Fill_Date_Time_in_DirInfo(psDirInfo);
    //
    error = Write_Dir_Info(&gDatCache, rootSector, rootEntry, psDirInfo);
    if (error) return 2; // Dir Write Error    
    //
    // Success
    //
    return 0;
}

//----------------------------------------------------
// Fill current Date and Time in Directory Information
//----------------------------------------------------
void Fill_Date_Time_in_DirInfo(sDirInfo *psDirInfo)
{
    UI16 year;
    UI16 month;
    UI16 day;
    UI16 hour;
    UI16 minute;
    UI16 second;
    //
    // convert from BCD code to Binary
    // the range of year is 2000-2099 in this system, 
    // so 20 is added to the year (because of counting from 1980)
    //
    year   = (gDateTime.bcd_years    & 0x000f) + (((gDateTime.bcd_years    & 0x00f0) >> 4) * 10) + 20;
    month  = (gDateTime.bcd_months   & 0x000f) + (((gDateTime.bcd_months   & 0x00f0) >> 4) * 10);
    day    = (gDateTime.bcd_days     & 0x000f) + (((gDateTime.bcd_days     & 0x00f0) >> 4) * 10);
    hour   = (gDateTime.bcd_hours    & 0x000f) + (((gDateTime.bcd_hours    & 0x00f0) >> 4) * 10);
    minute = (gDateTime.bcd_minutes  & 0x000f) + (((gDateTime.bcd_minutes  & 0x00f0) >> 4) * 10);
    second = (gDateTime.bcd_seconds  & 0x000f) + (((gDateTime.bcd_seconds  & 0x00f0) >> 4) * 10);
    //
    // fill in directory information
    // note that granularity of second is 2-seconds
    //
    psDirInfo->DIR_WrtDate = ((year & 0x007f)<< 9) + ((month  & 0x000f)<<5) + ((day    & 0x001f));
    psDirInfo->DIR_WrtTime = ((hour & 0x001f)<<11) + ((minute & 0x003f)<<5) + ((second & 0x003f)>>1);
    psDirInfo->DIR_CrtTimeTenth = ((second % 2) == 0)? 0 : 100; // if second is even then 0, else 100(ms).
    //
    // copy them, because writing a file always means creating it, in this system, 
    //
    psDirInfo->DIR_CrtDate    = psDirInfo->DIR_WrtDate;
    psDirInfo->DIR_CrtTime    = psDirInfo->DIR_WrtTime;
    psDirInfo->DIR_LstAccDate = psDirInfo->DIR_WrtDate;
}

//---------------------------
// Read Directory Information
//---------------------------
UI08 Read_Dir_Info(sSectorCache *cache, UI32 sector, UI32 entry, sDirInfo *psDirInfo)
{
    UI08 error;
    UI08 i;
    UI08 dataHH;
    UI08 dataHL;
    UI08 dataLH;
    UI08 dataLL;
    UI32 offset;
    //
    // Entry size of Directory is limited in 16bit size
    //
    if (entry > 65536) return 1; // entry over
    //
    // Read Directory Information
    //
    offset = entry * 32;
    //
    for (i = 0; i < 11; i++)
    {
        error = Read_Data_from_Cache(cache, sector, offset + i, &(psDirInfo->DIR_Name[i]));
            if (error) return 2;     // Cache Read Error
    }
    //
    error = Read_Data_from_Cache(cache, sector, offset + 11, &(psDirInfo->DIR_Attr));
        if (error) return 2;     // Cache Read Error
  //error = Read_Data_from_Cache(cache, sector, offset + 12, &(psDirInfo->DIR_NTRes));
  //    if (error) return 2;     // Cache Read Error
    error = Read_Data_from_Cache(cache, sector, offset + 13, &(psDirInfo->DIR_CrtTimeTenth));
        if (error) return 2;     // Cache Read Error
    //
    error = Read_Data_from_Cache(cache, sector, offset + 14, &dataLL);
        if (error) return 2;     // Cache Read Error
    error = Read_Data_from_Cache(cache, sector, offset + 15, &dataLH);
        if (error) return 2;     // Cache Read Error
    psDirInfo->DIR_CrtTime = dataLL + (dataLH<<8);    
    //
    error = Read_Data_from_Cache(cache, sector, offset + 16, &dataLL);
        if (error) return 2;     // Cache Read Error
    error = Read_Data_from_Cache(cache, sector, offset + 17, &dataLH);
        if (error) return 2;     // Cache Read Error
    psDirInfo->DIR_CrtDate = dataLL + (dataLH<<8);    
    //
    error = Read_Data_from_Cache(cache, sector, offset + 18, &dataLL);
        if (error) return 2;     // Cache Read Error
    error = Read_Data_from_Cache(cache, sector, offset + 19, &dataLH);
        if (error) return 2;     // Cache Read Error
    psDirInfo->DIR_LstAccDate = dataLL + (dataLH<<8);    
    //
    error = Read_Data_from_Cache(cache, sector, offset + 20, &dataLL);
        if (error) return 2;     // Cache Read Error
    error = Read_Data_from_Cache(cache, sector, offset + 21, &dataLH);
        if (error) return 2;     // Cache Read Error
    psDirInfo->DIR_FstClusHI = dataLL + (dataLH<<8);    
    //
    error = Read_Data_from_Cache(cache, sector, offset + 22, &dataLL);
        if (error) return 2;     // Cache Read Error
    error = Read_Data_from_Cache(cache, sector, offset + 23, &dataLH);
        if (error) return 2;     // Cache Read Error
    psDirInfo->DIR_WrtTime = dataLL + (dataLH<<8);    
    //
    error = Read_Data_from_Cache(cache, sector, offset + 24, &dataLL);
        if (error) return 2;     // Cache Read Error
    error = Read_Data_from_Cache(cache, sector, offset + 25, &dataLH);
        if (error) return 2;     // Cache Read Error
    psDirInfo->DIR_WrtDate = dataLL + (dataLH<<8);    
    //
    error = Read_Data_from_Cache(cache, sector, offset + 26, &dataLL);
        if (error) return 2;     // Cache Read Error
    error = Read_Data_from_Cache(cache, sector, offset + 27, &dataLH);
        if (error) return 2;     // Cache Read Error
    psDirInfo->DIR_FstClusLO = dataLL + (dataLH<<8);    
    //
    error = Read_Data_from_Cache(cache, sector, offset + 28, &dataLL);
        if (error) return 2;     // Cache Read Error
    error = Read_Data_from_Cache(cache, sector, offset + 29, &dataLH);
        if (error) return 2;     // Cache Read Error
    error = Read_Data_from_Cache(cache, sector, offset + 30, &dataHL);
        if (error) return 2;     // Cache Read Error
    error = Read_Data_from_Cache(cache, sector, offset + 31, &dataHH);
        if (error) return 2;     // Cache Read Error
    psDirInfo->DIR_FileSize= dataLL + (dataLH<<8) + (dataHL<<16) + (dataHH<<24);    
    //
    // Success
    //
    return 0;
}

//----------------------------
// Write Directory Information
//----------------------------
UI08 Write_Dir_Info(sSectorCache *cache, UI32 sector, UI32 entry, sDirInfo *psDirInfo)
{
    UI08 error;
    UI08 i;
    UI08 dataHH;
    UI08 dataHL;
    UI08 dataLH;
    UI08 dataLL;
    UI32 offset;
    //
    // Entry size of Directory is limited in 16bit size
    //
    if (entry > 65536) return 1; // entry over
    //
    // Write Directory Information
    //
    offset = entry * 32;
    //
    for (i = 0; i < 11; i++)
    {
        error = Write_Data_to_Cache(cache, sector, offset + i, psDirInfo->DIR_Name[i]);
            if (error) return 2;     // Cache Write Error
    }
    //
    error = Write_Data_to_Cache(cache, sector, offset + 11, psDirInfo->DIR_Attr);
        if (error) return 2;     // Cache Write Error
    error = Write_Data_to_Cache(cache, sector, offset + 12, 0); //DIR_NTRes
        if (error) return 2;     // Cache Write Error
    error = Write_Data_to_Cache(cache, sector, offset + 13, psDirInfo->DIR_CrtTimeTenth);
        if (error) return 2;     // Cache Write Error
    //
    dataLL =  psDirInfo->DIR_CrtTime & 0x00ff;
    dataLH = (psDirInfo->DIR_CrtTime & 0xff00) >> 8;
    error = Write_Data_to_Cache(cache, sector, offset + 14, dataLL);
        if (error) return 2;     // Cache Write Error
    error = Write_Data_to_Cache(cache, sector, offset + 15, dataLH);
        if (error) return 2;     // Cache Write Error
    //
    dataLL =  psDirInfo->DIR_CrtDate & 0x00ff;
    dataLH = (psDirInfo->DIR_CrtDate & 0xff00) >> 8;
    error = Write_Data_to_Cache(cache, sector, offset + 16, dataLL);
        if (error) return 2;     // Cache Write Error
    error = Write_Data_to_Cache(cache, sector, offset + 17, dataLH);
        if (error) return 2;     // Cache Write Error
    //
    dataLL =  psDirInfo->DIR_LstAccDate & 0x00ff;
    dataLH = (psDirInfo->DIR_LstAccDate & 0xff00) >> 8;
    error = Write_Data_to_Cache(cache, sector, offset + 18, dataLL);
        if (error) return 2;     // Cache Write Error
    error = Write_Data_to_Cache(cache, sector, offset + 19, dataLH);
        if (error) return 2;     // Cache Write Error
    //
    dataLL =  psDirInfo->DIR_FstClusHI & 0x00ff;
    dataLH = (psDirInfo->DIR_FstClusHI & 0xff00) >> 8;
    error = Write_Data_to_Cache(cache, sector, offset + 20, dataLL);
        if (error) return 2;     // Cache Write Error
    error = Write_Data_to_Cache(cache, sector, offset + 21, dataLH);
        if (error) return 2;     // Cache Write Error
    //
    dataLL =  psDirInfo->DIR_WrtTime & 0x00ff;
    dataLH = (psDirInfo->DIR_WrtTime & 0xff00) >> 8;
    error = Write_Data_to_Cache(cache, sector, offset + 22, dataLL);
        if (error) return 2;     // Cache Write Error
    error = Write_Data_to_Cache(cache, sector, offset + 23, dataLH);
        if (error) return 2;     // Cache Write Error
    //
    dataLL =  psDirInfo->DIR_WrtDate & 0x00ff;
    dataLH = (psDirInfo->DIR_WrtDate & 0xff00) >> 8;
    error = Write_Data_to_Cache(cache, sector, offset + 24, dataLL);
        if (error) return 2;     // Cache Write Error
    error = Write_Data_to_Cache(cache, sector, offset + 25, dataLH);
        if (error) return 2;     // Cache Write Error
    //
    dataLL =  psDirInfo->DIR_FstClusLO & 0x00ff;
    dataLH = (psDirInfo->DIR_FstClusLO & 0xff00) >> 8;
    error = Write_Data_to_Cache(cache, sector, offset + 26, dataLL);
        if (error) return 2;     // Cache Write Error
    error = Write_Data_to_Cache(cache, sector, offset + 27, dataLH);
        if (error) return 2;     // Cache Write Error
    //
    dataLL =  psDirInfo->DIR_FileSize & 0x000000ff;
    dataLH = (psDirInfo->DIR_FileSize & 0x0000ff00) >>  8;
    dataHL = (psDirInfo->DIR_FileSize & 0x00ff0000) >> 16;
    dataHH = (psDirInfo->DIR_FileSize & 0xff000000) >> 24;
    error = Write_Data_to_Cache(cache, sector, offset + 28, dataLL);
        if (error) return 2;     // Cache Write Error
    error = Write_Data_to_Cache(cache, sector, offset + 29, dataLH);
        if (error) return 2;     // Cache Write Error
    error = Write_Data_to_Cache(cache, sector, offset + 30, dataHL);
        if (error) return 2;     // Cache Write Error
    error = Write_Data_to_Cache(cache, sector, offset + 31, dataHH);
        if (error) return 2;     // Cache Write Error
    //
    // Success
    //
    return 0;
}

//---------------
// Read FAT Value
//---------------
UI08 Read_FAT_Value(UI32 cluster, UI16 *FAT16ClusEntryVal)
{
    UI32 FATOffset;
    UI32 ThisFATSecNum;
    UI32 ThisFATEntOffset;
    UI08 FATL;
    UI08 FATH;
    UI08 error;
    //
    // Prepare Sector and Offset
    //
    FATOffset = cluster * 2; // because of FAT16
    ThisFATSecNum    = gBS.BPB_RsvdSecCnt + (FATOffset / gBS.BPB_BytsPerSec);
    ThisFATEntOffset = FATOffset % gBS.BPB_BytsPerSec;
    //
    // Read from FAT Cache
    //
    error = Read_Data_from_Cache(&gFatCache, ThisFATSecNum, ThisFATEntOffset    , &FATL);
    if (error) return error;
    error = Read_Data_from_Cache(&gFatCache, ThisFATSecNum, ThisFATEntOffset + 1, &FATH);
    if (error) return error;
    //
    // Return the Value
    //
    *FAT16ClusEntryVal = FATL + (FATH<<8);
    //
    // Success
    //
    return 0;
}

//----------------
// Write FAT Value
//----------------
UI08 Write_FAT_Value(UI32 cluster, UI16 FAT16ClusEntryVal)
{
    UI32 FATOffset;
    UI32 ThisFATSecNum;
    UI32 ThisFATEntOffset;
    UI08 FATL;
    UI08 FATH;
    UI08 error;
    //
    // Prepare Sector and Offset
    //
    FATOffset = cluster * 2; // because of FAT16
    ThisFATSecNum    = gBS.BPB_RsvdSecCnt + (FATOffset / gBS.BPB_BytsPerSec);
    ThisFATEntOffset = FATOffset % gBS.BPB_BytsPerSec;
    //
    // Prepare each byte of FAT value
    //
    FATL =  FAT16ClusEntryVal & 0x00ff;
    FATH = (FAT16ClusEntryVal & 0xff00) >> 8;
    //
    // Write to FAT Cache
    //
    error = Write_Data_to_Cache(&gFatCache, ThisFATSecNum, ThisFATEntOffset    , FATL);
    if (error) return error;
    error = Write_Data_to_Cache(&gFatCache, ThisFATSecNum, ThisFATEntOffset + 1, FATH);
    if (error) return error;
    //
    // Success
    //
    return 0;
}

//--------------------------------
// Compute First Sector of Cluster
//--------------------------------
UI32 First_Sector_of_Cluster(UI32 cluster)
{
    UI32 FirstSectorofCluster;
    
    FirstSectorofCluster = ((cluster - 2) * gBS.BPB_SecPerClus) + gFI.FirstDataSector;
    return FirstSectorofCluster;
}

//-------------------------
// Check FAT Data Structure
//-------------------------
UI08 Check_FAT_Data_Structure(void)
{
    //
    // Bytes per sector must be 512 in this application
    //
    if (gBS.BPB_BytsPerSec != 512) return 255;
    //
    // Count of Sectors occupied by the Root Directory
    //
    gFI.RootDirSectors = ((gBS.BPB_RootEntCnt * 32) + (gBS.BPB_BytsPerSec - 1)) / gBS.BPB_BytsPerSec;
    //
    // Start of the data region, the first sector of cluster 2
    //
    gFI.FATSz = gBS.BPB_FATSz16;
    if (gFI.FATSz ==0) return 32; // can not accept "maybe-FAT32" format in this program
    gFI.FirstDataSector = gBS.BPB_RsvdSecCnt + (gBS.BPB_NumFATs * gFI.FATSz) + gFI.RootDirSectors;
    //
    // Count of Sectors in the data region of the volume
    //
    if (gBS.BPB_TotSec16 != 0)
    {
        gFI.TotSec = gBS.BPB_TotSec16;
    }
    else
    {
        gFI.TotSec = gBS.BPB_TotSec32;
    }
    gFI.DataSec = gFI.TotSec - (gBS.BPB_RsvdSecCnt + (gBS.BPB_NumFATs * gFI.FATSz) + gFI.RootDirSectors);
    //
    // Count of Clusters
    //
    gFI.CountofClusters = gFI.DataSec / gBS.BPB_SecPerClus;
    //
    // Now determine the FAT type : THIS IS THE ONE AND ONLY WAY TO DETERMINE THE FAT TYPE.
    //
    if (gFI.CountofClusters <   4085) return 12; // can not accept FAT12
    if (gFI.CountofClusters >= 65525) return 32; // can not accept FAT32
    //
    // First Sector of Root Directory
    //    
    gFI.FirstRootDirSecNum = gBS.BPB_RsvdSecCnt + (gBS.BPB_NumFATs * gFI.FATSz);
    //
    // Success
    //
    return 0;
}

//-------------
// Mount Volume
//-------------
UI08 SD_Mount(void)
{
    return Mount_Volume(DO_MOUNT);
}
UI08 SD_UnMount(void)
{
    return Mount_Volume(DO_UNMOUNT);
}
//
UI08 Mount_Volume(UI08 mount)
{
    static UI08 mounted = 0; // if this volume is mounted or not
    UI08 error;
    //
    // Do Mount
    //
    if (mount == DO_MOUNT)
    {
        if (SD_Check_Ready())
        {
            if (mounted == 0) // not mounted
            {
                //
                // Initialize Sector Cache
                //
                gFatCache.valid = 0;
                gDatCache.valid = 0;
                //
                // Read Volume Information
                //
                error = Read_Volume_Info();
                if (error) return (error);
                //
                error = Check_FAT_Data_Structure();
                if (error) return (error);
                //
              //Print_Volume_Info();
                //
                mounted = 1;
            }
            else // already mounted
            {
                // do nothing
            }
        }
        else
        {
            //
            // Invalidate Sector Cache
            //
            gFatCache.valid = 0;
            gDatCache.valid = 0;
            //
            mounted = 0;
            return 2; // can not mount
        }
    }
    //
    // Do UnMount
    //
    else // if (mount == DO_UNMOUNT)
    {
        if (mounted)
        {
            if (SD_Check_Ready())
            {
                error = Purge_Cache(&gFatCache);
                if (error) return (error);
                //
                error = Purge_Cache(&gDatCache);
                if (error) return (error);
                //
                mounted = 0;
            }
            else
            {
                mounted = 0;
                return 2; // SD not ready before unmount
            }
        }
        else // not mounted
        {
            // do nothing
        }
    }
    //
    // Success
    //
    return 0;
}

//-------------------
// Purge Sector Cache
//-------------------
UI08 Purge_Cache(sSectorCache *cache)
{
    UI32 addr;
    UI08 response;
    //
    // if Cache is Valid and Dirty, need write back to volume
    //
    if ((cache->valid) & (cache->dirty))
    {
        addr = (cache->sector) << SD_BLOCK_SIZE;
        response = SD_Write_Block(addr, cache->buffer);
        if (response) return response; // keep the cache
    }
    //
    // Invalidate Cache
    //
    cache->valid = 0;
    //
    // Success
    //
    return 0;
}

//----------------------------
// Read Data from Sector Cache
//----------------------------
UI08 Read_Data_from_Cache(sSectorCache *cache, UI32 sector, UI32 offset, UI08 *data)
{
    UI32 addr;
    UI08 response;
    //
    // Convert Sector Number for this volume
    //
    sector = sector + gPT.firstSectorPosition;
    //
    // if Cache is not Valid, read one sector to the buffer
    //
    if (cache->valid == 0)
    {
        addr = sector << SD_BLOCK_SIZE;
        response = SD_Read_Block(addr, cache->buffer);
        if (response) return response;
        cache->sector = sector;
        cache->valid  = 1;
        cache->dirty  = 0;
        *data = cache->buffer[offset];
    }
    //
    // if Cache Valid but Miss
    //
    else if (cache->sector != sector)
    {
        // if dirty, Write back Current Cache Contents to the Volume
        if (cache->dirty)
        {
            addr = (cache->sector) << SD_BLOCK_SIZE;
            response = SD_Write_Block(addr, cache->buffer);
            if (response) return response; // keep the cache
        }
        // Cache Fill
        addr = sector << SD_BLOCK_SIZE;
        response = SD_Read_Block(addr, cache->buffer);
        if (response) return response;
        cache->sector = sector;
        cache->valid  = 1;
        cache->dirty  = 0;
        *data = cache->buffer[offset];
    }
    //
    // if Cache Valid and Hit
    //
    else
    {
        *data = cache->buffer[offset];
    }
    //
    // Success
    //
    return 0;
}

//---------------------------
// Write Data to Sector Cache
//---------------------------
UI08 Write_Data_to_Cache(sSectorCache *cache, UI32 sector, UI32 offset, UI08 data)
{
    UI32 addr;
    UI08 response;
    //
    // Convert Sector Number for this volume
    //
    sector = sector + gPT.firstSectorPosition;
    //
    // if Cache is not Valid, read one sector to the buffer
    //
    if (cache->valid == 0)
    {
        addr = sector << SD_BLOCK_SIZE;
        response = SD_Read_Block(addr, cache->buffer);
        if (response) return response;
        cache->sector = sector;
        cache->valid  = 1;
        cache->dirty  = 1;
        cache->buffer[offset] = data;
    }
    //
    // if Cache Valid but Miss
    //
    else if (cache->sector != sector)
    {
        // if dirty, Write back Current Cache Contents to the Volume
        if (cache->dirty)
        {
            addr = (cache->sector) << SD_BLOCK_SIZE;
            response = SD_Write_Block(addr, cache->buffer);
            if (response) return response; // keep the cache
        }
        // Store data to New Cache
        addr = sector << SD_BLOCK_SIZE;
        response = SD_Read_Block(addr, cache->buffer);
        if (response) return response;
        cache->sector = sector;
        cache->valid  = 1;
        cache->dirty  = 1;
        cache->buffer[offset] = data;
    }
    //
    // if Cache Valid and Hit
    //
    else
    {
        cache->dirty  = 1;
        cache->buffer[offset] = data;
    }
    //
    // Success
    //
    return 0;
}

//------------------------
// Read Volume Information
//------------------------
UI08 Read_Volume_Info(void)
{
    UI08 error;
    UI16 offset;
    UI08 buf[512];
    //
    // Read First Block in SD Card
    //
    error = SD_Read_Block(0, buf);
    if (error) return error;
    //
    // Store Partition Table Information
    //
    offset = 0x1be; // partition 0
    gPT.bootDescriptor       = buf[offset+ 0];
    gPT.firstPartitionSector = buf[offset+ 1] + (buf[offset+ 2]<<8) + (buf[offset+ 3]<<16);
    gPT.fileSystemDescriptor = buf[offset+ 4];
    gPT.lastPartitionSector  = buf[offset+ 5] + (buf[offset+ 6]<<8) + (buf[offset+ 7]<<16);
    gPT.firstSectorPosition  = buf[offset+ 8] + (buf[offset+ 9]<<8) + (buf[offset+10]<<16) + (buf[offset+11]<<24);
    gPT.numberOfSectors      = buf[offset+12] + (buf[offset+13]<<8) + (buf[offset+14]<<16) + (buf[offset+15]<<24);
    //
    // Check if there is only one partition
    //
    for (offset = 0x1ce; offset < 0x1de; offset++) // partition 1
    {
        if (buf[offset] != 0) return offset;
    }
    for (offset = 0x1de; offset < 0x1ee; offset++) // partition 2
    {
        if (buf[offset] != 0) return offset;
    }
    for (offset = 0x1ee; offset < 0x1fe; offset++) // partition 3
    {
        if (buf[offset] != 0) return offset;
    }
    //
    // Check Partition Table Signature
    //
    if (buf[0x1fe] != 0x55) return 0x55;
    if (buf[0x1ff] != 0xaa) return 0xaa;
    //
    // Read Boot Sector (of partition 0)
    //
    error = SD_Read_Block(gPT.firstSectorPosition << SD_BLOCK_SIZE, buf);
    if (error) return error;
    //
    // Store Boot Sector Information
    //
    gBS.BPB_BytsPerSec = buf[11] + (buf[12]<<8);
    gBS.BPB_SecPerClus = buf[13];
    gBS.BPB_RsvdSecCnt = buf[14] + (buf[15]<<8);
    gBS.BPB_NumFATs    = buf[16];
    gBS.BPB_RootEntCnt = buf[17] + (buf[18]<<8);
    gBS.BPB_TotSec16   = buf[19] + (buf[20]<<8);
    gBS.BPB_Media      = buf[21];
    gBS.BPB_FATSz16    = buf[22] + (buf[23]<<8);
    gBS.BPB_HiddSec    = buf[28] + (buf[29]<<8) + (buf[30]<<16) + (buf[31]<<24);
    gBS.BPB_TotSec32   = buf[32] + (buf[33]<<8) + (buf[34]<<16) + (buf[35]<<24);
    //
    // Check Boot Sector Signature
    //
    if (buf[0x1fe] != 0x55) return 0x55;
    if (buf[0x1ff] != 0xaa) return 0xaa;
    //
    // Success
    //
    return 0;
}

//-------------------------
// Print Volume Information
//-------------------------
void Print_Volume_Info(void)
{
    UI08 i;
    //
    // Print Partition Table Information
    //
    i = 0;
    printf("<< Partition Table >>\n");
    printf("bootDescriptor       = 0x%02x\n", (int) gPT.bootDescriptor);
    printf("firstPartitionSector = 0x%08x\n", (int) gPT.firstPartitionSector);
    printf("fileSystemDescriptor = 0x%02x\n", (int) gPT.fileSystemDescriptor);
    printf("lastPartitionSector  = 0x%08x\n", (int) gPT.lastPartitionSector);
    printf("firstSectorPosition  = 0x%08x\n", (int) gPT.firstSectorPosition);
    printf("numberOfSectors      = 0x%08x\n", (int) gPT.numberOfSectors);
    //
    // Print Boot Sector Information
    //
    printf("<< Boot Sector >>\n");
    printf("BPB_BytsPerSec = %d\n",   (int) gBS.BPB_BytsPerSec);
    printf("BPB_SecPerClus = %d\n",   (int) gBS.BPB_SecPerClus);
    printf("BPB_RsvdSecCnt = %d\n",   (int) gBS.BPB_RsvdSecCnt);
    printf("BPB_NumFATs    = %d\n",   (int) gBS.BPB_NumFATs);
    printf("BPB_RootEntCnt = %d\n",   (int) gBS.BPB_RootEntCnt);
    printf("BPB_TotSec16   = %d\n",   (int) gBS.BPB_TotSec16);
    printf("BPB_Media      = 0x%x\n", (int) gBS.BPB_Media);
    printf("BPB_FATSz16    = %d\n",   (int) gBS.BPB_FATSz16);
    printf("BPB_HiddSec    = %d\n",   (int) gBS.BPB_HiddSec);
    printf("BPB_TotSec32   = %d\n",   (int) gBS.BPB_TotSec32);
    //
    // Print FAT Structure Information
    //
    printf("<< FAT16 Information >>\n");
    printf("RootDirSectors     = %d\n", (int) gFI.RootDirSectors);
    printf("FATSz              = %d\n", (int) gFI.FATSz);
    printf("FirstDataSector    = %d\n", (int) gFI.FirstDataSector);
    printf("TotSec             = %d\n", (int) gFI.TotSec);
    printf("DataSec            = %d\n", (int) gFI.DataSec);
    printf("CountofClusters    = %d\n", (int) gFI.CountofClusters);
    printf("FirstRootDirSecNum = %d\n", (int) gFI.FirstRootDirSecNum);
    //
    // Dump Some sectors
    //
  //Dump_Sector(0); // Boot Sector
  //Dump_Sector(gFI.FirstRootDirSecNum); // Root Directory Sector
}

//----------------------------
// Print Directory Information
//----------------------------
void Print_Dir_Info(sDirInfo *psDirInfo)
{
    UI08 i;

    printf("<< Directory Information >>\n");
    printf("DIR_Name = \n");
    for (i = 0; i < 11; i++) printf("%c", (int) psDirInfo->DIR_Name[i]);
    printf(" \n");
    printf("DIR_Attr         = 0x%x\n", (int) psDirInfo->DIR_Attr);
    printf("DIR_CrtTimeTenth = 0x%x\n", (int) psDirInfo->DIR_CrtTimeTenth);
    printf("DIR_CrtTime      = 0x%x\n", (int) psDirInfo->DIR_CrtTime);
    printf("DIR_CrtDate      = 0x%x\n", (int) psDirInfo->DIR_CrtDate);
    printf("DIR_LstAccDate   = 0x%x\n", (int) psDirInfo->DIR_LstAccDate);
    printf("DIR_FstClusHI    = %d\n",   (int) psDirInfo->DIR_FstClusHI);
    printf("DIR_WrtTime      = 0x%x\n", (int) psDirInfo->DIR_WrtTime);
    printf("DIR_WrtDate      = 0x%x\n", (int) psDirInfo->DIR_WrtDate);
    printf("DIR_FstClusLO    = %d\n",   (int) psDirInfo->DIR_FstClusLO);
    printf("DIR_FileSize     = %d\n",   (int) psDirInfo->DIR_FileSize);
}

//----------------------------
// Dump Sector (Physical Dump)
//----------------------------
void Dump_Sector(UI32 sector)
{
    UI32 addr;
    UI08 response;
    UI08 x, y;
    UI08 buf[512];

    sector = sector + gPT.firstSectorPosition; // convert sector number for this volume
    //
    addr = sector << SD_BLOCK_SIZE;
    response = SD_Read_Block(addr, buf);
    printf("*** SD_Read_Sector ***\n");
    printf("Sector=%d 0x%x Addr=0x%08x Response=%d\n", (int) sector, (int) sector, (int) addr, (int) response);
    for (y = 0; y < 32; y++)
    {
        printf("%08x ", (int) addr);
        for (x = 0; x < 16; x++) printf("%02x ", buf[y*16+x]);
        printf(" \n");
        addr = addr + 16;
    }
    printf(" \n");
}

//--------------------------
// Dump Cache (Logical Dump)
//--------------------------
void Dump_Cache(sSectorCache *cache, UI32 sector)
{
    UI32 offset;
    UI08 data;
    UI08 x, y;

    printf("*** Dump FAT16 Cache ***\n");
    if (cache == &gFatCache) printf("Fat Cache\n");
    if (cache == &gDatCache) printf("Dat Cache\n");
    printf("Sector=%d 0x%x\n", (int) sector, (int) sector);
    offset = 0;
    for (y = 0; y < 32; y++)
    {
        printf("%08x ", (int) offset);
        for (x = 0; x < 16; x++)
        {
            Read_Data_from_Cache(cache, sector, offset, &data);
            printf("%02x ", data);
            offset = offset + 1;
        }
        printf(" \n");
    }
    printf(" \n");
}

//---------------
// End of Program
//---------------
