//=========================================================
// H8SX/1655 Project
//=========================================================
// File Name : touch.c
// Function  : Control Touch Panel
//---------------------------------------------------------
// Rev.01 2009.09.08 Takanaga Yamazaki
//---------------------------------------------------------
// Copyright (C) 2009, 2010 Takanaga Yamazaki
//=========================================================
// ---- License Information -------------------------------
// Anyone can FREELY use this code fully or partially
// under conditions shown below.
// 1. You may use this code only for individual purpose,
//    and educational purpose.
//    Do not use this code for business even if partially.
// 2. You should use this code under the GNU GPL.
// 3. You should remain this header text in your codes
//   including Copyright credit and License Information.
// 4. Your codes should inherit this license information.
//=========================================================
// ---- Patent Notice -------------------------------------
// I have not cared whether this system (hw + sw) causes 
// infringement on the patent, copyright, trademark,
// or trade secret rights of others. You have all 
// responsibilities for determining if your designs
// and products infringe on the intellectual property
// rights of others, when you use technical information
// included in this system for your business.
//=========================================================
// ---- Disclaimers ---------------------------------------
// The function and reliability of this system are not 
// guaranteed. They may cause any damages to loss of
// properties, data, money, profits, life, or business.
// By adopting this system even partially, you assume
// all responsibility for its use.
//=========================================================

#include "api.h"

//================
// Globals
//================
UI32 gTouch = 0;
UI32 gTouch_X;
UI32 gTouch_Y;
//------------------
// Calibration Data
//------------------
FLOT gfCal_LL_X = 107.0;
FLOT gfCal_LL_Y = 173.0;
FLOT gfCal_HL_X = 900.0;
FLOT gfCal_HL_Y = 166.0;
FLOT gfCal_LH_X = 104.0;
FLOT gfCal_LH_Y = 860.0;
FLOT gfCal_HH_X = 910.0;
FLOT gfCal_HH_Y = 860.0;

//==========================
// Get Touch Information
//==========================
UI08 TOUCH_Get_Info(UI32 *touch_x, UI32 *touch_y)
{
    UI08 result = gTouch;
    FLOT fTouch_X = (FLOT) gTouch_X;
    FLOT fTouch_Y = (FLOT) gTouch_Y;;
    FLOT fCal_Xmin = (gfCal_LL_X + gfCal_LH_X) / 2;
    FLOT fCal_Xmax = (gfCal_HL_X + gfCal_HH_X) / 2;
    FLOT fCal_Ymin = (gfCal_LL_Y + gfCal_HL_Y) / 2;
    FLOT fCal_Ymax = (gfCal_LH_Y + gfCal_HH_Y) / 2;
    FLOT fposX = (fTouch_X - fCal_Xmin) / (fCal_Xmax - fCal_Xmin) * 320;
    FLOT fposY = (fTouch_Y - fCal_Ymin) / (fCal_Ymax - fCal_Ymin) * 240;
    fposX = (fposX < 0)? 0 : (fposX > 319)? 319 : fposX;
    fposY = (fposY < 0)? 0 : (fposY > 239)? 239 : fposY;

    *touch_x = (UI32) fposX;
    *touch_y = (UI32) fposY;
    return result;
}

//==========================
// Get Touch Raw Data
//==========================
UI08 TOUCH_Get_Raw_Data(UI32 *touch_x, UI32 *touch_y)
{
    *touch_x = gTouch_X;
    *touch_y = gTouch_Y;
    return gTouch;
}

//===========================
// Set State in IDLE
//===========================
void TOUCH_Set_State_IDLE(void)
{
    // PF3 : X1 = HIZ
    // PF4 : X2 = HIZ
    // PM0 : Y1 = HIZ
    // PM1 : Y2 = HIZ
    // PA0 : SS = HIZ
    Set_DDR_PF(Get_DDR_PF() & (~0x18));
    Set_DDR_PM(Get_DDR_PM() & (~0x03));
    Set_DDR_PA(Get_DDR_PA() & (~0x01));
}

//===========================
// Set State in SENSE
//===========================
void TOUCH_Set_State_SENSE(void)
{
    // PF3 : X1 = HIGH
    // PF4 : X2 = HIGH
    // PM0 : Y1 = HIZ
    // PM1 : Y2 = HIZ
    // PA0 : SS = LOW
    // Read P53 : Y1in (0=Release, 1=Touch)
    PF.DR.BYTE = PF.DR.BYTE | ( 0x18);
    PA.DR.BYTE = PA.DR.BYTE & (~0x01);
    Set_DDR_PF(Get_DDR_PF() | ( 0x18));
    Set_DDR_PM(Get_DDR_PM() & (~0x03));
    Set_DDR_PA(Get_DDR_PA() | ( 0x01));
}

//===========================
// Set State in MES_X
//===========================
void TOUCH_Set_State_MES_X(void)
{
    // PF3 : X1 = HIGH
    // PF4 : X2 = LOW
    // PM0 : Y1 = HIZ
    // PM1 : Y2 = HIZ
    // PA0 : SS = HIZ
    // Read P53 : Y1in
    PF.DR.BYTE = PF.DR.BYTE | ( 0x08);
    PF.DR.BYTE = PF.DR.BYTE & (~0x10);
    Set_DDR_PF(Get_DDR_PF() | ( 0x18));
    Set_DDR_PM(Get_DDR_PM() & (~0x03));
    Set_DDR_PA(Get_DDR_PA() & (~0x01));
}

//===========================
// Set State in MES_Y
//===========================
void TOUCH_Set_State_MES_Y(void)
{
    // PF3 : X1 = HIZ
    // PF4 : X2 = HIZ
    // PM0 : Y1 = LOW
    // PM1 : Y2 = HIGH
    // PA0 : SS = HIZ
    // Read P52 : X1in
    PM.DR.BIT.B0 = 0;
    PM.DR.BIT.B1 = 1;
    Set_DDR_PF(Get_DDR_PF() & (~0x18));
    Set_DDR_PM(Get_DDR_PM() | ( 0x03));
    Set_DDR_PA(Get_DDR_PA() & (~0x01));
}

//================================
// Control Touch Panel
//================================
// TMR Interrupt
//-----------------------------------
// STATE_IDLE        : Set  IDLE
//-----------------------------------
// STATE_SET_SENSE_0 : Set  SENSE
// STATE_GET_SENSE_1 : Read SENSE (if release, repeat)
//-----------------------------------
// STATE_SET_SENSE_0 : Set  SENSE
// STATE_GET_SENSE_1 : Read SENSE (sense touch!)
// STATE_GET_SENSE_2 : Read SENSE (sense touch!) + Set MES_X
//-----------------------------------
// STATE_ADC_X_CH3   : Start A/D Ch3(Y1in)
// STATE_GET_X_CH3   : Read Ch3(Y1in) + Set MEX_Y
// STATE_ADC_Y_CH2   : Start A/D Ch2(X1in)
// STATE_GET_Y_XH2   : Read Ch2(X1in) + Set SENSE
//-----------------------------------
// STATE_GET_SENSE   : Read SENSE
// ...
//-----------------------------------
enum TOUCH_STATE
{
    STATE_IDLE,
    STATE_SET_SENSE,   // Set Port State in SENSE
    STATE_GET_SENSE_0, // Remove Chattering
    STATE_GET_SENSE_1, // Remove Chattering
    STATE_GET_SENSE_2, // Remove Chattering, and Set Port State in MES_X
    STATE_ADC_X_CH3,   // Start ADC to Measure X position
    STATE_GET_X_CH3,   // Get X position, and Set Port State in MES_Y
    STATE_ADC_Y_CH2,   // Start ADC to Measure Y position
    STATE_GET_Y_CH2,   // Get Y position, and Set Port State in SENSE
    STATE_CHK_SENSE_0, // Check if still touched?
    STATE_CHK_SENSE_1, // Check if still touched?
    STATE_CHK_SENSE_2  // Check if still touched?, if OK, Set Information in Globals.
};

void TOUCH_Control(void)
{
    static UI08 state = STATE_IDLE;
    static UI32 touch_x;
    static UI32 touch_y;

    switch(state)
    {
        // Set Port State in SENSE
        case STATE_IDLE :
        {
            TOUCH_Set_State_IDLE();
            state = STATE_SET_SENSE;
            gTouch = 0;
            break;
        }
        // Remove Chattering
        case STATE_SET_SENSE :
        {
            TOUCH_Set_State_SENSE();
            state = STATE_GET_SENSE_0;
            break;
        }
        // Remove Chattering
        case STATE_GET_SENSE_0 :
        {
            if (P5.PORT.BIT.B3)
            {
                state = STATE_GET_SENSE_1;
            }
            else
            {
                gTouch = 0;
                state = STATE_SET_SENSE;
            }
            break;
        }
        // Remove Chattering
        case STATE_GET_SENSE_1 :
        {
            if (P5.PORT.BIT.B3)
            {
                state = STATE_GET_SENSE_2;
            }
            else
            {
                gTouch = 0;
                state = STATE_SET_SENSE;
            }
            break;
        }
        // Remove Chattering, and Set Port State in MES_X
        case STATE_GET_SENSE_2 :
        {
            if (P5.PORT.BIT.B3)
            {
                TOUCH_Set_State_MES_X();
                state = STATE_ADC_X_CH3;
            }
            else
            {
                gTouch = 0;
                state = STATE_SET_SENSE;
            }
            break;
        }
        // Start ADC to Measure X position
        case STATE_ADC_X_CH3 :
        {
            state = STATE_GET_X_CH3;
            break;
        }
        // Get X position, and Set Port State in MES_Y
        case STATE_GET_X_CH3 :
        {
            touch_x = ADC_Get_Data(3);
            TOUCH_Set_State_MES_Y();
            state = STATE_ADC_Y_CH2;
            break;
        }
        // Start ADC to Measure Y position
        case STATE_ADC_Y_CH2 :
        {
            state = STATE_GET_Y_CH2;
            break;
        }
        // Get Y position, and Set Port State in SENSE
        case STATE_GET_Y_CH2 :
        {
            touch_y = ADC_Get_Data(2);
            TOUCH_Set_State_SENSE();
            state = STATE_CHK_SENSE_0;
            break;
        }
        // Check if still touched?
        case STATE_CHK_SENSE_0 :
        {
            if (P5.PORT.BIT.B3)
            {
                state = STATE_CHK_SENSE_1;
            }
            else
            {
                gTouch = 0;
                state = STATE_SET_SENSE;
            }
            break;
        }
        // Check if still touched?
        case STATE_CHK_SENSE_1 :
        {
            if (P5.PORT.BIT.B3)
            {
                state = STATE_CHK_SENSE_2;
            }
            else
            {
                gTouch = 0;
                state = STATE_SET_SENSE;
            }
            break;
        }
        // Check if still touched?, if OK, Set Information in Globals.
        case STATE_CHK_SENSE_2 :
        {
            if (P5.PORT.BIT.B3)
            {
                gTouch_X = touch_x;
                gTouch_Y = touch_y;
                gTouch = 1;
            }
            else
            {
                gTouch = 0;
            }
            state = STATE_SET_SENSE;
            break;
        }
        default :
        {
            state = STATE_IDLE;
            break;
        }
    }
}

//======================================================
// End of Program
//======================================================

