//=========================================================
// H8SX/1655 Project
//=========================================================
// File Name : lcdpanel.c
// Function  : Control TFT LCD Panel
//---------------------------------------------------------
// Rev.01 2009.09.13 Takanaga Yamazaki
//---------------------------------------------------------
// Copyright (C) 2009, 2010 Takanaga Yamazaki
//=========================================================
// ---- License Information -------------------------------
// Anyone can FREELY use this code fully or partially
// under conditions shown below.
// 1. You may use this code only for individual purpose,
//    and educational purpose.
//    Do not use this code for business even if partially.
// 2. You should use this code under the GNU GPL.
// 3. You should remain this header text in your codes
//   including Copyright credit and License Information.
// 4. Your codes should inherit this license information.
//=========================================================
// ---- Patent Notice -------------------------------------
// I have not cared whether this system (hw + sw) causes 
// infringement on the patent, copyright, trademark,
// or trade secret rights of others. You have all 
// responsibilities for determining if your designs
// and products infringe on the intellectual property
// rights of others, when you use technical information
// included in this system for your business.
//=========================================================
// ---- Disclaimers ---------------------------------------
// The function and reliability of this system are not 
// guaranteed. They may cause any damages to loss of
// properties, data, money, profits, life, or business.
// By adopting this system even partially, you assume
// all responsibility for its use.
//=========================================================

#include <stdlib.h>
#include "api.h"

//==========================
// TFT LCD Timing Parameters
//==========================
    // [Specification of TFT LCD Panel (QVGA)]
    //     DCLK = 6MHz
    //     Th  = 408 DCLK (horizontal whole cycle)
    //     Thb =  38 DCLK (horizontal back porch)
    //     Thd = 320 DCLK (horizontal display period)
    //     Thf =  50 DCLK (horizontal front porch)
    //
    //     Tv  = 262 Th (vertical whole period)
    //     Tvb =  18 Th (vertical back porch)
    //     Tvd = 240 Th (vertical display period)
    //     Tvf =   4 Th (vertical front porch)
    //
#define TFT_HCYC 408
#define TFT_HBKP  38
#define TFT_HDIS 320
#define TFT_HFRP  50
#define TFT_VCYC 262
#define TFT_VBKP  18
#define TFT_VDIS 240
#define TFT_VFRP   4
#define TFT_EXDA   2 // EXDMAC Timing Adjustment
//
#define ADDR_EXT_SRAM 0x00200000 // top address of external SRAM

//==============
// Globals
//==============
UI32 gFMEM_TOP;
UI16 gTFT_VCOUNT = (TFT_VBKP+ TFT_VDIS + 1); // VSYNC Counter
// initial value above is a default for TFT_Panel_Wait_for_V_Blank_Start().
UI32 gTFT_ACTIVE; // is 1 if VCOUNT is in active region

UI08 gKey = 0x00;

//======================
// Initialize TFT Panel
//======================
void Init_TFT_Panel(sFrame *psFrame, UI16 color)
{
    UI32 x, y;
    UI16 **ppFrame;
    //-----------------------------
    // Setup Frame Memory
    //-----------------------------
    gFMEM_TOP = ADDR_EXT_SRAM;
    ppFrame = (UI16**) malloc(sizeof(UI16*) * TFT_VDIS);
    //
    psFrame->sizeX = TFT_HDIS;
    psFrame->sizeY = TFT_VDIS;
    psFrame->ppFrame = ppFrame;
    //
    for (y = 0; y < TFT_VDIS; y++)
    {
        ppFrame[y] = (UI16*) (gFMEM_TOP + y * TFT_HDIS * sizeof(UI16));
        for (x = 0; x < TFT_HDIS; x++)
        {
            ppFrame[y][x] = color;
        }
    }
    //---------------------------
    // Send Global Reset Command
    //---------------------------
    // Wait 10ms
    TMR_Wait_One_Tick();
    // Global Reset ON
    TFT_SPI_Tx_Register(0x04, 0x01);
    // Wait 10ms
    TMR_Wait_One_Tick();
    // Global Reset OFF
    TFT_SPI_Tx_Register(0x04, 0x00);
    // Wait 10ms
    TMR_Wait_One_Tick();
    //-----------------------------------
    // Start System Clock of TPU & EXDMAC
    //-----------------------------------
    MSTP.CRA.BIT._TPUL   = 0;
    MSTP.CRA.BIT._EXDMAC = 0;
    //---------------------------------------------------
    // {Pin Assign]
    //     P21 : TPU3 TIOCA3 TFT_HSYNC
    //     P22 : TPU3 TIOCC3 TFT_VSYNC
    //     P24 : TPU4 TIOCB4 TFT EXDMA Request /EDREQ1
    //     P25 : TPU4 TIOCA4 TFT_DE
    //     P26 : TPU5 TIOCA5 TFT_DCLK
    //     P16 : EXDMAC1 /EDACK1
    //---------------------------------------------------
    //-----------------------
    // TPU Common Setting
    //-----------------------
    TPU.TSTR.BYTE = 0x00;   // Stop TCNT
    //
    TPU.TSYR.BIT.SYNC0 = 0;
    TPU.TSYR.BIT.SYNC1 = 0;
    TPU.TSYR.BIT.SYNC2 = 1; // SYNC TCNT2
    TPU.TSYR.BIT.SYNC3 = 1; // SYNC TCNT3
    TPU.TSYR.BIT.SYNC4 = 1; // SYNC TCNT4
    TPU.TSYR.BIT.SYNC5 = 0;
    //---------------------------------------------------
    // [TCNT2]
    //     Mode = Normal
    //     Count Clock = PCLK / 1 = 24MHz
    //     Count Clear = TGRA = (TFT_HCYC * 4 - 1)
    //     Interrupt   = TGRB = ((TFT_HFRP + TFT_HBKP) * 4)
    //---------------------------------------------------
    TPU2.TCR.BIT.CCLR = 1; // Clear by TGRA Compare Match
    TPU2.TCR.BIT.CKEG = 1; // Count Up at Rising Edge of Count Clock
    TPU2.TCR.BIT.TPSC = 0; // Count Clock = PCLK / 1
    //
    TPU2.TMDR.BIT.MD = 0;  // Normal Mode
    //
    TPU2.TIOR.BIT.IOA = 0; // TGRA : Output Compare (output disable)
    TPU2.TIOR.BIT.IOB = 0; // TGRB : Output Compare (output disable)
    //
    TPU2.TIER.BYTE = 0x02; // Enable TGRB Interrupt (TGIEB = 1)
    //
    TPU2.TCNT = 0x0000;
    //
    TPU2.TGRA = (TFT_HCYC * 4 - 1);
    TPU2.TGRB = ((TFT_HFRP + TFT_HBKP) * 4);
    //---------------------------------------------------
    // [TCNT3]
    //     Mode = PWM1
    //     Count Clock = PCLK / 1 = 24MHz
    //     Count Clear = Synchronized with TCNT2 Clear
    //     Initial Value of TIOCA3 (TGRA) = 1 (HSYNC)
    //     Initial Value of TIOCC3 (TGRC) = 1 (VSYNC)
    //     Output Value of TGRA = 1
    //     Output Value of TGRB = 0
    //     Output Value of TGRC = 1
    //     Output Value of TGRD = 0
    //     TGRA = ((TFT_HFRP + 2) * 4)
    //     TGRB = (TFT_HFRP * 4)
    //     TGRC = Active Period : (TFT_HFRP * 4) (make TIOCC3 high)
    //            Blank  Period : 0xffff
    //     TGRD = Active Period : (TFT_HFRP * 4) (make TIOCC3 low)
    //            Blank  Period : 0xffff
    //---------------------------------------------------
    TPU.TSTR.BIT.CST3 = 0; // Stop TCNT
    //
    TPU3.TCR.BIT.CCLR = 7; // Clear by Synchronized Channel
    TPU3.TCR.BIT.CKEG = 1; // Count Up at Rising Edge of Count Clock
    TPU3.TCR.BIT.TPSC = 0; // Count Clock = PCLK / 1
    //
    TPU3.TMDR.BIT.BFA = 0; // No Buffer Operation A
    TPU3.TMDR.BIT.BFB = 0; // No Buffer Operation A
    TPU3.TMDR.BIT.MD  = 2; // PWM Mode 1
    //
    TPU3.TIOR.BIT.IOA = 6; // TRGA Initial=1, Output=1
    TPU3.TIOR.BIT.IOB = 1; // TRGB Initial=0, Output=0 (Initial is ignored)
    TPU3.TIOR.BIT.IOC = 6; // TRGC Initial=1, Output=1
    TPU3.TIOR.BIT.IOD = 1; // TRGB Initial=0, Output=0 (Initial is ignored)
    //
    TPU3.TIER.BYTE = 0x00; // No Interrupt
    //
    TPU3.TCNT = 0x0000;
    //
    TPU3.TGRA = ((TFT_HFRP + 2) * 4);
    TPU3.TGRB = (TFT_HFRP * 4);
    TPU3.TGRC = (0xffff);
    TPU3.TGRD = (0xffff);

    //---------------------------------------------------
    // [TCNT4]
    //     Mode = PWM2
    //     Count Clock = PCLK / 1 = 24MHz
    //     Count Clear = Synchronized with TCNT2 Clear
    //     Initial Value of TIOCA4 (TGRA) = 0 (DE)
    //     Initial Value of TIOCB4 (TGRB) = 1 (/EDREQ1)
    //     Output Value of TGRA = 1
    //     Output Value of TGRB = 0
    //     TGRA = Active Period : ((TFT_HFRP + TFT_HBKP) * 4)
    //            Blank  Period : 0xffff
    //     TGRB = Active Period : ((TFT_HFRP + TFT_HBKP - TFT_EXDA) * 4)
    //            Blank  Period : 0xffff
    //---------------------------------------------------
    TPU.TSTR.BIT.CST4 = 0; // Stop TCNT
    //
    TPU4.TCR.BIT.CCLR = 3; // Clear by Synchronized Channel
    TPU4.TCR.BIT.CKEG = 1; // Count Up at Rising Edge of Count Clock
    TPU4.TCR.BIT.TPSC = 0; // Count Clock = PCLK / 1
    //
    TPU4.TMDR.BIT.MD = 3;  // PWM Mode 2
    //
    TPU4.TIOR.BIT.IOA = 2; // TGRB Initial=0, Output=1
    TPU4.TIOR.BIT.IOB = 5; // TGRB Initial=1, Output=0
    //
    TPU4.TIER.BYTE = 0x00; // No Interrupt
    //
    TPU4.TCNT = 0x0000;
    //
    TPU4.TGRA = (0xffff);
    TPU4.TGRB = (0xffff);
    //---------------------------------------------------
    // [TCNT5]
    //     Mode = PWM2
    //     Count Clock = PCLK = 24MHz
    //     Count Clear = TGRB = (4 - 1)
    //     Initial Value of TIOCA5 (TGRA) = 0 (DCLK)
    //     Output Value of TGRA = 1
    //     TGRA = 2
    //---------------------------------------------------
    TPU.TSTR.BIT.CST5 = 0; // Stop TCNT
    //
    TPU5.TCR.BIT.CCLR = 2; // Clear by TGRB Compare Match
    TPU5.TCR.BIT.CKEG = 0; // Count Up at Falling Edge of Count Clock
    TPU5.TCR.BIT.TPSC = 0; // Count Clock = PCLK / 1
    //
    TPU5.TMDR.BIT.MD  = 3; // PWM Mode 2
    //
    TPU5.TIOR.BIT.IOA = 2; // TGRA Initial=0, Output=1
    TPU5.TIOR.BIT.IOB = 0; // TGRB Output Compare (output disable)
    //
    TPU5.TIER.BYTE = 0x00; // No Interrupt
    //
    TPU5.TCNT = 0x0000;
    //
    TPU5.TGRA = (2   );
    TPU5.TGRB = (4- 1);
    //--------------------
    // Initialize Globals
    //--------------------
    gTFT_VCOUNT = 0;
    gTFT_ACTIVE = 0;
    //----------------------------
    // Set Interrupt Request Level
    //----------------------------
    INTC.IPRG.BIT._TPU2 = 6;
    //-------------------
    // Start Timer Count
    //-------------------
    TPU.TSTR.BYTE = 0x3c; // Start TCNT2,3,4,5
}

//==============================
// TFT SPI Send Data to Register
//==============================
void TFT_SPI_Tx_Register(UI08 addr, UI08 data)
{
    // Set SPI Baud Rate 100kbps
    SPI_Set_BaudRate(100);
    // Assert CS
    SPI_Control_CS(SPI_CS_TFT, SPI_CS_ASSERT);
    // Send Address
    SPI_TxRx(addr);
    // Send Data
    SPI_TxRx(data);
    // Negate CS
    SPI_Control_CS(SPI_CS_TFT, SPI_CS_NEGATE);
}

//====================================
// TFT SPI Receive Data from Register
//====================================
// [Note] As for TFT Panel, there is no Serial Output Pin,
//        so this routine has no meanings.
UI08 TFT_SPI_Rx_Register(UI08 addr)
{
    UI08 data;
    // Set SPI Baud Rate 100kbps
    SPI_Set_BaudRate(100);
    // Assert CS
    SPI_Control_CS(SPI_CS_TFT, SPI_CS_ASSERT);
    // Send Address
    SPI_TxRx(addr | 0x80);
    // Send Data
    data = SPI_TxRx(0xff);
    // Negate CS
    SPI_Control_CS(SPI_CS_TFT, SPI_CS_NEGATE);

    return data;
}

//========================
// Wait for V-Blank Start
//========================
void TFT_Panel_Wait_for_V_Blank_Start(void)
{
    while(gTFT_VCOUNT != (TFT_VBKP+ TFT_VDIS + 1));
}

//=============================
// Is it in Active period?
//=============================
UI32 TFT_Is_Active_Period(void)
{
    return gTFT_ACTIVE; // active period
}

//================================
// Interrupt Handler : TPU2 TGRB4
//================================
// Just after beginning of (pseudo) active timing (even if its in blanking).
void Int_Handler_TPU2_TGI2B(void)
{
    //----------------------
    // Active Period or not?
    //----------------------
    if (gTFT_VCOUNT == TFT_VBKP) gTFT_ACTIVE = 1;
    if (gTFT_VCOUNT == (TFT_VBKP + TFT_VDIS)) gTFT_ACTIVE = 0;
    //-----------------------------------------
    // Configure DE and EXDREQ1 Waveform
    //-----------------------------------------
    if (TFT_Is_Active_Period())
    {
        TPU4.TGRA = ((TFT_HFRP + TFT_HBKP           ) * 4); // DE    (TIOCA4) fall
        TPU4.TGRB = ((TFT_HFRP + TFT_HBKP - TFT_EXDA) * 4); // EXDREQ(TIOCB4) rise
    }
    else
    {
        TPU4.TGRA = (0xffff); // DE     (TIOCA4) fall
        TPU4.TGRB = (0xffff); // EXDREQ1(TIOCB4) rise
    }
    //---------------------------------
    // Configure VSYNC Waveform
    //---------------------------------
    if (gTFT_VCOUNT == 0)
    {
        TPU3.TGRC = (0xffff);   // VSYNC(TIOCC3) rise
        TPU3.TGRD = (TFT_HFRP * 4); // VSYNC(TIOCC3) fall
    }
    else if (gTFT_VCOUNT == 1)
    {
        TPU3.TGRC = (TFT_HFRP * 4); // VSYNC(TIOCC3) rise
        TPU3.TGRD = (0xffff);   // VSYNC(TIOCC3) fall
    }
    else
    {
        TPU3.TGRC = (0xffff);   // VSYNC(TIOCC3) rise
        TPU3.TGRD = (0xffff);   // VSYNC(TIOCC3) fall
    }
    //------------------------
    // Configure EXDMAC1
    //------------------------
    // [EXDMAC1]
    //     Address Mode = Single Address Mode
    //     Transfer Mode = Block Transfer Mode
    //     No Repeat Area (Continue Source Address Increment)
    //     Block Size(EDBSR) = 2bytes * 320pixel = 640bytes
    //     EDSAR = 0x00200000 (initial value)
    //     EDDAR = 0x00000000 (unused)
    //     EDOFR = 0x00000000 (unused)
    //     EDTCR = 320 * 240 * 2 = 153600bytes
    //     EDBSR = (320 << 16) + (320 << 0)
    //     EDMDR = DTE:1
    //             EDACKE:1
    //             ETENDE:0
    //             EDRAKE:0
    //             EDREQS:1 (falling edge)
    //             NRD:0
    //             DTSZ:1 (word size)
    //             MDS:1 (block transfer mode)
    //             TSEIE:0
    //             ESIE:0
    //             DTIE:0
    //             DTF:3 (external request)
    //             EDMAP:7 (highest priority)
    //     EDACR = AMS:1 (single address mode)
    //             DIRS:0 (EDSAR is source address)
    //             RPTIE:0
    //             ARS:2 (no block/repeat area)
    //             SAT:2 (increment source address)
    //             DAT:0 (unused)
    //             SARIE:0
    //             SARA:0
    //             DARIE:0
    //             DARA:0
    //
    if (gTFT_VCOUNT == TFT_VBKP) // start active period
    {
        EXDMAC1.EDSAR = gFMEM_TOP;
        EXDMAC1.EDDAR = 0x00000000;
        EXDMAC1.EDOFR = 0x00000000;
        EXDMAC1.EDTCR = 320 * 240 * 2;
        EXDMAC1.EDBSR = (320 << 16) + (320 << 0);
        EXDMAC1.EDMDR.LONG = 0x480050c7; // 0100_1000_0000_0000_0101_0000_1100_0111
        EXDMAC1.EDACR.LONG = 0x83200000; // 1000_0011_0010_0000_0000_0000_0000_0000
        //
        EXDMAC1.EDMDR.BIT.DTE = 1; // start and wait for EDREQ1
    }
    else if (gTFT_VCOUNT == (TFT_VBKP+ TFT_VDIS)) // start blanking period
    {
        UI32 edmdr;
        edmdr = EXDMAC1.EDMDR.LONG;
        EXDMAC1.EDMDR.LONG = 0x00000000; // clear all flags
    }
    //------------------------
    // Update Timing Globals
    //------------------------
    gTFT_VCOUNT++;
    if (gTFT_VCOUNT >= TFT_VCYC)
    {
        gTFT_VCOUNT = 0;
    }
    //-----------------------
    // Clear Interrupt Flag
    //-----------------------
    TPU2.TSR.BIT.TGFB = 0;
}

//======================================================
// End of Program
//======================================================
