//=========================================================
// H8SX/1655 Project
//=========================================================
// File Name : rtc.c
// Function  : Control RTC (DS1305)
//---------------------------------------------------------
// Rev.01 2009.09.05 Takanaga Yamazaki
//---------------------------------------------------------
// Copyright (C) 2009, 2010 Takanaga Yamazaki
//=========================================================
// ---- License Information -------------------------------
// Anyone can FREELY use this code fully or partially
// under conditions shown below.
// 1. You may use this code only for individual purpose,
//    and educational purpose.
//    Do not use this code for business even if partially.
// 2. You should use this code under the GNU GPL.
// 3. You should remain this header text in your codes
//   including Copyright credit and License Information.
// 4. Your codes should inherit this license information.
//=========================================================
// ---- Patent Notice -------------------------------------
// I have not cared whether this system (hw + sw) causes 
// infringement on the patent, copyright, trademark,
// or trade secret rights of others. You have all 
// responsibilities for determining if your designs
// and products infringe on the intellectual property
// rights of others, when you use technical information
// included in this system for your business.
//=========================================================
// ---- Disclaimers ---------------------------------------
// The function and reliability of this system are not 
// guaranteed. They may cause any damages to loss of
// properties, data, money, profits, life, or business.
// By adopting this system even partially, you assume
// all responsibility for its use.
//=========================================================

#include "api.h"

//====================
// Global Variables
//====================
sDateTime gDateTime; // Current Clock Database

//==================================
// Define RTC Register Address
//==================================
#define RTC_SEC   0x00
#define RTC_MIN   0x01
#define RTC_HOUR  0x02
#define RTC_WEEK  0x03
#define RTC_DATE  0x04
#define RTC_MONTH 0x05
#define RTC_YEAR  0x06
#define RTC_ALARM0_SEC  0x07
#define RTC_ALARM0_MIN  0x08
#define RTC_ALARM0_HOUR 0x09
#define RTC_ALARM0_DAY  0x0a
#define RTC_ALARM1_SEC  0x0b
#define RTC_ALARM1_MIN  0x0c
#define RTC_ALARM1_HOUR 0x0d
#define RTC_ALARM1_DAY  0x0e
#define RTC_CONTROL     0x0f
#define RTC_STATUS      0x10
#define RTC_CHARGER     0x11
#define RTC_RAM(addr)  (0x20 + addr)

//====================
// RTC Hex from BCD
//====================
UI08 RTC_Hex_from_BCD(UI08 bcd)
{
    UI08 hex;

    hex = bcd & 0x0f;
    hex = hex + ((bcd >> 4) & 0x0f) * 10;
    return hex;
}

//====================
// RTC BCD from Hex
//====================
UI08 RTC_BCD_from_Hex(UI08 hex)
{
    UI08 bcd;

    bcd = hex % 10;
    bcd = bcd | ((hex / 10) << 4);
    return bcd;
}

//============================
// RTC Write Register
//============================
void RTC_Write(UI08 addr, UI08 data)
{
    UI08 imask;

    // Wait for Blanking Period
    // (to avoid TFT display corruption during interrupt mask)
    TFT_Panel_Wait_for_V_Blank_Start();

    // Set SPI Baud Rate 10kbps
    SPI_Set_BaudRate(10);
    // Assert CS
    SPI_Control_CS(SPI_CS_RTC, SPI_CS_ASSERT);
    // Wait for 100us, or over
    CPU_Wait_N_us(100);
    // Save Current Interrupt Level and Mask All Interrupts
    // (not to make gap between address and data)
    imask = get_imask_exr();
    set_imask_exr(7);
    // Send Address
    SPI_Tx(addr | 0x80);
    // Send Data
    SPI_Tx(data);
    // Revert Interrupt Mask Level
    set_imask_exr(imask);
    // Wait for 100us, or over
    CPU_Wait_N_us(100);
    // Negate CS
    SPI_Control_CS(SPI_CS_RTC, SPI_CS_NEGATE);
    // Wait for 100us, or over
    CPU_Wait_N_us(100);
}

//============================
// RTC Read Register
//============================
UI08 RTC_Read(UI08 addr)
{
    UI08 data;
    UI08 imask;

    // Wait for Blanking Period
    // (to avoid TFT display corruption during interrupt mask)
    TFT_Panel_Wait_for_V_Blank_Start();

    // Set SPI Baud Rate 10kbps
    SPI_Set_BaudRate(10);
    // Assert CS
    SPI_Control_CS(SPI_CS_RTC, SPI_CS_ASSERT);
    // Wait for 100us, or over
    CPU_Wait_N_us(100);
    // Save Current Interrupt Level and Mask All Interrupts
    // (not to make gap between address and data)
    imask = get_imask_exr();
    set_imask_exr(7);
    // Send Address
    SPI_Tx(addr | 0x00);
    // Receive Data
    data = SPI_Rx();
    // Revert Interrupt Mask Level
    set_imask_exr(imask);
    // Wait for 100us, or over
    CPU_Wait_N_us(100);
    // Negate CS
    SPI_Control_CS(SPI_CS_RTC, SPI_CS_NEGATE);
    // Wait for 100us, or over
    CPU_Wait_N_us(100);

    return data;
}

//======================
// Initialize RTC
//======================
void Init_RTC(void)
{
    RTC_Write(RTC_CONTROL, 0x00); // EOSC_N=0, WP=0
    RTC_Write(RTC_CHARGER, 0x5c); // same as initial state
}

//=============================
// Set RTC
//=============================
void RTC_Set(sDateTime *psDateTime)
{
    RTC_Write(RTC_SEC,   psDateTime->bcd_seconds);
    RTC_Write(RTC_MIN,   psDateTime->bcd_minutes);
    RTC_Write(RTC_HOUR,  psDateTime->bcd_hours);
    RTC_Write(RTC_DATE,  psDateTime->bcd_days);
    RTC_Write(RTC_MONTH, psDateTime->bcd_months);
    RTC_Write(RTC_YEAR,  psDateTime->bcd_years);
    RTC_Write(RTC_WEEK,  psDateTime->weekdays);
}

//=============================
// Get RTC
//=============================
void RTC_Get(sDateTime *psDateTime)
{
    psDateTime->bcd_seconds = RTC_Read(RTC_SEC);
    psDateTime->bcd_minutes = RTC_Read(RTC_MIN);
    psDateTime->bcd_hours   = RTC_Read(RTC_HOUR);
    psDateTime->bcd_days    = RTC_Read(RTC_DATE);
    psDateTime->bcd_months  = RTC_Read(RTC_MONTH);
    psDateTime->bcd_years   = RTC_Read(RTC_YEAR);
    psDateTime->weekdays    = RTC_Read(RTC_WEEK);
}

//===========================
// Check RTC Power Failure
//===========================
// if OK, return 1, else 0
UI08 RTC_Check_Power(void)
{
    static UI32 count = 0; // noise reduction
    UI08 result;

    result = 1;
    if (PA.PORT.BIT.B6 == 1)
    {
        count = 0;
    }
    else
    {
        count++;
        if (count >= 16)
        {
            count = 0;
            result = 0;
        }
    }

    return result;
}

//======================================================
// End of Program
//======================================================
