//=========================================================
// H8SX/1655 Project
//=========================================================
// File Name : tmr.c
// Function  : TMR Control
//---------------------------------------------------------
// Rev.01 2009.08.13 Takanaga Yamazaki
//---------------------------------------------------------
// Copyright (C) 2009, 2010 Takanaga Yamazaki
//=========================================================
// ---- License Information -------------------------------
// Anyone can FREELY use this code fully or partially
// under conditions shown below.
// 1. You may use this code only for individual purpose,
//    and educational purpose.
//    Do not use this code for business even if partially.
// 2. You should use this code under the GNU GPL.
// 3. You should remain this header text in your codes
//   including Copyright credit and License Information.
// 4. Your codes should inherit this license information.
//=========================================================
// ---- Patent Notice -------------------------------------
// I have not cared whether this system (hw + sw) causes 
// infringement on the patent, copyright, trademark,
// or trade secret rights of others. You have all 
// responsibilities for determining if your designs
// and products infringe on the intellectual property
// rights of others, when you use technical information
// included in this system for your business.
//=========================================================
// ---- Disclaimers ---------------------------------------
// The function and reliability of this system are not 
// guaranteed. They may cause any damages to loss of
// properties, data, money, profits, life, or business.
// By adopting this system even partially, you assume
// all responsibility for its use.
//=========================================================

#include "api.h"
#include "application.h"

#define TMR0_CYCLE_16 (24*1000*1000 * 16/(1000*1000) / 2)  // A/D Cycle = 16us
#define TMR0_CYCLE_20 (24*1000*1000 * 20/(1000*1000) / 2)  // A/D Cycle = 20us
#define TMR1_CYCLE  (24*1000*1000 * 80/(1000*1000) / 8)  // A/D Cycle = 80us
#define TMR67_CYCLE (24*1000*1000 * TICK_UNIT/1000 / 8) // System Tick Cycle

//=============
// Globals
//=============
UI32 gTicks;

//==================================
// Initialize TMR0 (8bit)
//==================================
void Init_TMR0(UI08 cyc)
{
    //=========================================
    // Configure TMR0 for A/D Conversion Cycle
    //=========================================
    //--------------------------
    // Start TMR01 System Clock
    //--------------------------
    MSTP.CRA.BIT._TMR01  = 0;
    //----------------------------
    // Set Interrupt Request Level
    //----------------------------
    INTC.IPRH.BIT._TMR0 = 3;
    //----------------------------
    // Start TMR0 for A/D Trigger
    //----------------------------
    TMR0.TCORA = (cyc == 0)? (UI08) TMR0_CYCLE_16 : (UI08) TMR0_CYCLE_20;
    //
    TMR0.TCR.BIT.CCLR = 1;  // Clear 8bit TCNT by CMA
    TMR0.TCSR.BIT.ADTE = 1; // Generate A/D Trigger
    //
    TMR0.TCR.BIT.CKS = 1;   // PCLK/2
    TMR0.TCCR.BIT.ICKS = 1; // CPLK/2
    //
    TMR0.TCR.BIT.CMIEA = 1;
}

//===========================
// Interrupt Handler TMR0
//===========================
void Int_Handler_TMR0_CMIA(void)
{
    //---------------------
    // Hook for Application
    //---------------------
    Hook_Int_Handler_TMR0_CMIA();
    //----------------------
    // Clear Interrupt Flag
    //----------------------
    TMR0.TCSR.BIT.CMFA = 0;
}

//==================================
// Initialize TMR1 (8bit)
//==================================
void Init_TMR1(void)
{
    //=========================================
    // Configure TMR1 for D/A Wave Output
    //=========================================
    //--------------------------
    // Start TMR01 System Clock
    //--------------------------
    MSTP.CRA.BIT._TMR01  = 0;
    //----------------------------
    // Set Interrupt Request Level
    //----------------------------
    INTC.IPRH.BIT._TMR1 = 3;
    //----------------------------
    // Start TMR1 for A/D Trigger
    //----------------------------
    TMR1.TCORA = (UI08) TMR1_CYCLE;
    //
    TMR1.TCR.BIT.CCLR = 1;  // Clear 8bit TCNT by CMA
    //
    TMR1.TCR.BIT.CKS = 1;   // PCLK/8
    TMR1.TCCR.BIT.ICKS = 2; // CPLK/8
    //
    TMR1.TCR.BIT.CMIEA = 1;
}

//===========================
// Interrupt Handler TMR1
//===========================
void Int_Handler_TMR1_CMIA(void)
{
    //---------------------
    // Hook for Application
    //---------------------
    Hook_Int_Handler_TMR1_CMIA();
    //----------------------
    // Clear Interrupt Flag
    //----------------------
    TMR1.TCSR.BIT.CMFA = 0;
}

//==================================
// Initialize TMR67 (16bit Cascade)
//==================================
void Init_TMR67(void)
{
    //=========================================
    // Configure TMR67 for System Tick
    //=========================================
    //--------------------------
    // Start TMR67 System Clock
    //--------------------------
    MSTP._CRC.BIT._TMR67 = 0;
    //----------------------------
    // Set Interrupt Request Level
    //----------------------------
    INTC.IPRR.BIT._TMR47 = 4;
    //---------------------
    // Start TMR67
    //---------------------
    TMR6.TCORA = (UI08) (TMR67_CYCLE >> 8);
    TMR7.TCORA = (UI08) (TMR67_CYCLE & 0x00ff);
    //
    TMR6.TCR.BIT.CCLR = 1; // Clear 16bit TCNT by CMA
    //
    TMR6.TCR.BIT.CKS   = 4; // Count up at Overflow of TMR7
    TMR6.TCCR.BIT.ICKS = 0; // Count up at Overflow of TMR7
    TMR7.TCR.BIT.CKS   = 1; // PCLK/8
    TMR7.TCCR.BIT.ICKS = 2; // PCLK/8
    //
    TMR6.TCR.BIT.CMIEA = 1;
    //-----------------------
    // Initialize Globals
    //-----------------------
    gTicks = 0;
}

//===========================
// Interrupt Handler TMR6
//===========================
void Int_Handler_TMR6_CMIA(void)
{
    static UI32 count_1sec = 0;
    //------------------------
    // Increment Global Tick
    //------------------------
    gTicks++;
    //--------------
    // Toggle LED
    //--------------
    if (count_1sec == 0)
    {
        LED(LED_TOGGLE);
    }
    count_1sec = (count_1sec + 1) % (100 / 2); // make 1sec cycle
    //---------------------
    // Hook for Application
    //---------------------
    Hook_Int_Handler_TMR6_CMIA();
    //----------------------
    // Clear Interrupt Flag
    //----------------------
    TMR6.TCSR.BIT.CMFA = 0;
}

//====================
// Get Tick Value
//====================
UI32 TMR_Get_Ticks(void)
{
    return gTicks;
}

//=============================
// Wait for 1 tick, at least
//=============================
void TMR_Wait_One_Tick(void)
{
    UI32 tick_prev;

    tick_prev = TMR_Get_Ticks();
    while(TMR_Get_Ticks() < (tick_prev + 2));
}

//======================================
// Wait for over N us, without using TMR
// (Software Wait Routine)
//======================================
void CPU_Wait_N_us(UI32 N)
{
    volatile UI32 i;

    if (N == 0) N = 1;
    for (i = 0; i < (N * 1000 / 20); i++);
}

//======================================================
// End of Program
//======================================================
