//=========================================================
// H8SX/1655 Project
//=========================================================
// File Name : stub.c
// Function  : GDB Stub Support Routines
//---------------------------------------------------------
// Rev.01 2009.08.23 Takanaga Yamazaki
//---------------------------------------------------------
// Copyright (C) 2009, 2010 Takanaga Yamazaki
//=========================================================
// ---- License Information -------------------------------
// Anyone can FREELY use this code fully or partially
// under conditions shown below.
// 1. You may use this code only for individual purpose,
//    and educational purpose.
//    Do not use this code for business even if partially.
// 2. You should use this code under the GNU GPL.
// 3. You should remain this header text in your codes
//   including Copyright credit and License Information.
// 4. Your codes should inherit this license information.
//=========================================================
// ---- Patent Notice -------------------------------------
// I have not cared whether this system (hw + sw) causes
// infringement on the patent, copyright, trademark,
// or trade secret rights of others. You have all
// responsibilities for determining if your designs
// and products infringe on the intellectual property
// rights of others, when you use technical information
// included in this system for your business.
//=========================================================
// ---- Disclaimers ---------------------------------------
// The function and reliability of this system are not
// guaranteed. They may cause any damages to loss of
// properties, data, money, profits, life, or business.
// By adopting this system even partially, you assume
// all responsibility for its use.
//=========================================================

/* Remote communication protocol.

   A debug packet whose contents are <data>
   is encapsulated for transmission in the form:

	$ <data> # CSUM1 CSUM2

	<data> must be ASCII alphanumeric and cannot include characters
	'$' or '#'.  If <data> starts with two characters followed by
	':', then the existing stubs interpret this as a sequence number.

	CSUM1 and CSUM2 are ascii hex representation of an 8-bit 
	checksum of <data>, the most significant nibble is sent first.
	the hex digits 0-9,a-f are used.

   Receiver responds with:

	+	- if CSUM is correct and ready for next packet
	-	- if CSUM is incorrect

   <data> is as follows:
   All values are encoded in ascii hex digits.

	Request		Packet

	read registers  g
	reply		XX....X		Each byte of register data
					is described by two hex digits.
					Registers are in the internal order
					for GDB, and the bytes in a register
					are in the same order the machine uses.
			or ENN		for an error.

	write regs	GXX..XX		Each byte of register data
					is described by two hex digits.
	reply		OK		for success
			ENN		for an error

        write reg	Pn...=r...	Write register n... with value r...,
					which contains two hex digits for each
					byte in the register (target byte
					order).
	reply		OK		for success
			ENN		for an error
	(not supported by all stubs).

	read mem	mAA..AA,LLLL	AA..AA is address, LLLL is length.
	reply		XX..XX		XX..XX is mem contents
					Can be fewer bytes than requested
					if able to read only part of the data.
			or ENN		NN is errno

	write mem	MAA..AA,LLLL:XX..XX
					AA..AA is address,
					LLLL is number of bytes,
					XX..XX is data
	reply		OK		for success
			ENN		for an error (this includes the case
					where only part of the data was
					written).

	cont		cAA..AA		AA..AA is address to resume
					If AA..AA is omitted,
					resume at same address.

	step		sAA..AA		AA..AA is address to resume
					If AA..AA is omitted,
					resume at same address.

	last signal     ?               Reply the current reason for stopping.
                                        This is the same reply as is generated
					for step or cont : SAA where AA is the
					signal number.

	There is no immediate reply to step or cont.
	The reply comes when the machine stops.
	It is		SAA		AA is the "signal number"

	or...		TAAn...:r...;n:r...;n...:r...;
					AA = signal number
					n... = register number
					r... = register contents
	or...		WAA		The process exited, and AA is
					the exit status.  This is only
					applicable for certains sorts of
					targets.
	kill request	k

	toggle debug	d		toggle debug flag (see 386 & 68k stubs)
	reset		r		reset -- see sparc stub.
	reserved	<other>		On other requests, the stub should
					ignore the request and send an empty
					response ($#<checksum>).  This way
					we can extend the protocol and GDB
					can tell whether the stub it is
					talking to uses the old or the new.
	search		tAA:PP,MM	Search backwards starting at address
					AA for a match with pattern PP and
					mask MM.  PP and MM are 4 bytes.
					Not supported by all stubs.

	general query	qXXXX		Request info about XXXX.
	general set	QXXXX=yyyy	Set value of XXXX to yyyy.
	query sect offs	qOffsets	Get section offsets.  Reply is
					Text=xx;Data=yy;Bss=zz
	console output	Otext		Send text to stdout.  Only comes from
					remote target.

	Responses can be run-length encoded to save space.  A '*' means that
	the next character is an ASCII encoding giving a repeat count which
	stands for that many repititions of the character preceding the '*'.
	The encoding is n+29, yielding a printable character where n >=3 
	(which is where rle starts to win).  Don't use an n > 126. 

	So 
	"0* " means the same as "0000".  */

//=======================
// Include Files
//=======================
#include "sys.h"
#include "SysMemMap.h"   // USB
#include "CatProType.h"  // USB
#include <string.h>
#include <setjmp.h>

//=========================
// Define UBC Registers
//=========================
#define REG_UBC_TOP        0x00fffa00
#define REG_UBC_BAR(ch)    (*(volatile UI32 *) (REG_UBC_TOP + 0x0000 + (ch) * 8))
#define REG_UBC_BAMR(ch)   (*(volatile UI32 *) (REG_UBC_TOP + 0x0004 + (ch) * 8))
#define REG_UBC_BRCR(ch)   (*(volatile UI16 *) (REG_UBC_TOP + 0x0028 + (ch) * 4))

//======================
// Define H8SX Registers
//======================
//
// BUFMAX defines the maximum number of characters in inbound/outbound
// buffers. At least NUMREGBYTES*2 are needed for register packets.
//
#define BUFMAX 1024
//
// Number of registers
//
//#define NUMREGS 20
#define NUMREGS 18
//
// Number of bytes for registers
//
#define NUMREGBYTES (NUMREGS * 4)
//
// Register Names
//
enum regnames
{
    er0, er1, er2, er3, er4, er5, er6, sp,
    ccr, pc, cycles, exr, inst, insts,
    mach, macl, sbr, vbr
};
//
// Register Buffer
//
UI32 registers[NUMREGBYTES / 4];

//===========================
// Hex Utilities
//===========================
static const UI08 hexchars[] = "0123456789abcdef";
//
static UI08 highhex(UI32 x)
{
    return hexchars[(x >> 4) & 0xf];
}
//
static UI08 lowhex(UI32 x)
{
    return hexchars[x & 0xf];
}
//
static SI32 hex (UI08 ch)
{
    if ((ch >= 'a') && (ch <= 'f'))
        return (SI32) (ch - 'a' + 10);
    if ((ch >= '0') && (ch <= '9'))
        return (SI32) (ch - '0');
    if ((ch >= 'A') && (ch <= 'F'))
        return (SI32) (ch - 'A' + 10);
    return (-1);
}
//
// convert the memory, pointed to by mem into hex, placing result in buf
// return a pointer to the last char put in buf (null)
//
static UI08 *mem2hex (UI08 *mem, UI08 *buf, UI32 count)
{
    UI32 i;
    UI32 ch;
    for (i = 0; i < count; i++)
    {
        ch = *mem++;
        *buf++ = highhex (ch);
        *buf++ = lowhex (ch);
    }
    *buf = 0;
    return (buf);
}
//
// convert the hex array pointed to by buf into binary, to be placed in mem
// return a pointer to the character after the last byte written
//
static UI08 *hex2mem (UI08 *buf, UI08 *mem, UI32 count)
{
    UI32 i;
    UI08 ch;
    for (i = 0; i < count; i++)
    {
        ch = (UI08) hex (*buf++) << 4;
        ch = ch + (UI08) hex (*buf++);
        *mem++ = ch;
    }
    return (mem);
}
//
// WHILE WE FIND NICE HEX CHARS, BUILD AN INT
// RETURN NUMBER OF CHARS PROCESSED
//
static UI32 hexToInt (UI08 **ptr, UI32 *intValue)
{
    UI32 numChars = 0;
    SI32 hexValue;

    *intValue = 0;

    while (**ptr)
    {
        hexValue = hex (**ptr);
        if (hexValue >= 0)
        {
            *intValue = (*intValue << 4) | (UI32) hexValue;
            numChars++;
        }
        else
            break;

        (*ptr)++;
    }
    return (numChars);
}

//===========================
// COM Port Interface
//===========================
void putDebugChar (UI08 ch)
{
    STUB_putchar(ch);
}
//
UI08 getDebugChar (void)
{
    return STUB_getchar();
}

//===========================
// Packet Utilities
//===========================
static UI08 remcomInBuffer[BUFMAX];
static UI08 remcomOutBuffer[BUFMAX];
//---------------------------------------------
// scan for the sequence $<data>#<checksum>
//---------------------------------------------
UI08 *getpacket (void)
{
    UI08 *buffer = &remcomInBuffer[0];
    UI08 checksum;
    UI08 xmitcsum;
    UI32 count;
    UI08 ch;

    while (1)
    {
        // wait around for the start character, ignore all other characters
        while ((ch = getDebugChar ()) != '$');

    retry:
        checksum = 0;
        xmitcsum = -1;
        count = 0;

        // now, read until a # or end of buffer is found
        while (count < BUFMAX - 1)
        {
            ch = getDebugChar ();
            if (ch == '$') goto retry;
            if (ch == '#') break;
            checksum = checksum + ch;
            buffer[count] = ch;
            count = count + 1;
        }
        buffer[count] = 0;

        if (ch == '#')
        {
            ch = getDebugChar ();
            xmitcsum = hex (ch) << 4;
            ch = getDebugChar ();
            xmitcsum += hex (ch);

            if (checksum != xmitcsum)
            {
                putDebugChar ('-');   // failed checksum
            }
            else
            {
                putDebugChar ('+');   // successful transfer

                // if a sequence char is present, reply the sequence ID
                if (buffer[2] == ':')
                {
                    putDebugChar (buffer[0]);
                    putDebugChar (buffer[1]);

                    return &buffer[3];
                }
                return &buffer[0];
            }
        }
    }
}
//---------------------------------------------
// send the packet in buffer.
//---------------------------------------------
static void putpacket (UI08 *buffer)
{
    UI32 checksum;
    //
    //  $<packet info>#<checksum>.
    //
    do
    {
        UI08 *src = buffer;
        putDebugChar ('$');
        checksum = 0;

        while (*src)
        {
            UI32 runlen;

            // Do run length encoding
            for (runlen = 0; runlen < 100; runlen ++)
            {
                if ((src[0] != src[runlen]) || (runlen == 99))
                {
                    if (runlen > 3)
                    {
                        int encode;
                        // Got a useful amount
                        putDebugChar (*src);
                        checksum += *src;
                        putDebugChar ('*');
                        checksum += '*';
                        checksum += (encode = runlen + ' ' - 4);
                        putDebugChar (encode);
                        src += runlen;
                    }
                    else
                    {
                        putDebugChar (*src);
                        checksum += *src;
                        src++;
                    }
                    break;
                }
            }
        }

        putDebugChar ('#');
        putDebugChar (highhex(checksum));
        putDebugChar (lowhex(checksum));
    }
    while  (getDebugChar() != '+');
}

//=======================
// UBC Controls
//=======================
#define MAX_BRK 4
typedef struct
{
    UI32 enabled;
    UI32 address;
} sUBC_Channel;
//
sUBC_Channel gsUBC[MAX_BRK];
//--------------------
// Initialize UBC
//--------------------
void Init_UBC(void)
{
    UI32 i;

    for (i = 0; i < MAX_BRK; i++)
    {
        gsUBC[i].enabled = 0;
        REG_UBC_BRCR(i) = 0x0000;
    }
}

//--------------------
// Set UBC
//--------------------
void Set_UBC(void)
{
    UI32 i;

    for (i = 0; i < MAX_BRK; i++)
    {
        if (gsUBC[i].enabled)
        {
            REG_UBC_BAMR(i) = 0xff000000;
            REG_UBC_BAR(i)  = gsUBC[i].address;
            REG_UBC_BRCR(i) = 0x0214;
        }
        else
        {
            REG_UBC_BRCR(i) = 0x00000;
        }
    }
}

//-------------------
// Clear UBC Flag
//-------------------
void Clear_UBC_Flag(void)
{
    UI32 i;
    UI16 dummy;

    for (i = 0; i < MAX_BRK; i++)
    {
        dummy = REG_UBC_BRCR(i);
        REG_UBC_BRCR(i) = 0x00000;
    }
}

//=======================
// GDB STUB Globals
//=======================
UI32  in_nmi;   // Set when handling an NMI, so we don't reenter
UI32  dofault;  // Non zero, bus errors will raise exception
UI32  dotrace;  // if 1, set T-bit in EXR when exit from Stub
UI32  stepped;  // whether, now in step mode or not
UI32 *stub_sp;  // stack area address for GDB Stub
UI32  exceptionVector; // which exception raises the Stub
UI32  gFound_Ctrl_C;   // found Ctrl-C key-in
UI32  remote_debug;    // debug > 0 prints ill-formed commands in valid packets & checksum errors
jmp_buf remcomEnv;     // jump buffer used for setjmp/longjmp
UI32  ubc_break_at_main; // UBC break at main() in startup
UI32  dokill;   // do kill command

//==============================
// GDB STUB dedicated Stack Area
//==============================
#define stub_stack_size (1 * 1024 / 4) // unit = long word
UI32 stub_stack[stub_stack_size] __attribute__ ((section (".stack_stub"))) = {0};

//============================
// Break Point
//============================
#define BREAKPOINT()   asm("trapa  #3"::);
//
void breakpoint (void)
{
    BREAKPOINT ();
}

//====================================================
// a bus error has occurred, perform a longjmp
// to return execution and allow handling of the error
//====================================================
void handle_buserror (void)
{
    longjmp (remcomEnv, 1);
}

//===============================
// Kill Command
//===============================
void kill_command(void)
{
    //-------------------
    // Wait for GDB Quit
    //-------------------
    volatile UI32 i;
    for (i = 0; i < 5000000; i++);
    //-------------------
    // Generate WDT Reset
    //-------------------
    WDT.WRITE.RSTCSR = (0x5A << 8) | (1<<6);
    WDT.WRITE.TCSR   = (0xA5 << 8) | (1<<6) | (1<<5);
    WDT.WRITE.TCSR   = (0x5A << 8) | 0xff; // write TCNT
    //-------------------
    // Wait for Reset...
    //-------------------
    while(1);
}

//=================================
// Generate Signal No. to Reply
//=================================
#define GDB_STUB_VEC_UNDEF      4
#define GDB_STUB_VEC_TRACE      5
#define GDB_STUB_VEC_NMI        7
#define GDB_STUB_VEC_TRAPA_0    8
#define GDB_STUB_VEC_TRAPA_1    9
#define GDB_STUB_VEC_TRAPA_2   10
#define GDB_STUB_VEC_TRAPA_3   11
#define GDB_STUB_VEC_CPUAERR   12
#define GDB_STUB_VEC_DMAAERR   13
#define GDB_STUB_VEC_UBC       14
#define GDB_STUB_VEC_INT_UART_Rx 161
#define GDB_STUB_VEC_INT_USB     234
//
#define GDB_STUB_SOFT_START     0
#define GDB_STUB_TERMINATE    255
//
static UI32 computeSignal (UI32 exceptionVector)
{
    UI32 sigval;
    switch (exceptionVector)
    {
        case GDB_STUB_VEC_UNDEF:
            sigval = 4;
            break;
        case GDB_STUB_VEC_TRACE:
            sigval = 5;
            break;
        case GDB_STUB_VEC_INT_USB:
        case GDB_STUB_VEC_INT_UART_Rx:
        case GDB_STUB_VEC_NMI:
            sigval = 2;
            break;
        case GDB_STUB_VEC_TRAPA_0:
        case GDB_STUB_VEC_TRAPA_1:
        case GDB_STUB_VEC_TRAPA_2:
        case GDB_STUB_VEC_TRAPA_3:
            sigval = 5;
            break;
        case GDB_STUB_VEC_CPUAERR:
        case GDB_STUB_VEC_DMAAERR:
            sigval = 10;
            break;
        case GDB_STUB_VEC_UBC:
            sigval = 5;
            break;
        case GDB_STUB_TERMINATE:
            sigval = 15;
            break;
        default:
            sigval = 7; //"software generated"
            break;
    }
    return (sigval);
}

//=========================
// Step Support
//=========================
void doSStep(void)
{
    stepped = 1;
    dotrace = 1; // Use Trace Exception H8SX CPU has.
}
//
static void undoSStep(void)
{
    if (stepped)
    {
        dotrace = 0;
    }
    stepped = 0;
}

//=============================
// GDB Handle Exception
//=============================
void gdb_handle_exception (UI32 exceptionVector)
{
    UI32 sigval, stepping;
    UI32 addr, length;
    UI08 *ptr;
    UI32 i;

    if (exceptionVector > 0)
    {
        // reply to host that an exception has occurred
        sigval = computeSignal (exceptionVector);

        remcomOutBuffer[0] = 'S';
        remcomOutBuffer[1] = highhex(sigval);
        remcomOutBuffer[2] = lowhex (sigval);
        remcomOutBuffer[3] = 0;

        putpacket (remcomOutBuffer);
    }

    // exception TRAPA indicates a software trap
    // inserted in place of code ... so back up
    // PC by one instruction, since this instruction
    // will later be replaced by its original one!
    // [Note] TRAPA #0 is used as non-replaced break
    //        instruction which make break stop at
    //        next of the TRAMA #0, while
    //        TRAPA #1 ~ TRAPA #3 is used as
    //        "to-be-replaced" break instruction.
    if (exceptionVector == GDB_STUB_VEC_TRAPA_1
     || exceptionVector == GDB_STUB_VEC_TRAPA_2
     || exceptionVector == GDB_STUB_VEC_TRAPA_3)
      registers[pc] -= 2;

    // Do the things needed to undo
    // any stepping we may have done!
    undoSStep ();

    stepping = 0;

    dokill = 0;

    while (1)
    {
        remcomOutBuffer[0] = 0;
        ptr = getpacket ();

        switch (*ptr++)
        {
            case '?':
                remcomOutBuffer[0] = 'S';
                remcomOutBuffer[1] = highhex (sigval);
                remcomOutBuffer[2] = lowhex (sigval);
                remcomOutBuffer[3] = 0;
                break;
            case 'd':
                remote_debug = !(remote_debug); // toggle debug flag
                break;
            case 'g': // return the value of the CPU registers
                mem2hex ((UI08 *) registers, remcomOutBuffer, NUMREGBYTES);
                break;
            case 'G': // set the value of the CPU registers - return OK
                hex2mem (ptr, (UI08 *) registers, NUMREGBYTES);
                strcpy ((char*) remcomOutBuffer, "OK");
                break;

            // mAA..AA,LLLL  Read LLLL bytes at address AA..AA
            case 'm':
                if (setjmp (remcomEnv) == 0)
                {
                    dofault = 0;
                    // TRY, TO READ %x,%x.  IF SUCCEED, SET PTR = 0
                    if (hexToInt (&ptr, &addr))
                        if (*(ptr++) == ',')
                            if (hexToInt (&ptr, &length))
                            {
                                ptr = 0;
                                mem2hex ((UI08 *) addr, remcomOutBuffer, length);
                            }
                    if (ptr)
                        strcpy ((char*) remcomOutBuffer, "E01");
                }
                else
                    strcpy ((char*) remcomOutBuffer, "E03");

                // restore handler for bus error
                dofault = 1;
                break;

            // MAA..AA,LLLL: Write LLLL bytes at address AA.AA return OK
            case 'M':
                if (setjmp (remcomEnv) == 0)
                {
                    dofault = 0;

                    // TRY, TO READ '%x,%x:'.  IF SUCCEED, SET PTR = 0
                    if (hexToInt (&ptr, &addr))
                        if (*(ptr++) == ',')
                            if (hexToInt (&ptr, &length))
                                if (*(ptr++) == ':')
                                {
                                    hex2mem (ptr, (UI08 *) addr, length);
                                    ptr = 0;
                                    strcpy ((char*) remcomOutBuffer, "OK");
                                }
                    if (ptr)
                        strcpy ((char*) remcomOutBuffer, "E02");
                }
                else
                    strcpy ((char*) remcomOutBuffer, "E03");

                // restore handler for bus error
                dofault = 1;
                break;

            // cAA..AA    Continue at address AA..AA(optional)
            // sAA..AA   Step one instruction from AA..AA(optional)
            case 's':
                stepping = 1;
            case 'c':
                // TRY, to read optional parameter, pc unchanged if no parm
                if (hexToInt (&ptr, &addr))
                    registers[pc] = addr;
                if (stepping)
                    doSStep ();
                //
                Set_UBC();
                return; // GOOD LUCK !!
                break;

            // kill the program
            case 'k':
                dokill = 1;
                kill_command();
            case 'R':
            case 'r':
                goto silent;

            // Set Break Point
            // use UBC for both software break and hardware break
            case 'Z':
                if (*ptr == '0' || *ptr == '1')
                {
                    ptr++;
                    if (*(ptr++) == ',')
                    {
                        if (hexToInt(&ptr, &addr))
                        {
                            if (*(ptr++) == ',')
                            {
                                // ignore length field after ','
                                // search vacant break point
                                for (i = 0; i < MAX_BRK; i++)
                                {
                                    if (gsUBC[i].enabled == 0)
                                    {
                                        gsUBC[i].enabled = 1;
                                        gsUBC[i].address = addr;
                                        break; // from for loop
                                    }
                                }
                                if (i == MAX_BRK) // not found
                                {
                                    strcpy ((char*) remcomOutBuffer, "E00");
                                }
                                else
                                {
                                    strcpy ((char*) remcomOutBuffer, "OK");
                                }
                            }
                        }
                    }
                }
                break;

                // Clear Break Point
                // use UBC for both software break and hardware break
                case 'z':
                    if (*ptr == '0' || *ptr == '1')
                    {
                        ptr++;
                        if (*(ptr++) == ',')
                        {
                            if (hexToInt(&ptr, &addr))
                            {
                                if (*(ptr++) == ',')
                                {
                                    // ignore length field after ','
                                    // search same break point
                                    for (i = 0; i < MAX_BRK; i++)
                                    {
                                        if ((gsUBC[i].enabled == 1) && (gsUBC[i].address == addr))
                                        {
                                            gsUBC[i].enabled = 0;
                                            break; // from for loop
                                        }
                                    }
                                    if (i == MAX_BRK) // not found, strange?
                                    {
                                        strcpy ((char*) remcomOutBuffer, "E00");
                                    }
                                    else
                                    {
                                        strcpy ((char*) remcomOutBuffer, "OK");
                                    }
                                }
                            }
                        }
                    }
                    break;
        }  // switch

        // kill command
        if (dokill) kill_command();

        // reply to the request
        putpacket (remcomOutBuffer);

        // silent and continue
        silent:
            continue;
    }
}

//=============================
// Handle Exception
//=============================
#define GDBCOOKIE 0x5ac
static UI32 ingdbmode;
//-----------------------------------------------------------
// We've had an exception -
// choose to go into the monitor or the gdb stub
//-----------------------------------------------------------
void handle_exception(UI32 exceptionVector)
{
    //
    // Clear UBC Flags,
    // but keep contents of gsUBC[] structure, except for 1st break at main()
    //
    Clear_UBC_Flag();
    if (ubc_break_at_main)
    {
        gsUBC[0].enabled = 0;
        gsUBC[1].enabled = 0;
        gsUBC[2].enabled = 0;
        gsUBC[3].enabled = 0;
        ubc_break_at_main = 0;
    }
    //
    // GDB STUB handler
    //
#ifdef MONITOR
    if (ingdbmode != GDBCOOKIE)
      monitor_handle_exception (exceptionVector);
    else
#endif
      gdb_handle_exception (exceptionVector);
}
//--------------------------------
// So far, not used this function
//--------------------------------
void gdb_mode (void)
{
  ingdbmode = GDBCOOKIE;
  breakpoint();
}

//=====================================
// Exception Entry Point for User Mode
//=====================================
#ifndef CONFIG_GDB_STUB_ON
void EXC_UNDEF(void) __attribute__ ((interrupt_handler));
void EXC_TRACE(void) __attribute__ ((interrupt_handler));
void INT_NMI(void) __attribute__ ((interrupt_handler));
void EXC_TRAPA_0(void) __attribute__ ((interrupt_handler));
void EXC_TRAPA_1(void) __attribute__ ((interrupt_handler));
void EXC_TRAPA_2(void) __attribute__ ((interrupt_handler));
void EXC_TRAPA_3(void) __attribute__ ((interrupt_handler));
void EXC_CPUAERR(void) __attribute__ ((interrupt_handler));
void EXC_DMAAERR(void) __attribute__ ((interrupt_handler));
void INT_UBC(void) __attribute__ ((interrupt_handler));
//
void EXC_UNDEF(void)
{
    EXC_UNDEF_Handler();
}
void EXC_TRACE(void)
{
    EXC_TRACE_Handler();
}
void INT_NMI(void)
{
    INT_NMI_Handler();
}
void EXC_TRAPA_0(void)
{
    EXC_TRAPA_0_Handler();
}
void EXC_TRAPA_1(void)
{
    EXC_TRAPA_1_Handler();
}
void EXC_TRAPA_2(void)
{
    EXC_TRAPA_2_Handler();
}
void EXC_TRAPA_3(void)
{
    EXC_TRAPA_3_Handler();
}
void EXC_CPUAERR(void)
{
    EXC_CPUAERR_Handler();
}
void EXC_DMAAERR(void)
{
    EXC_DMAAERR_Handler();
}
void INT_UBC(void)
{
    INT_UBC_Handler();
}
#endif

#ifdef CONFIG_GDB_STUB_ON
//===================================
// Exception Entry Point for GDB STUB
//===================================
void GDB_STUB_Exception_Entry(void)
{
    asm("    .global  _EXC_UNDEF");
    asm("    .global  _EXC_TRACE");
    asm("    .global  _INT_NMI");
    asm("    .global  _EXC_TRAPA_0");
    asm("    .global  _EXC_TRAPA_1");
    asm("    .global  _EXC_TRAPA_2");
    asm("    .global  _EXC_TRAPA_3");
    asm("    .global  _EXC_CPUAERR");
    asm("    .global  _EXC_DMAAERR");
    asm("    .global  _INT_UBC");
    //--------------------
    // Define Vector No.
    //--------------------
    asm("    .equ     GDB_STUB_VEC_UNDEF,     4");
    asm("    .equ     GDB_STUB_VEC_TRACE,     5");
    asm("    .equ     GDB_STUB_VEC_NMI,       7");
    asm("    .equ     GDB_STUB_VEC_TRAPA_0,   8");
    asm("    .equ     GDB_STUB_VEC_TRAPA_1,   9");
    asm("    .equ     GDB_STUB_VEC_TRAPA_2,  10");
    asm("    .equ     GDB_STUB_VEC_TRAPA_3,  11");
    asm("    .equ     GDB_STUB_VEC_CPUAERR,  12");
    asm("    .equ     GDB_STUB_VEC_DMAAERR,  13");
    asm("    .equ     GDB_STUB_VEC_UBC,      14");
    asm("    .align   2");
    //-------------------------------
    // Entry of Undefined Instruction
    //-------------------------------
    asm("_EXC_UNDEF:");
    asm("    push.l   er0");
    asm("    mov.l    #GDB_STUB_VEC_UNDEF, er0");
    asm("    mov.l    er0, @_exceptionVector");
    asm("    pop.l    er0");
    asm("    jmp      saveRegisters");
    //-------------------------------
    // Entry of TRACE
    //-------------------------------
    asm("_EXC_TRACE:");
    asm("    push.l   er0");
    asm("    mov.l    #GDB_STUB_VEC_TRACE, er0");
    asm("    mov.l    er0, @_exceptionVector");
    asm("    pop.l    er0");
    asm("    jmp      saveRegisters");
    //-------------------------------
    // Entry of NMI (prevent reenter)
    //-------------------------------
    asm("_INT_NMI:");
    asm("    push.l   er0");
    asm("    mov.l    @_in_nmi, er0");
    asm("    beq      _int_nmi_go");
    asm("    pop.l    er0");
    asm("    rte");
    //
    asm("_int_nmi_go:");
    asm("    mov.l    #1, @_in_nmi");
    asm("    mov.l    #GDB_STUB_VEC_NMI, er0");
    asm("    mov.l    er0, @_exceptionVector");
    asm("    pop.l    er0");
    asm("    jmp      saveRegisters");
    //------------------
    // Entry of TRAPA #0
    //------------------
    asm("_EXC_TRAPA_0:");
    asm("    push.l   er0");
    asm("    mov.l    #GDB_STUB_VEC_TRAPA_0, er0");
    asm("    mov.l    er0, @_exceptionVector");
    asm("    pop.l    er0");
    asm("    jmp      saveRegisters");
    //------------------
    // Entry of TRAPA #1
    //------------------
    asm("_EXC_TRAPA_1:");
    asm("    push.l   er0");
    asm("    mov.l    #GDB_STUB_VEC_TRAPA_1, er0");
    asm("    mov.l    er0, @_exceptionVector");
    asm("    pop.l    er0");
    asm("    jmp      saveRegisters");
    //------------------
    // Entry of TRAPA #2
    //------------------
    asm("_EXC_TRAPA_2:");
    asm("    push.l   er0");
    asm("    mov.l    #GDB_STUB_VEC_TRAPA_2, er0");
    asm("    mov.l    er0, @_exceptionVector");
    asm("    pop.l    er0");
    asm("    jmp      saveRegisters");
    //------------------
    // Entry of TRAPA #3
    //------------------
    asm("_EXC_TRAPA_3:");
    asm("    push.l   er0");
    asm("    mov.l    #GDB_STUB_VEC_TRAPA_3, er0");
    asm("    mov.l    er0, @_exceptionVector");
    asm("    pop.l    er0");
    asm("    jmp      saveRegisters");
    //---------------------------------------------
    // Entry of CPU Address Error (can be disabled)
    //---------------------------------------------
    asm("_EXC_CPUAERR:");
    asm("    push.l   er0");
    asm("    mov.l    @_dofault, er0");
    asm("    beq      _exc_cpuaerr_go");
    asm("    pop.l    er0");
    asm("    rte");
    //
    asm("_exc_cpuaerr_go:");
    asm("    jsr      _handle_buserror");
    // lines below have no means
    asm("    mov.l    #GDB_STUB_VEC_CPUAERR, er0");
    asm("    mov.l    er0, @_exceptionVector");
    asm("    pop.l    er0");
    asm("    jmp      saveRegisters");
    //---------------------------
    // Entry of DMA Address Error
    //---------------------------
    asm("_EXC_DMAAERR:");
    asm("    push.l   er0");
    asm("    mov.l    #GDB_STUB_VEC_DMAAERR, er0");
    asm("    mov.l    er0, @_exceptionVector");
    asm("    pop.l    er0");
    asm("    jmp      saveRegisters");
    //---------------------------
    // Entry of UBC
    //---------------------------
    asm("_INT_UBC:");
    asm("    push.l   er0");
    asm("    mov.l    #GDB_STUB_VEC_UBC, er0");
    asm("    mov.l    er0, @_exceptionVector");
    asm("    pop.l    er0");
    asm("    jmp      saveRegisters");
}
#endif

//==================================
// GDB Stub Main Entry and Exit
//==================================
void GDB_STUB_Main_Entry_Exit(void)
{
    asm("    .global  saveRegisters");
    asm("    .global  restoreRegisters");
    //--------------------
    // Save Registers
    //--------------------
    // 0x00 : er0    // 0x04 : er1
    // 0x08 : er2    // 0x0c : er3
    // 0x10 : er4    // 0x14 : er5
    // 0x18 : er6    // 0x1c : er7(sp)
    // 0x20 : ccr    // 0x24 : pc
    // 0x28 : cycles // 0x2c : exr
    // 0x30 : inst   // 0x34 : insts
    // 0x38 : mach   // 0x3c : macl
    // 0x40 : sbr    // 0x44 : vbr
    asm("saveRegisters:");
    asm("    mov.l    er0, @_registers");   // save er0
    //
    asm("    stc.b    exr, r0l");           // get exr
    asm("    and.b    #0xf8, r0l");         // clear lower 3bits (I-field)
    asm("    or.b     #0x06, r0l");         // set interrupt level to 6
    asm("    ldc.b    r0l, exr");           // set exr
    //
    asm("    mov.l    #_registers, er0");   // get pointer to save
    asm("    mov.l    er1, @(0x04, er0)");  // save er1
    asm("    mov.l    er2, @(0x08, er0)");  // save er2
    asm("    mov.l    er3, @(0x0c, er0)");  // save er3
    asm("    mov.l    er4, @(0x10, er0)");  // save er4
    asm("    mov.l    er5, @(0x14, er0)");  // save er5
    asm("    mov.l    er6, @(0x18, er0)");  // save er6
    //
    asm("    mov.l    #0, er1");            // use er1 as work
    asm("    mov.b    @er7+, r1l");         // get exr
    asm("    and.b    #0x7f, r1l");         // clear T-bit
    asm("    shll.l   #24, er1");           // align exr(8bit) on HH lane
    asm("    mov.l    er1, @(0x2c, er0)");  // save exr
    asm("    add      #1, er7");            // skip sp
    //
    asm("    mov.l    #0, er1");            // use er1 as work
    asm("    mov.b    @er7, r1L");          // get ccr
    asm("    shll.l   #24, er1");           // align ccr(8bit) on HH lane
    asm("    mov.l    er1, @(0x20, er0)");  // save ccr
    //
    asm("    mov.l    @er7, er1");          // get {ccr:8, pc:24}
    asm("    and.l    #0xffffff, er1");     // get only pc
    asm("    mov.l    er1, @(0x24, er0)");  // save pc
    //
    asm("    add      #4, er7");            // revert er7(sp)
    asm("    mov.l    er7, @(0x1c, er0)");  // save er7(sp)
    //
    asm("    stmac    mach, er1");          // get mach
    asm("    mov.l    er1, @(0x38, er0)");  // save mach
    //
    asm("    stmac    macl, er1");          // get macl
    asm("    mov.l    er1, @(0x3c, er0)");  // save macl
    //
    asm("    stc.l    sbr, er1");           // get sbr
    asm("    mov.l    er1, @(0x40, er0)");  // save sbr
    //
    asm("    stc.l    vbr, er1");           // get vbr
    asm("    mov.l    er1, @(0x44, er0)");  // save vbr
    //
    asm("    mov.l    #0, @_dotrace");          // clear dotrace flag
    asm("    mov.l    @_stub_sp, er7");         // load er7(sp)
    asm("    mov.l    @_exceptionVector, er0"); // get vector No.
    asm("    jsr      _handle_exception");      // Do stub body
    asm("    mov.l    er7, @_stub_sp");         // save er7(sp)
    //------------------------
    // Restore Registers
    //------------------------
    // 0x00 : er0    // 0x04 : er1
    // 0x08 : er2    // 0x0c : er3
    // 0x10 : er4    // 0x14 : er5
    // 0x18 : er6    // 0x1c : er7(sp)
    // 0x20 : ccr    // 0x24 : pc
    // 0x28 : cycles // 0x2c : exr
    // 0x30 : inst   // 0x34 : insts
    // 0x38 : mach   // 0x3c : macl
    // 0x40 : sbr    // 0x44 : vbr
    asm("restoreRegisters:");
    asm("    mov.l    #_registers, er0");   // get pointer to restore
    //
    asm("    mov.l    @(0x1c, er0), er7");  // restore er7(sp) as user stack
    //
    asm("    mov.l    @(0x24, er0), er1");  // take saved pc
    asm("    mov.l    @(0x20, er0), er2");  // take saved ccr (aligned on MSB side)
    asm("    or.l     er2, er1");           // make {ccr:8, pc:24}
    asm("    push.l   er1");                // push {ccr:8, pc:24} in user stack
    //
    asm("    mov.l    @(0x2c, er0), er1");  // take saved exr
    asm("    shlr.l   #16, er1");           // align exr(8bit) on LH lane
    asm("    mov.l    @_dotrace, er2");     // get dotrace
    asm("    bld      #0, r2l");            // get LSB of dotrace in C
    asm("    bst      #7, r1h");            // set dotrace in T-bit of exr in LH lane
    asm("    push.w   r1");                 // push exr in user stack (LL lane is dummy)
    //
    asm("    mov.l    @(0x44, er0), er1");  // take saved vbr
    asm("    ldc.l    er1, vbr");           // restore vbr
    //
    asm("    mov.l    @(0x40, er0), er1");  // take saved sbr
    asm("    ldc.l    er1, sbr");           // restore sbr
    //
    asm("    mov.l    @(0x3c, er0), er1");  // take saved macl
    asm("    ldmac    er1, macl");          // restore macl
    //
    asm("    mov.l    @(0x38, er0), er1");  // take saved mach
    asm("    ldmac    er1, mach");          // restore mach
    //
    asm("    mov.l    @(0x18, er0), er6");  // restore er6
    asm("    mov.l    @(0x14, er0), er5");  // restore er5
    asm("    mov.l    @(0x10, er0), er4");  // restore er4
    asm("    mov.l    @(0x0c, er0), er3");  // restore er3
    asm("    mov.l    @(0x08, er0), er2");  // restore er2
    asm("    mov.l    @(0x04, er0), er1");  // restore er1
    //
    asm("    mov.b    #0, @_in_nmi");       // in_nmi = 0;
    //
    asm("    mov.l    @_registers, er0");   // restore er0
    //
    asm("    rte");                         // restore exr, ccr, and pc
}

//============================
// USB Rx Check Ctrl-C
//============================
UI32 USB_Rx_Check_CTRL_C(void)
{
    UI32 result;
    //-----------------------------------
    // Set Default Value
    //-----------------------------------
    result = 0;
#ifdef CONFIG_GDB_STUB_USB_ON
    //------------------
    // Enabled follows
    //------------------
    UI08 ch;
    UI08 *readPtr;
    UI08 *writePtr;
    //---------------------------------
    // Get Current USB Rx FIFO Pointers
    //---------------------------------
    readPtr  = dataBuffGPtr[BULKOUT_IDX].readPtr;
    writePtr = dataBuffGPtr[BULKOUT_IDX].writePtr;
    //----------------------------------
    // If Rx has detected, Found Ctrl-C?
    //----------------------------------
    while (readPtr != writePtr)
    {
        ch = *readPtr;
        if (ch == 0x03) // Ctrl-C ?
        {
            result = 1;
            //-----------------------------------------------
            // Skip Read Pointer, and abandon data in Rx FIFO
            //-----------------------------------------------
            dataBuffGPtr[BULKOUT_IDX].readPtr = readPtr;
            break;
        }
        //
        if (readPtr >= dataBuffGPtr[BULKOUT_IDX].endAriaPtr)
        {
            readPtr = (UI08*) dataBuffGPtr[BULKOUT_IDX].beginAriaPtr;
        }
        else
        {
            readPtr++;
        }
    }
#endif
    return result;
}

//============================
// UART Rx Check Ctrl-C
//============================
UI32 UART_Rx_Check_CTRL_C(void)
{
    UI32 result;
    //-----------------------------------
    // Set Default Value
    //-----------------------------------
    result = 0;
#ifndef CONFIG_GDB_STUB_USB_ON
    //------------------
    // Enabled follows
    //------------------
    UI08 ch;
    //-----------------
    // Check Ctrl-C
    //-----------------
    if (UART_Found_Ctrl_C())
    {
        result = 1;
        //------------------------
        // Abandon data in Rx FIFO
        //------------------------
        while(UART_Rx(&ch));
    }
#endif
    return result;
}

//=============================
// GDB STUB USB Interrupt Entry
//=============================
void GDB_STUB_INT_USB_ENTRY(void) // dummy function name
{
    asm("    .global  _INT_USB_ENTRY");
    //--------------------
    // Define Vector No.
    //--------------------
    asm("    .equ     GDB_STUB_VEC_INT_USB, 234");
    //--------------------
    // Entry of INT_USB
    //--------------------
    asm("    .align   2");
    asm("_INT_USB_ENTRY:");
    asm("    mov.l    er6, @-er7");
    asm("    stm.l    er4-er5, @-er7");
    asm("    stm.l    er0-er3, @-er7");
    //----------------------------------
    // USB Interrupt Handle
    //----------------------------------
    INT_USB_Handler(); // normal subroutine
#ifdef CONFIG_GDB_STUB_ON
    //-----------------
    // Found Ctrl-C?
    //-----------------
    gFound_Ctrl_C = USB_Rx_Check_CTRL_C();
    //
    asm("    mov.l    @_gFound_Ctrl_C, er0");
    asm("    beq      INT_USB_EXIT"); // not found
    //-----------------------------------------
    // If Detect Ctrl-C, GDB Stub takes Control
    //-----------------------------------------
    asm("    mov.l    #GDB_STUB_VEC_INT_USB, @_exceptionVector");
    asm("    ldm.l    @er7+, er0-er3");
    asm("    ldm.l    @er7+, er4-er5");
    asm("    mov.l    @er7+, er6");
    asm("    jmp      saveRegisters:24");
#endif
    //--------------------
    // Exit
    //--------------------
    asm("INT_USB_EXIT:");
    asm("    ldm.l    @er7+, er0-er3");
    asm("    ldm.l    @er7+, er4-er5");
    asm("    mov.l    @er7+, er6");
    asm("    rte");
}

//=================================
// GDB STUB UART Rx Interrupt Entry
//=================================
void GDB_STUB_INT_UART_Rx_ENTRY(void) // dummy function name
{
    asm("    .global  _INT_UART_Rx_ENTRY");
    //--------------------
    // Define Vector No.
    //--------------------
    asm("    .equ     GDB_STUB_VEC_INT_UART_Rx, 161");
    //----------------------
    // Entry of INT_UART_Rx
    //----------------------
    asm("    .align   2");
    asm("_INT_UART_Rx_ENTRY:");
    asm("    mov.l    er6, @-er7");
    asm("    stm.l    er4-er5, @-er7");
    asm("    stm.l    er0-er3, @-er7");
    //--------------------------
    // UART Rx Interrupt Handler
    //--------------------------
    Int_Handler_UART_Rx(); // normal subroutine
#ifdef CONFIG_GDB_STUB_ON
    //-----------------
    // Found Ctrl-C?
    //-----------------
    gFound_Ctrl_C = UART_Rx_Check_CTRL_C();
    //
    asm("    mov.l    @_gFound_Ctrl_C, er0");
    asm("    beq      INT_UART_Rx_EXIT"); // not found
    //-----------------------------------------
    // If Detect Ctrl-C, GDB Stub takes Control
    //-----------------------------------------
    asm("    mov.l    #GDB_STUB_VEC_INT_UART_Rx, @_exceptionVector");
    asm("    ldm.l    @er7+, er0-er3");
    asm("    ldm.l    @er7+, er4-er5");
    asm("    mov.l    @er7+, er6");
    asm("    jmp      saveRegisters:24");
#endif
    //--------------------
    // Exit
    //--------------------
    asm("INT_UART_Rx_EXIT:");
    asm("    ldm.l    @er7+, er0-er3");
    asm("    ldm.l    @er7+, er4-er5");
    asm("    mov.l    @er7+, er6");
    asm("    rte");
}

//============================================
// GDB Stub Main Routine
//============================================
extern int main(void);
//
void GDB_STUB_main(void)
{
    //----------------------------
    // Initialize GDB STUB Globals
    //----------------------------
    in_nmi = 0;
    dofault = 1;
    dotrace = 0;
    stepped = 0;
    stub_sp = stub_stack + stub_stack_size;

#ifdef CONFIG_GDB_STUB_ON
    //----------------------------
    // Use GDB Stub, Forever Loop
    //----------------------------
    while(1)
    {
        //---------------------------
        // Initialize System Hardware
        //---------------------------
        // Set All Internal Clocks in Maximum Speed
        // EXTAL = 12MHz, MD_CLK = 0
        // ICLK = EXTAL x 4 (48MHz)
        // PCLK = EXTAL x 2 (24MHz)
        // BCLK = EXTAL x 4 (48MHz)
        SCKCR.BIT.ICK = 0;
        SCKCR.BIT.PCK = 1;
        SCKCR.BIT.BCK = 0;
        //---------------------------
        // Initialize INTC
        //---------------------------
        INTC.INTCR.BIT.INTM = 2;
        set_imask_exr(0);
#ifdef CONFIG_GDB_STUB_USB_ON
        //---------------------------
        // Initialize USB
        //---------------------------
        Init_USB();
        INTC.IPRR.BIT._USB  = 7;
#else
        //---------------------------
        // Initialize UART
        //---------------------------
        Init_UART();
        INTC.IPRL.BIT._SCI4 = 7;
#endif
        //---------------------------
        // Initialize UBC
        //---------------------------
        MSTP.CRB.BIT._UBC = 0;
        //----------------------------
        // Initialize GDB STUB Globals
        //----------------------------
        in_nmi = 0;
        dofault = 1;
        dotrace = 0;
        stepped = 0;
        stub_sp = stub_stack + stub_stack_size;
        //------------------------------------------------------------
        // Set UBC to break at main() (wide traps after setting frame)
        //------------------------------------------------------------
        Init_UBC();
        gsUBC[0].enabled = 1;
        gsUBC[1].enabled = 1;
        gsUBC[2].enabled = 1;
        gsUBC[3].enabled = 1;
        gsUBC[0].address = ((UI32) main + 0x0006);
        gsUBC[1].address = ((UI32) main + 0x0008);
        gsUBC[2].address = ((UI32) main + 0x000a);
        gsUBC[3].address = ((UI32) main + 0x000c);
        Set_UBC();
        ubc_break_at_main = 1;
        //-------------------
        // User Main Routine
        //-------------------
        main();
    }
#else
    //-------------------------
    // Do main()
    //-------------------------
    main();
    while(1);
#endif
}

//======================================================
// End of Program
//======================================================
