//=========================================================
// H8SX/1655 Project
//=========================================================
// File Name : oscilloscope.c
// Function  : Application of Oscilloscope
//---------------------------------------------------------
// Rev.01 2009.10.12 Takanaga Yamazaki
//---------------------------------------------------------
// Copyright (C) 2009, 2010 Takanaga Yamazaki
//=========================================================
// ---- License Information -------------------------------
// Anyone can FREELY use this code fully or partially
// under conditions shown below.
// 1. You may use this code only for individual purpose,
//    and educational purpose.
//    Do not use this code for business even if partially.
// 2. You should use this code under the GNU GPL.
// 3. You should remain this header text in your codes
//   including Copyright credit and License Information.
// 4. Your codes should inherit this license information.
//=========================================================
// ---- Patent Notice -------------------------------------
// I have not cared whether this system (hw + sw) causes 
// infringement on the patent, copyright, trademark,
// or trade secret rights of others. You have all 
// responsibilities for determining if your designs
// and products infringe on the intellectual property
// rights of others, when you use technical information
// included in this system for your business.
//=========================================================
// ---- Disclaimers ---------------------------------------
// The function and reliability of this system are not 
// guaranteed. They may cause any damages to loss of
// properties, data, money, profits, life, or business.
// By adopting this system even partially, you assume
// all responsibility for its use.
//=========================================================

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "application.h"
#include "api.h"

//========================
// RTC Global
//========================
extern sDateTime gDateTime;  // Current Clock Database

//=========================
// CH1/CH2 V-Scale Controls
//=========================
#define CH1_INDEX_MAX 2
#define CH2_INDEX_MAX 2
#define VSCALE_INDEX_MAX  6
//
volatile UI32 gCH1_INDEX = 0; // Current CH1 Mode Index
volatile UI32 gCH2_INDEX = 0; // Current CH1 Mode Index
//
volatile UI32 gCH1_VSCALE_INDEX = 2; // CH1 V-Scale Mode Index
volatile UI32 gCH2_VSCALE_INDEX = 2; // CH2 V-Scale Mode Index
//
const UI08* CH1_STR[] = {(UI08*) "CH1", (UI08*) "off"};
const UI08* CH2_STR[] = {(UI08*) "CH2", (UI08*) "off"};
const UI08* VSCALE_STR[] =
{
    (UI08*) "5V", (UI08*) "2V", (UI08*) "1V",
    (UI08*) "0.5V", (UI08*) "0.2V", (UI08*) "0.1V"
};
const FLOT VSCALE_MUL[] = // vertical pixel counts corresponding to 1LSB of A/D
{                         // (25pixels = 1div)
    ((3.3/1024.0) * (25 / 5)),
    ((3.3/1024.0) * (25 / 2)),
    ((3.3/1024.0) * (25 / 1)),
    ((3.3/1024.0) * (25 / 0.5)),
    ((3.3/1024.0) * (25 / 0.2)),
    ((3.3/1024.0) * (25 / 0.1))
};

//====================
// Trigger Controls
//====================
#define TRG_INDEX_MAX 5
#define TRG_LEVEL_MIN 0
#define TRG_LEVEL_MAX 3.3
//
volatile UI32 gTRG_INDEX = 0;    // Trigger Mode Index
volatile FLOT gTRG_fLevel = 1.0; // Trigger Voltage Level
//
const UI08* TRG_STR[]   =
{
    (UI08*) "Free",
    (UI08*) "CH1+", (UI08*) "CH1-", (UI08*) "CH2+", (UI08*) "CH2-"
};
const UI08  TRG_CH[]    = {0, 1, 1, 2, 2}; // Trigger Channel
const UI08  TRG_SLOPE[] = {0, 0, 0, 1, 1}; // 0:positive, 1:negative
//
#define SMA_MAX 4 // Simple Moving Average (SMA) Length (2^n preferred)
volatile UI16 gCH1_Wave_for_SMA[SMA_MAX]; // Wave Memory to calculate SMA of CH1
volatile UI16 gCH2_Wave_for_SMA[SMA_MAX]; // Wave Memory to calculate SMA of CH2
volatile UI16 gTRG_Value; // Trigger Level by A/D 10bit value
volatile UI08 gDetect_trigger = 0; // Trigger Detected, or not

//====================
// T-Scale Controls
//====================
#define TSCALE_INDEX_MAX 15
#define HOLD_OFF 20 // Timing Gap between Sweeps
//
volatile UI32 gTSCALE_INDEX = 2; // T-Scale Mode Index
//
const UI08* TSCALE_STR[] = {
    (UI08*) "0.4ms", (UI08*) "0.8ms", (UI08*) "1ms",
    (UI08*) "2ms",   (UI08*) "4ms",   (UI08*) "8ms",   (UI08*) "10ms",
    (UI08*) "20ms",  (UI08*) "40ms",  (UI08*) "80ms",  (UI08*) "0.1s",
    (UI08*) "0.2s",  (UI08*) "0.4s",  (UI08*) "0.8s",  (UI08*) "1s"
};
const UI32 TSCALE_SPL[] = // Storage A/D Sampled data once every this cycles.
{
    (    400 / 25 / 16),
    (    800 / 25 / 16),
    (   1000 / 25 / 20),
    (   2000 / 25 / 16),
    (   4000 / 25 / 16),
    (   8000 / 25 / 16),
    (  10000 / 25 / 16),
    (  20000 / 25 / 16),
    (  40000 / 25 / 16),
    (  80000 / 25 / 16),
    ( 100000 / 25 / 16),
    ( 200000 / 25 / 16),
    ( 400000 / 25 / 16),
    ( 800000 / 25 / 16),
    (1000000 / 25 / 16)
};
const UI08 TSCALE_TMR0_CYC[] =  // Sampling Cycle ... 0:16us, 1:20us
{
    0, 0, 1,
    0, 0, 0, 0,
    0, 0, 0, 0,
    0, 0, 0, 0
};

//==========================
// Run/Capture Mode Control
//==========================
#define CAPTURE_INDEX_MAX 2
//
volatile UI32 gCAPTURE_INDEX = 0; // Run/Capture Mode Index
//
const UI08* CAPTURE_STR[] = {(UI08*) "RUN", (UI08*) "CAP"};

//======================
// GND Level Controls
//======================
volatile UI32 gCH1_GND =  85; // Vertical Position of CH1 GND Level
volatile UI32 gCH2_GND = 185; // Vertical Position of CH2 GND Level

//======================
// Wave Storages
//======================
volatile UI16 gCH1_Wave[300]; // Wave Memory of CH1
volatile UI16 gCH2_Wave[300]; // Wave Memory of CH1
volatile UI16 gCH1_Wave_Prev_posY[300]; // Previous Vertical Position of CH1 to erase
volatile UI16 gCH2_Wave_Prev_posY[300]; // Previous Vertical Position of CH2 to erase
volatile UI16 gCH1_Wave_Prev_Pixel[300]; // Previous Pixel Value of CH1 to erase
volatile UI16 gCH2_Wave_Prev_Pixel[300]; // Previous Pixel Value of CH2 to erase
//
volatile UI32 gSample_Count = 0; // Sample Wave when this value is 0
volatile UI32 gSample_Index = 0; // Wave Memory Index to sample
volatile UI32 gCapturing = 0;    // Now, do Capturing Waves, or not
//
volatile UI32 gWave_Index = 0; // Wave Memory Index to draw

//==============================
// Calibration Waves
//==============================
#define CAL_WAVE_SAMPLE_COUNT 125
volatile UI08 gCH1_Cal_Wave[CAL_WAVE_SAMPLE_COUNT];
volatile UI08 gCH2_Cal_Wave[CAL_WAVE_SAMPLE_COUNT];

//===============================
// Oscilloscope Do Sampling
// (Called by Int_Handler_AD0())
//===============================
void OSCILLOSCOPE_Sampling(void)
{
    UI08 i;
    static UI08 HOLD_OFF_Count = 0;
    static UI16 CH1_SMA_prev;
    static UI16 CH2_SMA_prev;
    UI16 CH1_SMA_curr;
    UI16 CH2_SMA_curr;
    UI16 CH1_Value;
    UI16 CH2_Value;

    //---------------------------
    // If not in capturing, quit
    //---------------------------
    if (gCapturing == 0) // may be in drawing
    {
        return;
    }
    //--------------
    // Get A/D Data
    //--------------
    CH1_Value = ADC_Get_Data(0);
    CH2_Value = ADC_Get_Data(1);
    //----------------------------
    // SMA (Simple Moving Average)
    //----------------------------
    CH1_SMA_curr = 0;
    CH2_SMA_curr = 0;
    //
    for (i = 0; i < (SMA_MAX - 1); i++)
    {
        gCH1_Wave_for_SMA[i] = gCH1_Wave_for_SMA[i + 1];
        gCH2_Wave_for_SMA[i] = gCH2_Wave_for_SMA[i + 1];
        //
        CH1_SMA_curr = CH1_SMA_curr + (UI32) gCH1_Wave_for_SMA[i];
        CH2_SMA_curr = CH2_SMA_curr + (UI32) gCH2_Wave_for_SMA[i];
    }
    gCH1_Wave_for_SMA[SMA_MAX - 1] = CH1_Value;
    gCH2_Wave_for_SMA[SMA_MAX - 1] = CH2_Value;
    //
    CH1_SMA_curr = CH1_SMA_curr + (UI32) CH1_Value;
    CH2_SMA_curr = CH2_SMA_curr + (UI32) CH2_Value;
    //
    CH1_SMA_curr = CH1_SMA_curr / SMA_MAX;
    CH2_SMA_curr = CH2_SMA_curr / SMA_MAX;
    //---------------------------
    // Hold Off Timing
    // (And Wait for Stable SMA)
    //---------------------------
    if (HOLD_OFF_Count < HOLD_OFF)
    {
        CH1_SMA_prev = CH1_SMA_curr; // Save SMA Value
        CH2_SMA_prev = CH2_SMA_curr; // Save SMA Value
        //
        HOLD_OFF_Count++;
        return;
    }
    //-------------------------
    // Trigger Operation
    //-------------------------
    switch(gTRG_INDEX)
    {
        //-------------------
        // Free Run Trigger
        //-------------------
        case 0 :
        {
            gDetect_trigger = 1;
            break;
        }
        //-----------------
        // CH1+ Trigger
        //-----------------
        case 1 :
        {
            if (gDetect_trigger == 0)
            {
                if ((CH1_SMA_prev < gTRG_Value) && (CH1_SMA_curr >= gTRG_Value))
                {
                    gDetect_trigger = 1;
                }
            }
            break;
        }
        //-----------------
        // CH1- Trigger
        //-----------------
        case 2 :
        {
            if (gDetect_trigger == 0)
            {
                if ((CH1_SMA_prev > gTRG_Value) && (CH1_SMA_curr <= gTRG_Value))
                {
                    gDetect_trigger = 1;
                }
            }
            break;
        }
        //-----------------
        // CH2+ Trigger
        //-----------------
        case 3 :
        {
            if (gDetect_trigger == 0)
            {
                if ((CH2_SMA_prev < gTRG_Value) && (CH2_SMA_curr >= gTRG_Value))
                {
                    gDetect_trigger = 1;
                }
            }
            break;
        }
        //-----------------
        // CH2- Trigger
        //-----------------
        case 4 :
        {
            if (gDetect_trigger == 0)
            {
                if ((CH2_SMA_prev > gTRG_Value) && (CH2_SMA_curr <= gTRG_Value))
                {
                    gDetect_trigger = 1;
                }
            }
            break;
        }
        //-----------------
        // Others
        //-----------------
        default :
        {
            gDetect_trigger = 0;
            break;
        }
    }
    //-----------------
    // Save SMA Value
    //-----------------
    CH1_SMA_prev = CH1_SMA_curr;
    CH2_SMA_prev = CH2_SMA_curr;
    //---------------------------
    // Do Sampling if Triggered
    //---------------------------
    if (gDetect_trigger == 1)
    {
        if (gSample_Count == 0)
        {
            gCH1_Wave[gSample_Index] = CH1_Value;
            gCH2_Wave[gSample_Index] = CH2_Value;
            //
            gSample_Index++;
            if (gSample_Index >= 300)
            {
                gCapturing = 0;
                gSample_Count = 0;
                gSample_Index = 0;
                //
                HOLD_OFF_Count = 0;
                gDetect_trigger = 0;
            }
        }
        //
        gSample_Count++;
        if (gSample_Count >= TSCALE_SPL[gTSCALE_INDEX])
        {
            gSample_Count = 0;
        }
    }
}

//===========================
// Oscilloscope Main
//===========================
void OSCILLOSCOPE_main(void)
{
    UI32 i;
    sFrame  sFMEM;
    UI32 x, y;
    UI32 touch, touch_x, touch_y;
    UI08* TRG_sLevel[8];
    UI32 CH1_GND_prev;
    UI32 CH2_GND_prev;
    UI08 stop_redraw = 0;
    //
    sButton sButton_CH1;
    sButton sButton_CH1_SCALE_UP;
    sButton sButton_CH1_SCALE_DN;
    sButton sButton_CH2;
    sButton sButton_CH2_SCALE_UP;
    sButton sButton_CH2_SCALE_DN;
    sButton sButton_TRG;
    sButton sButton_TRG_VOLTAGE_UP;
    sButton sButton_TRG_VOLTAGE_DN;
    sButton sButton_TIME_SCALE_UP;
    sButton sButton_TIME_SCALE_DN;
    sButton sButton_CAPTURE;
    //--------------------------------------
    // Initialize Frame Memory and TFT Panel
    //--------------------------------------
    Init_TFT_Panel(&sFMEM, BLK);
    //-------------------
    // Set TMR0 Cycle
    //-------------------
    Init_TMR0(TSCALE_TMR0_CYC[gTSCALE_INDEX]);
    //--------------------------
    // Prepare Calibration Wave
    //--------------------------
    OSCILLOSCOPE_Init_Cal_Wave();
    //-----------------------
    // Initialize Wave Data
    //-----------------------
    for (i = 0; i < 300; i++)
    {
        gCH1_Wave[i] = 0;
        gCH2_Wave[i] = 0;
        gCH1_Wave_Prev_posY[i] = gCH1_GND;
        gCH2_Wave_Prev_posY[i] = gCH2_GND;
        gCH1_Wave_Prev_Pixel[i] = BLU;
        gCH2_Wave_Prev_Pixel[i] = BLU;
    }
    //--------------------
    // Clear SMA Buffer
    //--------------------
    for (i = 0; i < SMA_MAX; i++)
    {
        gCH1_Wave_for_SMA[i] = 0;
        gCH2_Wave_for_SMA[i] = 0;
    }
    //-----------------
    // Draw Scale Lines
    //-----------------
    for (x = 10; x <= 310; x = x + 25)
    {
        GC_Draw_Line(&sFMEM, x, x, 10, 185, BLU);
    }
    for (y = 10; y <= 185; y = y + 25)
    {
        GC_Draw_Line(&sFMEM, 10, 310, y, y, BLU);
    }
    //------------------------------------
    // Configure sButton_CH1
    //------------------------------------
    sButton_CH1.posX  =  10;
    sButton_CH1.posY  = 200;
    sButton_CH1.sizeX =  30;
    sButton_CH1.sizeY =  15;
    sButton_CH1.pString = (UI08*) CH1_STR[gCH1_INDEX];
    sButton_CH1.color_button_off = DGR;
    sButton_CH1.color_string_off = YEL;
    sButton_CH1.color_button_on  = GRY;
    sButton_CH1.color_string_on  = BLK;
    sButton_CH1.redraw = 1;
    sButton_CH1.int_status = 0;
    //------------------------------------
    // Configure sButton_CH1_SCALE_UP
    //------------------------------------
    sButton_CH1_SCALE_UP.posX  =  45;
    sButton_CH1_SCALE_UP.posY  = 200;
    sButton_CH1_SCALE_UP.sizeX =  20;
    sButton_CH1_SCALE_UP.sizeY =  15;
    sButton_CH1_SCALE_UP.pString = (UI08*) "UP";
    sButton_CH1_SCALE_UP.color_button_off = DGR;
    sButton_CH1_SCALE_UP.color_string_off = YEL;
    sButton_CH1_SCALE_UP.color_button_on  = GRY;
    sButton_CH1_SCALE_UP.color_string_on  = BLK;
    sButton_CH1_SCALE_UP.redraw = 1;;
    sButton_CH1_SCALE_UP.int_status = 0;
    //------------------------------------
    // Configure sButton_CH1_SCALE_DN
    //------------------------------------
    sButton_CH1_SCALE_DN.posX  =  45;
    sButton_CH1_SCALE_DN.posY  = 220;
    sButton_CH1_SCALE_DN.sizeX =  20;
    sButton_CH1_SCALE_DN.sizeY =  15;
    sButton_CH1_SCALE_DN.pString = (UI08*) "DN";
    sButton_CH1_SCALE_DN.color_button_off = DGR;
    sButton_CH1_SCALE_DN.color_string_off = YEL;
    sButton_CH1_SCALE_DN.color_button_on  = GRY;
    sButton_CH1_SCALE_DN.color_string_on  = BLK;
    sButton_CH1_SCALE_DN.redraw = 1;;
    sButton_CH1_SCALE_DN.int_status = 0;
    //------------------------------------
    // Configure sButton_CH2
    //------------------------------------
    sButton_CH2.posX  =  80;
    sButton_CH2.posY  = 200;
    sButton_CH2.sizeX =  30;
    sButton_CH2.sizeY =  15;
    sButton_CH2.pString = (UI08*) CH2_STR[gCH2_INDEX];
    sButton_CH2.color_button_off = DGR;
    sButton_CH2.color_string_off = YEL;
    sButton_CH2.color_button_on  = GRY;
    sButton_CH2.color_string_on  = BLK;
    sButton_CH2.redraw = 1;;
    sButton_CH2.int_status = 0;
    //------------------------------------
    // Configure sButton_CH2_SCALE_UP
    //------------------------------------
    sButton_CH2_SCALE_UP.posX  = 115;
    sButton_CH2_SCALE_UP.posY  = 200;
    sButton_CH2_SCALE_UP.sizeX =  20;
    sButton_CH2_SCALE_UP.sizeY =  15;
    sButton_CH2_SCALE_UP.pString = (UI08*) "UP";
    sButton_CH2_SCALE_UP.color_button_off = DGR;
    sButton_CH2_SCALE_UP.color_string_off = YEL;
    sButton_CH2_SCALE_UP.color_button_on  = GRY;
    sButton_CH2_SCALE_UP.color_string_on  = BLK;
    sButton_CH2_SCALE_UP.redraw = 1;;
    sButton_CH2_SCALE_UP.int_status = 0;
    //------------------------------------
    // Configure sButton_CH2_SCALE_DN
    //------------------------------------
    sButton_CH2_SCALE_DN.posX  = 115;
    sButton_CH2_SCALE_DN.posY  = 220;
    sButton_CH2_SCALE_DN.sizeX =  20;
    sButton_CH2_SCALE_DN.sizeY =  15;
    sButton_CH2_SCALE_DN.pString = (UI08*) "DN";
    sButton_CH2_SCALE_DN.color_button_off = DGR;
    sButton_CH2_SCALE_DN.color_string_off = YEL;
    sButton_CH2_SCALE_DN.color_button_on  = GRY;
    sButton_CH2_SCALE_DN.color_string_on  = BLK;
    sButton_CH2_SCALE_DN.redraw = 1;;
    sButton_CH2_SCALE_DN.int_status = 0;
    //------------------------------------
    // Configure sButton_TRG
    //------------------------------------
    sButton_TRG.posX  = 150;
    sButton_TRG.posY  = 200;
    sButton_TRG.sizeX =  35;
    sButton_TRG.sizeY =  15;
    sButton_TRG.pString = (UI08*) TRG_STR[gTRG_INDEX];
    sButton_TRG.color_button_off = DGR;
    sButton_TRG.color_string_off = YEL;
    sButton_TRG.color_button_on  = GRY;
    sButton_TRG.color_string_on  = BLK;
    sButton_TRG.redraw = 1;;
    sButton_TRG.int_status = 0;
    //------------------------------------
    // Configure sButton_TRG_VOLTAGE_UP
    //------------------------------------
    sButton_TRG_VOLTAGE_UP.posX  = 190;
    sButton_TRG_VOLTAGE_UP.posY  = 200;
    sButton_TRG_VOLTAGE_UP.sizeX =  20;
    sButton_TRG_VOLTAGE_UP.sizeY =  15;
    sButton_TRG_VOLTAGE_UP.pString = (UI08*) "UP";
    sButton_TRG_VOLTAGE_UP.color_button_off = DGR;
    sButton_TRG_VOLTAGE_UP.color_string_off = YEL;
    sButton_TRG_VOLTAGE_UP.color_button_on  = GRY;
    sButton_TRG_VOLTAGE_UP.color_string_on  = BLK;
    sButton_TRG_VOLTAGE_UP.redraw = 1;;
    sButton_TRG_VOLTAGE_UP.int_status = 0;
    //------------------------------------
    // Configure sButton_TRG_VOLTAGE_DN
    //------------------------------------
    sButton_TRG_VOLTAGE_DN.posX  = 190;
    sButton_TRG_VOLTAGE_DN.posY  = 220;
    sButton_TRG_VOLTAGE_DN.sizeX =  20;
    sButton_TRG_VOLTAGE_DN.sizeY =  15;
    sButton_TRG_VOLTAGE_DN.pString = (UI08*) "DN";
    sButton_TRG_VOLTAGE_DN.color_button_off = DGR;
    sButton_TRG_VOLTAGE_DN.color_string_off = YEL;
    sButton_TRG_VOLTAGE_DN.color_button_on  = GRY;
    sButton_TRG_VOLTAGE_DN.color_string_on  = BLK;
    sButton_TRG_VOLTAGE_DN.redraw = 1;;
    sButton_TRG_VOLTAGE_DN.int_status = 0;
    //------------------------------------
    // Configure sButton_TIME_SCALE_DN
    //------------------------------------
    sButton_TIME_SCALE_DN.posX  = 225;
    sButton_TIME_SCALE_DN.posY  = 200;
    sButton_TIME_SCALE_DN.sizeX =  20;
    sButton_TIME_SCALE_DN.sizeY =  15;
    sButton_TIME_SCALE_DN.pString = (UI08*) "DN";
    sButton_TIME_SCALE_DN.color_button_off = DGR;
    sButton_TIME_SCALE_DN.color_string_off = YEL;
    sButton_TIME_SCALE_DN.color_button_on  = GRY;
    sButton_TIME_SCALE_DN.color_string_on  = BLK;
    sButton_TIME_SCALE_DN.redraw = 1;;
    sButton_TIME_SCALE_DN.int_status = 0;
    //------------------------------------
    // Configure sButton_TIME_SCALE_UP
    //------------------------------------
    sButton_TIME_SCALE_UP.posX  = 250;
    sButton_TIME_SCALE_UP.posY  = 200;
    sButton_TIME_SCALE_UP.sizeX =  20;
    sButton_TIME_SCALE_UP.sizeY =  15;
    sButton_TIME_SCALE_UP.pString = (UI08*) "UP";
    sButton_TIME_SCALE_UP.color_button_off = DGR;
    sButton_TIME_SCALE_UP.color_string_off = YEL;
    sButton_TIME_SCALE_UP.color_button_on  = GRY;
    sButton_TIME_SCALE_UP.color_string_on  = BLK;
    sButton_TIME_SCALE_UP.redraw = 1;;
    sButton_TIME_SCALE_UP.int_status = 0;
    //------------------------------------
    // Configure sButton_CAPTURE
    //------------------------------------
    sButton_CAPTURE.posX  = 285;
    sButton_CAPTURE.posY  = 200;
    sButton_CAPTURE.sizeX =  25;
    sButton_CAPTURE.sizeY =  35;
    sButton_CAPTURE.pString = (UI08*) CAPTURE_STR[gCAPTURE_INDEX];
    sButton_CAPTURE.color_button_off = DGR;
    sButton_CAPTURE.color_string_off = YEL;
    sButton_CAPTURE.color_button_on  = GRY;
    sButton_CAPTURE.color_string_on  = BLK;
    sButton_CAPTURE.redraw = 1;
    sButton_CAPTURE.int_status = 0;
    //------------------------------------
    // Draw Control Panel
    //------------------------------------
    GC_Draw_Fill_Rect(&sFMEM,   5, 195,  65,  45, WHT); // CH1
    GC_Draw_Fill_Rect(&sFMEM,  75, 195,  65,  45, RED); // CH2
    GC_Draw_Fill_Rect(&sFMEM, 145, 195,  70,  45, CYN); // TRG
    GC_Draw_Fill_Rect(&sFMEM, 220, 195,  55,  45, BLU); // TIME
    GC_Draw_Fill_Rect(&sFMEM, 280, 195,  35,  45, YEL); // CAPTURE
    //------------------------------------
    // Draw Parameters
    //------------------------------------
    GC_Draw_String((UI08 *) VSCALE_STR[gCH1_VSCALE_INDEX], &sFMEM, 15, 220, BLK);
    GC_Draw_String((UI08 *) VSCALE_STR[gCH2_VSCALE_INDEX], &sFMEM, 85, 220, BLK);
    sprintf((char*) TRG_sLevel, "%3.1fV", gTRG_fLevel);
    GC_Draw_String((UI08*) TRG_sLevel,  &sFMEM, 155,  220, BLK);
    GC_Draw_String((UI08 *) TSCALE_STR[gTSCALE_INDEX], &sFMEM, 230, 220, BLK);
    //------------------------------------
    // Draw GND Marker
    //------------------------------------
    GC_Draw_Fill_Rect(&sFMEM, 0, gCH1_GND - 3,  10,  6, WHT); // CH1
    GC_Draw_Fill_Rect(&sFMEM, 0, gCH2_GND - 3,  10,  6, RED); // CH2
    CH1_GND_prev = gCH1_GND;
    CH2_GND_prev = gCH2_GND;
    //------------------------------------
    // Main Loop
    //------------------------------------
    gSample_Count = 0;
    gSample_Index = 0;
    gWave_Index = 0;
    gCapturing = 1;
    //
    while(1)
    {
        //----------------
        // Draw Waves
        //----------------
        UI32 CH1_posX,  CH2_posX;
        UI32 CH1_posY,  CH2_posY;
        FLOT CH1_fPosY, CH2_fPosY;
        //
        if ((stop_redraw == 0) && (gCapturing == 0))
        {
            for (i = 0; i < 10; i++) // to make wave drawing fast
            {
                //
                // Update posX
                //
                CH1_posX = gWave_Index + 10;
                CH2_posX = gWave_Index + 10;
                //
                // Erase Dots
                //
                GC_Draw_Dot(&sFMEM, CH1_posX, gCH1_Wave_Prev_posY[gWave_Index], gCH1_Wave_Prev_Pixel[gWave_Index]);
                if (gCH1_Wave_Prev_posY[gWave_Index] != gCH2_Wave_Prev_posY[gWave_Index])
                {
                    GC_Draw_Dot(&sFMEM, CH2_posX, gCH2_Wave_Prev_posY[gWave_Index], gCH2_Wave_Prev_Pixel[gWave_Index]);
                }
                //
                // Get posY
                //
                CH1_fPosY = (FLOT) gCH1_GND - (FLOT) gCH1_Wave[gWave_Index] * VSCALE_MUL[gCH1_VSCALE_INDEX];
                CH2_fPosY = (FLOT) gCH2_GND - (FLOT) gCH2_Wave[gWave_Index] * VSCALE_MUL[gCH2_VSCALE_INDEX];
                CH1_posY = (UI32) CH1_fPosY;
                CH2_posY = (UI32) CH2_fPosY;
                //
                // Draw CH1 Dots with storing previous pixel information
                //
                if ((CH1_fPosY >= 10) && (CH1_fPosY <= 185))
                {
                    gCH1_Wave_Prev_Pixel[gWave_Index] = GC_Get_Pixel(&sFMEM, CH1_posX, CH1_posY);
                    gCH1_Wave_Prev_posY[gWave_Index] = CH1_posY;
                    //
                    if (gCH1_INDEX == 0)
                    {
                        GC_Draw_Dot(&sFMEM, CH1_posX, CH1_posY, WHT);
                    }
                }
                //
                // Draw CH2 Dots with storing previous pixel information
                //
                if ((CH2_fPosY >= 10) && (CH2_fPosY <= 185))
                {
                    gCH2_Wave_Prev_Pixel[gWave_Index] = GC_Get_Pixel(&sFMEM, CH2_posX, CH2_posY);
                    gCH2_Wave_Prev_posY[gWave_Index] = CH2_posY;
                    //
                    if (gCH2_INDEX == 0)
                    {
                        if (CH2_posY != CH1_posY)
                        {
                            GC_Draw_Dot(&sFMEM, CH2_posX, CH2_posY, RED);
                        }
                    }
                }
                //
                // Increment Time Index
                //
                gWave_Index++;
                if (gWave_Index >= 300)
                {
                    gWave_Index = 0;
                    gCapturing = 1;
                    //
                    if (gCAPTURE_INDEX) stop_redraw = 1;
                }
            }
        }
        //-------------------
        // GND Marker CH1/CH2
        //-------------------
        touch = TOUCH_Get_Info(&touch_x, &touch_y);
        //
        if ((touch) && (touch_x < 20) && (touch_y > gCH1_GND - 10) && (touch_y < gCH1_GND + 10))
        {
            while((touch = TOUCH_Get_Info(&touch_x, &touch_y)))
            {
                if ((touch_y > 10) && (touch_y < 187))
                {
                    gCH1_GND = touch_y;
                    if (gCH1_GND != CH1_GND_prev)
                    {
                        GC_Draw_Fill_Rect(&sFMEM, 0, CH1_GND_prev - 3,  10,  6, BLK);
                        GC_Draw_Fill_Rect(&sFMEM, 0, gCH1_GND - 3,  10,  6, WHT);
                        GC_Draw_Fill_Rect(&sFMEM, 0, gCH2_GND - 3,  10,  6, RED);
                        CH1_GND_prev = gCH1_GND;
                    }
                }
            }
        }
        //
        if ((touch) && (touch_x < 20) && (touch_y > gCH2_GND - 10) && (touch_y < gCH2_GND + 10))
        {
            while((touch = TOUCH_Get_Info(&touch_x, &touch_y)))
            {
                if ((touch_y > 10) && (touch_y < 187))
                {
                    gCH2_GND = touch_y;
                    if (gCH2_GND != CH2_GND_prev)
                    {
                        GC_Draw_Fill_Rect(&sFMEM, 0, CH2_GND_prev - 3,  10,  6, BLK);
                        GC_Draw_Fill_Rect(&sFMEM, 0, gCH1_GND - 3,  10,  6, WHT);
                        GC_Draw_Fill_Rect(&sFMEM, 0, gCH2_GND - 3,  10,  6, RED);
                        CH2_GND_prev = gCH2_GND;
                    }
                }
            }
        }
        //---------------------------------
        // Button : CH1 Control and VScale
        //---------------------------------
        if (GC_Button(&sButton_CH1, &sFMEM) == BUTTON_UP)
        {
            gCH1_INDEX = (gCH1_INDEX == (CH1_INDEX_MAX - 1))? 0 : gCH1_INDEX  + 1;
            sButton_CH1.pString = (UI08*) CH1_STR[gCH1_INDEX];
            sButton_CH1.redraw = 1;
            GC_Button(&sButton_CH1, &sFMEM);
            //
            gWave_Index = 0;
            gCapturing = 1;
        }
        //
        if (GC_Button(&sButton_CH1_SCALE_UP, &sFMEM) == BUTTON_UP)
        {
            gCH1_VSCALE_INDEX = (gCH1_VSCALE_INDEX == (VSCALE_INDEX_MAX - 1))? gCH1_VSCALE_INDEX : gCH1_VSCALE_INDEX  + 1;
            GC_Draw_Fill_Rect(&sFMEM, 15, 220, 30, 15, WHT);
            GC_Draw_String((UI08 *) VSCALE_STR[gCH1_VSCALE_INDEX], &sFMEM, 15, 220, BLK);
            //
            gWave_Index = 0;
            gCapturing = 1;
        }
        //
        if (GC_Button(&sButton_CH1_SCALE_DN, &sFMEM) == BUTTON_UP)
        {
            gCH1_VSCALE_INDEX = (gCH1_VSCALE_INDEX == 0)? gCH1_VSCALE_INDEX : gCH1_VSCALE_INDEX  - 1;
            GC_Draw_Fill_Rect(&sFMEM, 15, 220, 30, 15, WHT);
            GC_Draw_String((UI08 *) VSCALE_STR[gCH1_VSCALE_INDEX], &sFMEM, 15, 220, BLK);
            //
            gWave_Index = 0;
            gCapturing = 1;
        }
        //--------------------------------
        // Button : CH2 Control and VScale
        //--------------------------------
        if (GC_Button(&sButton_CH2, &sFMEM) == BUTTON_UP)
        {
            gCH2_INDEX = (gCH2_INDEX == (CH2_INDEX_MAX - 1))? 0 : gCH2_INDEX  + 1;
            sButton_CH2.pString = (UI08*) CH2_STR[gCH2_INDEX];
            sButton_CH2.redraw = 1;
            GC_Button(&sButton_CH2, &sFMEM);
            //
            gWave_Index = 0;
            gCapturing = 1;
        }
        //
        if (GC_Button(&sButton_CH2_SCALE_UP, &sFMEM) == BUTTON_UP)
        {
            gCH2_VSCALE_INDEX = (gCH2_VSCALE_INDEX == (VSCALE_INDEX_MAX - 1))? gCH2_VSCALE_INDEX : gCH2_VSCALE_INDEX  + 1;
            GC_Draw_Fill_Rect(&sFMEM, 85, 220, 30, 15, RED);
            GC_Draw_String((UI08 *) VSCALE_STR[gCH2_VSCALE_INDEX], &sFMEM, 85, 220, BLK);
            //
            gWave_Index = 0;
            gCapturing = 1;
        }
        //
        if (GC_Button(&sButton_CH2_SCALE_DN, &sFMEM) == BUTTON_UP)
        {
            gCH2_VSCALE_INDEX = (gCH2_VSCALE_INDEX == 0)? gCH2_VSCALE_INDEX : gCH2_VSCALE_INDEX  - 1;
            GC_Draw_Fill_Rect(&sFMEM, 85, 220, 30, 15, RED);
            GC_Draw_String((UI08 *) VSCALE_STR[gCH2_VSCALE_INDEX], &sFMEM, 85, 220, BLK);
            //
            gWave_Index = 0;
            gCapturing = 1;
        }
        //-------------------------------------------
        // Button : Trigger Control and Trigger Level
        //-------------------------------------------
        if (GC_Button(&sButton_TRG, &sFMEM) == BUTTON_UP)
        {
            gTRG_INDEX = (gTRG_INDEX == (TRG_INDEX_MAX - 1))? 0 : gTRG_INDEX  + 1;
            sButton_TRG.pString = (UI08*) TRG_STR[gTRG_INDEX];
            sButton_TRG.redraw = 1;
            GC_Button(&sButton_TRG, &sFMEM);
            //
            gTRG_Value =  (UI16) (gTRG_fLevel * 1023 / 3.3);
            //
            gWave_Index = 0;
            gCapturing = 1;
            //
            gDetect_trigger = 0;
        }
        //
        if (GC_Button(&sButton_TRG_VOLTAGE_UP, &sFMEM) == BUTTON_UP)
        {
            gTRG_fLevel = gTRG_fLevel + 0.1;
            if (gTRG_fLevel > TRG_LEVEL_MAX) gTRG_fLevel = TRG_LEVEL_MAX;
            sprintf((char*) TRG_sLevel, "%3.1fV", gTRG_fLevel);
            GC_Draw_Fill_Rect(&sFMEM, 155, 220, 30, 15, CYN);
            GC_Draw_String((UI08*) TRG_sLevel,  &sFMEM, 155,  220, BLK);
            //
            gTRG_Value =  (UI16) (gTRG_fLevel * 1023 / 3.3);
            //
            gWave_Index = 0;
            gCapturing = 1;
            //
            gDetect_trigger = 0;
        }
        //
        if (GC_Button(&sButton_TRG_VOLTAGE_DN, &sFMEM) == BUTTON_UP)
        {
            gTRG_fLevel = gTRG_fLevel - 0.1;
            if (gTRG_fLevel < TRG_LEVEL_MIN) gTRG_fLevel = TRG_LEVEL_MIN;
            sprintf((char*) TRG_sLevel, "%3.1fV", gTRG_fLevel);
            GC_Draw_Fill_Rect(&sFMEM, 155, 220, 30, 15, CYN);
            GC_Draw_String((UI08*) TRG_sLevel,  &sFMEM, 155,  220, BLK);
            //
            gTRG_Value =  (UI16) (gTRG_fLevel * 1023 / 3.3);
            //
            gWave_Index = 0;
            gCapturing = 1;
            //
            gDetect_trigger = 0;
        }
        //---------------------
        // Button : Time Scale
        //---------------------
        if (GC_Button(&sButton_TIME_SCALE_UP, &sFMEM) == BUTTON_UP)
        {
            gTSCALE_INDEX = (gTSCALE_INDEX == (TSCALE_INDEX_MAX - 1))? gTSCALE_INDEX : gTSCALE_INDEX  + 1;
            GC_Draw_Fill_Rect(&sFMEM, 230, 220, 40, 15, BLU);
            GC_Draw_String((UI08 *) TSCALE_STR[gTSCALE_INDEX], &sFMEM, 230, 220, BLK);
            //
            Init_TMR0(TSCALE_TMR0_CYC[gTSCALE_INDEX]);
            //
            gWave_Index = 0;
            gCapturing = 1;
        }
        //
        if (GC_Button(&sButton_TIME_SCALE_DN, &sFMEM) == BUTTON_UP)
        {
            gTSCALE_INDEX = (gTSCALE_INDEX == 0)? gTSCALE_INDEX : gTSCALE_INDEX - 1;
            GC_Draw_Fill_Rect(&sFMEM, 230, 220, 40, 15, BLU);
            GC_Draw_String((UI08 *) TSCALE_STR[gTSCALE_INDEX], &sFMEM, 230, 220, BLK);
            //
            Init_TMR0(TSCALE_TMR0_CYC[gTSCALE_INDEX]);
            //
            gWave_Index = 0;
            gCapturing = 1;
        }
        //----------------------------
        // Button : Capture and Sample
        //----------------------------
        if (GC_Button(&sButton_CAPTURE, &sFMEM) == BUTTON_UP)
        {
            gCAPTURE_INDEX = (gCAPTURE_INDEX == (CAPTURE_INDEX_MAX - 1))? 0 : gCAPTURE_INDEX  + 1;
            sButton_CAPTURE.pString = (UI08*) CAPTURE_STR[gCAPTURE_INDEX];
            sButton_CAPTURE.redraw = 1;
            GC_Button(&sButton_CAPTURE, &sFMEM);
            //
            if (gCAPTURE_INDEX == 0) stop_redraw = 0;
        }
        //------------------
        // RTC Display
        //------------------
#if 1
        UI08 *WEEK[] = {(UI08*)"SUN", (UI08*)"MON", (UI08*)"TUE", (UI08*)"WED", (UI08*)"THU", (UI08*)"FRI", (UI08*)"SAT"};
        UI08 str0[64];
        UI08 str1[64];
        UI08 sec_prev = 255;
        //
        if ((gCapturing == 0) && (gWave_Index == 0))
        {
            if (RTC_Check_Power())
            {
                RTC_Get(&gDateTime);
                if (gDateTime.bcd_seconds != sec_prev)
                {
                    sprintf((char*) str0, "[H8SX]  %02x:%02x:%02x", gDateTime.bcd_hours, gDateTime.bcd_minutes, gDateTime.bcd_seconds);
                    sprintf((char*) str1, "20%02x/%02x/%02x (%s)", gDateTime.bcd_years, gDateTime.bcd_months, gDateTime.bcd_days, WEEK[gDateTime.weekdays - 1]);
                    //
                    LCD1602_Print_String(0, str0);
                    LCD1602_Print_String(1, str1);
                    //
                    sec_prev = gDateTime.bcd_seconds;
                }
            }
            else
            {
                LCD1602_Print_String(0, (UI08*) "[H8SX] ERROR !");
                LCD1602_Print_String(1, (UI08*) "RTC Power Fail.");
                while(1);
            }
        }
#endif
    }
}

//==========================================
// Oscilloscope Initialize Calibration Waves
//==========================================
void OSCILLOSCOPE_Init_Cal_Wave(void)
{
    UI32 i;
    FLOT fSin;
    FLOT fTri;

    for (i = 0; i < CAL_WAVE_SAMPLE_COUNT; i++)
    {
        fSin = (1.5 + 1.5 * _SIN_(2 * _PI_ * (FLOT) i / CAL_WAVE_SAMPLE_COUNT)) * 255 / 3.3;
        fTri = (i < (CAL_WAVE_SAMPLE_COUNT / 2))?
               ((FLOT) i / (CAL_WAVE_SAMPLE_COUNT / 2)) * 255 * 3.0 / 3.3 :
               ((FLOT) (CAL_WAVE_SAMPLE_COUNT - i) / (CAL_WAVE_SAMPLE_COUNT / 2)) * 255 * 3.0 / 3.3;
        fSin = (fSin < 0)? 0.0 : (fSin > 255)? 255.0 : fSin;
        fTri = (fTri < 0)? 0.0 : (fTri > 255)? 255.0 : fTri;
        gCH1_Cal_Wave[i] = (UI08) fSin;
        gCH2_Cal_Wave[i] = (UI08) fTri;
    }
}

//=====================================
// Oscilloscope Calibration Wave Output
//=====================================
void OSCILLOSCOPE_Cal_Out(void)
{
    static UI32 i = 0;

    DA.DADR0H = gCH1_Cal_Wave[i];
    DA.DADR1H = gCH2_Cal_Wave[i];

    i++;
    if (i >= CAL_WAVE_SAMPLE_COUNT) i = 0;
}

//======================================================
// End of Program
//======================================================

