//=========================================================
// H8SX/1655 Project
//=========================================================
// File Name : uart.c
// Function  : Control UART (SCI4)
//---------------------------------------------------------
// Rev.01 2009.09.11 Takanaga Yamazaki
//---------------------------------------------------------
// Copyright (C) 2009, 2010 Takanaga Yamazaki
//=========================================================
// ---- License Information -------------------------------
// Anyone can FREELY use this code fully or partially
// under conditions shown below.
// 1. You may use this code only for individual purpose,
//    and educational purpose.
//    Do not use this code for business even if partially.
// 2. You should use this code under the GNU GPL.
// 3. You should remain this header text in your codes
//   including Copyright credit and License Information.
// 4. Your codes should inherit this license information.
//=========================================================
// ---- Patent Notice -------------------------------------
// I have not cared whether this system (hw + sw) causes 
// infringement on the patent, copyright, trademark,
// or trade secret rights of others. You have all 
// responsibilities for determining if your designs
// and products infringe on the intellectual property
// rights of others, when you use technical information
// included in this system for your business.
//=========================================================
// ---- Disclaimers ---------------------------------------
// The function and reliability of this system are not 
// guaranteed. They may cause any damages to loss of
// properties, data, money, profits, life, or business.
// By adopting this system even partially, you assume
// all responsibility for its use.
//=========================================================

#include "sys.h"

//=============================
// Global : UART Software FIFO
//=============================
#define UART_FIFO_DEPTH 256          // UART FIFO Depth
UI08 gUART_FIFO_Tx[UART_FIFO_DEPTH]; // Tx FIFO Buffer
UI32 gUART_FIFO_Tx_WP;               // Tx FIFO Read Pointer
UI32 gUART_FIFO_Tx_RP;               // Tx FIFO Write Pointer
UI32 gUART_FIFO_Tx_DC;               // Tx FIFO Data Count;
UI08 gUART_FIFO_Rx[UART_FIFO_DEPTH]; // Rx FIFO Buffer
UI32 gUART_FIFO_Rx_WP;               // Rx FIFO Read Pointer
UI32 gUART_FIFO_Rx_RP;               // Rx FIFO Write Pointer
UI32 gUART_FIFO_Rx_DC;               // Rx FIFO Data Count;

//======================
// Global : Ctrl-C
//======================
UI08 gUART_Ctrl_C;

//=============================
// Initialize UART (SCI4)
//=============================
void Init_UART(void)
{
    volatile UI32 i;
    static UI08 uart_init_done = 0;
    //---------------------------
    // Prevent Re-Initialization
    //---------------------------
    if (uart_init_done) return;
    uart_init_done = 1;
    //------------------------
    // P60 : SCI4 TxD4
    // P61 : SCI4 RxD4
    //------------------------
    P6.ICR.BIT.B1 = 1;
    //------------------------
    // Start SCI4 System Clock
    //------------------------
    MSTP.CRB.BIT._SCI4 = 0;
    //------------------------
    // Initialize FIFO
    //------------------------
    gUART_FIFO_Tx_WP = 0;
    gUART_FIFO_Tx_RP = 0;
    gUART_FIFO_Tx_DC = 0;
    gUART_FIFO_Rx_WP = 0;
    gUART_FIFO_Rx_RP = 0;
    gUART_FIFO_Rx_DC = 0;
    //------------------------
    // Initialize Global
    //------------------------
    gUART_Ctrl_C = 0;
    //-----------------------------
    // Set Interrupt Request Level
    //-----------------------------
    INTC.IPRL.BIT._SCI4 = 6;
    //------------------------
    // P60 : TxD4
    // P61 : RxD4
    //------------------------
    SCI4.SCR.BIT.TE  = 0;
    SCI4.SCR.BIT.RE  = 0;
    //------------------------
    // UART Mode
    //   8bit, No-Parity, 1-Stop
    //   Baud Rate = 38400bps
    //   PCLK = 24MHz
    //   CKS  = 0
    //   BRR = 24M / (64 * 2^(-1) * 38400) - 1 = 19
    //------------------------
    SCI4.SCR.BIT.CKE = 0;
    SCI4.SMR.BYTE = 0x00;  //00000000
    SCI4.BRR = 19;
    //------------------------
    // Wait for 1bit period (30us)
    //------------------------
    for (i = 0; i < 30000 / 20; i++);
    //------------------------
    // Enable Interrupt RXI
    //------------------------
    SCI4.SCR.BIT.RIE = 1;
    //------------------------
    // Set TE, RE
    //------------------------
    SCI4.SCR.BIT.TE  = 1;
    SCI4.SCR.BIT.RE  = 1;
}

//=====================
// UART Tx
//=====================
UI08 UART_Tx_Possible(void)
{
    UI08 possible;
    UI08 tie_bkup;

    //----------------------------------------------
    // Disable Interrupt TXI (still TDRE flag is ON)
    //----------------------------------------------
    tie_bkup = SCI4.SCR.BIT.TIE;
    SCI4.SCR.BIT.TIE = 0;
    //------------------------
    // Check FIFO Tx
    //------------------------
    possible = (gUART_FIFO_Tx_DC < UART_FIFO_DEPTH)? 1 : 0;
    //------------------------
    // Revert TIE
    //------------------------
    SCI4.SCR.BIT.TIE = tie_bkup;

    return possible;
}
//
UI08 UART_Tx(UI08 byte)
{
    UI08 possible;
    UI08 tie_bkup;

    //----------------------------------------------
    // Disable Interrupt TXI (still TDRE flag is ON)
    //----------------------------------------------
    tie_bkup = SCI4.SCR.BIT.TIE;
    //------------------------
    // Check FIFO Tx
    //------------------------
    possible = (gUART_FIFO_Tx_DC < UART_FIFO_DEPTH)? 1 : 0;
    //------------------------
    // Set the Tx Data in FIFO
    //------------------------
    if (possible)
    {
        gUART_FIFO_Tx[gUART_FIFO_Tx_WP] = byte;
        gUART_FIFO_Tx_WP = (gUART_FIFO_Tx_WP + 1) % UART_FIFO_DEPTH;
        gUART_FIFO_Tx_DC = gUART_FIFO_Tx_DC + 1;
        tie_bkup = 1;
    }
    //------------------------
    // Revert TIE
    //------------------------
    SCI4.SCR.BIT.TIE = tie_bkup;

    return possible;
}
//
void UART_Tx_Char(UI08 byte)
{
    while(UART_Tx(byte) == 0);
}

//=====================
// UART Rx
//=====================
UI08 UART_Rx_Possible(void)
{
    UI08 possible;
    UI08 rie_bkup;

    //------------------------
    // Disable Interrupt RXI
    //------------------------
    rie_bkup = SCI4.SCR.BIT.RIE;
    SCI4.SCR.BIT.RIE = 0;
    //------------------------
    // Check FIFO Rx
    //------------------------
    possible = (gUART_FIFO_Rx_DC > 0)? 1 : 0;
    //------------------------
    // Revert RIE
    //------------------------
    SCI4.SCR.BIT.RIE = rie_bkup;

    return possible;
}
//
UI08 UART_Rx(UI08 *byte)
{
    UI08 possible;
    UI08 rie_bkup;

    //------------------------
    // Disable Interrupt RXI
    //------------------------
    rie_bkup = SCI4.SCR.BIT.RIE;
    //------------------------
    // Check FIFO Rx
    //------------------------
    possible = (gUART_FIFO_Rx_DC > 0)? 1 : 0;
    //---------------------------
    // Get the Rx Data from FIFO
    //---------------------------
    if (possible)
    {
        *byte = gUART_FIFO_Rx[gUART_FIFO_Rx_RP];
        gUART_FIFO_Rx_RP = (gUART_FIFO_Rx_RP + 1) % UART_FIFO_DEPTH;
        gUART_FIFO_Rx_DC = gUART_FIFO_Rx_DC - 1;
        rie_bkup = 1;
    }
    //------------------------
    // Revert RIE
    //------------------------
    SCI4.SCR.BIT.RIE = rie_bkup;

    return possible;
}
//
UI08 UART_Rx_Char(void)
{
    UI08 byte;

    while(UART_Rx(&byte) == 0);
    return byte;
}

//============================
// Check if Found Ctrl-C ?
//============================
UI08 UART_Found_Ctrl_C(void)
{
    UI08 result;

    result = gUART_Ctrl_C;
    gUART_Ctrl_C = 0;
    return result;
}

//=============================
// Interrupt Handler : UART Tx
//=============================
void Int_Handler_UART_Tx(void)
{
    //------------------------
    // If FIFO Tx has Data
    //------------------------
    if (gUART_FIFO_Tx_DC > 0)
    {
        // Read FIFO Tx and Set TDR
        SCI4.TDR = gUART_FIFO_Tx[gUART_FIFO_Tx_RP];
        gUART_FIFO_Tx_RP = (gUART_FIFO_Tx_RP + 1) % UART_FIFO_DEPTH;
        gUART_FIFO_Tx_DC = gUART_FIFO_Tx_DC - 1;
        // Clear TDRE
        SCI4.SSR.BIT.TDRE = 0; // read 1 and write 0
    }
    //-------------------------
    // If FIFO Tx has no Data
    //-------------------------
    else
    {
        // Disable Interrupt TXI (still TDRE flag is ON)
        SCI4.SCR.BIT.TIE = 0;
    }
}

//=============================
// Interrupt Handler : UART Rx
//=============================
void Int_Handler_UART_Rx(void)
{
    UI08 byte;

    //------------------------
    // If FIFO Rx has Room
    //------------------------
    if (gUART_FIFO_Rx_DC < UART_FIFO_DEPTH)
    {
        // Read RDR and Set FIFO Rx
        byte = SCI4.RDR;
        if (byte == 0x03) // Ctrl-C ?
        {
            gUART_Ctrl_C = 1;
        }
        gUART_FIFO_Rx[gUART_FIFO_Rx_WP] = byte;
        gUART_FIFO_Rx_WP = (gUART_FIFO_Rx_WP + 1) % UART_FIFO_DEPTH;
        gUART_FIFO_Rx_DC = gUART_FIFO_Rx_DC + 1;
        // Clear RDRF
        SCI4.SSR.BIT.RDRF = 0; // read 1 and write 0
    }
    //------------------------
    // If FIFO Rx has no Room
    //------------------------
    else
    {
        // Disable Interrupt RXI (still RDRF flag is ON)
        SCI4.SCR.BIT.RIE = 0;
    }
}

//==================================
// Interrupt Handler : UART Rx Error
//==================================
void Int_Handler_UART_Rx_ERR(void)
{
    SCI4.SSR.BIT.FER = 0;
    SCI4.SSR.BIT.ORER = 0;
    SCI4.SSR.BIT.PER = 0;
    SCI4.SSR.BIT.RDRF = 0;
}

//==================================
// Interrupt Handler : UART Tx End
//==================================
void Int_Handler_UART_Tx_END(void)
{
    SCI4.SSR.BIT.TEND = 0;
}

//======================================================
// End of Program
//======================================================
